import I8BitIoDevice;

class CVHandCtrls implements I8BitIoDevice
{
	private static final char m_kpdResetMask		=	0xF0FF;
	private static final char m_kpd0Mask			=	0x0000;
	private static final char m_kpd1Mask			=	0x0001;
	private static final char m_kpd2Mask			=	0x0002;
	private static final char m_kpd3Mask			=	0x0003;
	private static final char m_kpd4Mask			=	0x0004;
	private static final char m_kpd5Mask			=	0x0005;
	private static final char m_kpd6Mask			=	0x0006;
	private static final char m_kpd7Mask			=	0x0007;
	private static final char m_kpd8Mask			=	0x0008;
	private static final char m_kpd9Mask			=	0x0009;
	private static final char m_kpdAsteriskMask	=	0x000A;
	private static final char m_kpdPoundMask		=	0x000B;
	private static final char m_kpdFire1Mask		=	0x0040;
	private static final char m_kpdFire2Mask		=	0x4000;

	// Joystick position constants
	private static final char joyResetMask		=  0x0F00;
	private static final char joyUpMask			=	0x0100;
	private static final char joyRightMask		=	0x0200;
	private static final char joyDownMask		=	0x0400;
	private static final char joyLeftMask		=	0x0800;

	public static final char m_CV_JOYPOS_CENTER		= 0;
	public static final char m_CV_JOYPOS_NORTH		= 1;
	public static final char m_CV_JOYPOS_SOUTH		= 2;
	public static final char m_CV_JOYPOS_EAST			= 3;
	public static final char m_CV_JOYPOS_WEST			= 4;
	public static final char m_CV_JOYPOS_NORTHEAST	= 5;
	public static final char m_CV_JOYPOS_NORTHWEST	= 6;
	public static final char m_CV_JOYPOS_SOUTHEAST	= 7;
	public static final char m_CV_JOYPOS_SOUTHWEST	= 8;

	private static final char m_CV_NUM_JOYDIRECTIONS	= 8;	// Not inluding "CENTER"

	public static final char m_CV_NUMPAD_NOTHING		= 0;
	public static final char m_CV_NUMPAD_0				= 1;
	public static final char m_CV_NUMPAD_1				= 2;
	public static final char m_CV_NUMPAD_2				= 3;
	public static final char m_CV_NUMPAD_3				= 4;
	public static final char m_CV_NUMPAD_4				= 5;
	public static final char m_CV_NUMPAD_5				= 6;
	public static final char m_CV_NUMPAD_6				= 7;
	public static final char m_CV_NUMPAD_7				= 8;
	public static final char m_CV_NUMPAD_8				= 9;
	public static final char m_CV_NUMPAD_9				= 10;
	public static final char m_CV_NUMPAD_POUND		= 11;
	public static final char m_CV_NUMPAD_ASTERISK	= 12;
	public static final char m_CV_FIRE1					= 13;
	public static final char m_CV_FIRE2					= 14;

	private static final char m_CV_NUM_KEYPAD_KEYS	= 14;	//	Don't inlude "NOTHING" type

	public static final char m_CV_BUTTON_PRESSED		= 0;
	public static final char m_CV_BUTTON_NOTPRESSED	= 1;

	public static final char m_CV_PRIMARY_CONTROLLER = 0;
	public static final char m_CV_SECONDARY_CONTROLLER = 1;

	static char m_rgbKeyCodes[] =
	{
		(char)0x0A,
		(char)0x0D,
		(char)0x07,
		(char)0x0C,
		(char)0x02,
		(char)0x03,
		(char)0x0E,
		(char)0x05,
		(char)0x01,
		(char)0x0B,
		(char)0x06,
		(char)0x09,
		(char)0x0F,
		(char)0x0F,
		(char)0x0F,
		(char)0x0F
	};

	static	char	m_rgwJoyPosMasks[] =
	{
		(char)(~joyResetMask),		// CV_JOYPOS_CENTER
		joyUpMask,						//	CV_JOYPOS_NORTH
		joyDownMask,					//	CV_JOYPOS_SOUTH
		joyRightMask,					//	CV_JOYPOS_EAST
		joyLeftMask,					//	CV_JOYPOS_WEST
		joyUpMask | joyRightMask,	//	CV_JOYPOS_NORTHEAST
		joyUpMask | joyLeftMask,	//	CV_JOYPOS_NORTHWEST
		joyDownMask | joyRightMask,//	CV_JOYPOS_SOUTHEAST
		joyDownMask | joyLeftMask	//	CV_JOYPOS_SOUTHWEST
	};

	static	char	m_rgwKeypadStateMasks[] =
	{
		0,						// Filler for CV_NUMPAD_NOTHING
		m_kpd0Mask,			//
		m_kpd1Mask,			//
		m_kpd2Mask,			//
		m_kpd3Mask,			//
		m_kpd4Mask,			//
		m_kpd5Mask,			//
		m_kpd6Mask,			//
		m_kpd7Mask,			//
		m_kpd8Mask,			//
		m_kpd9Mask,			//
		m_kpdPoundMask,		//
		m_kpdAsteriskMask,	//
		m_kpdFire1Mask,		//
		m_kpdFire2Mask		//
	};


	static char	m_rgwJoyPosMap[] =
	{
		m_CV_JOYPOS_CENTER,		// 1111, 0000
		m_CV_JOYPOS_NORTH,		// 1110, 0001
		m_CV_JOYPOS_EAST,		// 1101, 0010
		m_CV_JOYPOS_NORTHEAST,	// 1100, 0011
		m_CV_JOYPOS_SOUTH,		// 1011, 0100
		m_CV_JOYPOS_CENTER,		//	1010, 0101	// Invalid bit pattern
		m_CV_JOYPOS_SOUTHEAST,	// 1001, 0110
		m_CV_JOYPOS_CENTER,		//	1000, 0111	// Invalid bit pattern
		m_CV_JOYPOS_WEST,		// 0111, 1000
		m_CV_JOYPOS_NORTHWEST,	// 0110, 1001
		m_CV_JOYPOS_CENTER,		//	0101, 1010	// Invalid bit pattern
		m_CV_JOYPOS_CENTER,		//	0100, 1011	// Invalid bit pattern
		m_CV_JOYPOS_SOUTHWEST,	//	0011, 1100
		m_CV_JOYPOS_CENTER,		// 0010, 1101	// Invalid bit pattern
		m_CV_JOYPOS_CENTER,		// 0001, 1110	// Invalid bit pattern
		m_CV_JOYPOS_CENTER,		// 0000, 1111	// Invalid bit pattern
	};


	static char	m_kpdKeyMap[] =
	{
		m_CV_NUMPAD_NOTHING,
		m_CV_NUMPAD_NOTHING,
		m_CV_NUMPAD_NOTHING,
		m_CV_NUMPAD_NOTHING,
		m_CV_NUMPAD_ASTERISK,
		m_CV_NUMPAD_POUND,
		m_CV_NUMPAD_9,
		m_CV_NUMPAD_8,
		m_CV_NUMPAD_7,
		m_CV_NUMPAD_6,
		m_CV_NUMPAD_5,
		m_CV_NUMPAD_4,
		m_CV_NUMPAD_3,
		m_CV_NUMPAD_2,
		m_CV_NUMPAD_1,
		m_CV_NUMPAD_0
	};

	private char m_nPortBase = 0;
	private char m_bJoyMode = 0;	// Joyst. controller mode
	private char m_rgwJoyState[] = { 0, 0 };

	public CVHandCtrls()
	{
		reset();
	}


	public String toString()
	{
		return	"ColecoVision Hand Controllers Emulation";
	}

	// I8BitIoDevice implementation
	public boolean init( char nPortBase )
	{
		m_nPortBase = nPortBase;
		reset();
		return	true;
	}

	public char read( char nPort )
	{
		if ((nPort & 0xE0)==0xE0)
		{
			nPort = (char)((nPort >> 1) & 0x01); // Joystick ports
			nPort = (char)((0 != m_bJoyMode) ? m_rgwJoyState[ nPort ] >> 8	:
						((m_rgwJoyState[ nPort ] & 0xF0) | m_rgbKeyCodes[ m_rgwJoyState[ nPort ] & 0x0F ] ));
			return	(char)((nPort | 0xB0) & 0x7F);
		}

		return	0xFF;
	}

	public void write( char nPort, char bData )
	{
		switch (nPort & 0xE0)
		{
			case 0x80:
				m_bJoyMode = 0;
				return;
			
			case 0xC0:
				m_bJoyMode = 1;
				return;
		}
	}

	public boolean reset()
	{
		// Reset state variables
		m_bJoyMode = 0;
		m_rgwJoyState[ m_CV_PRIMARY_CONTROLLER ] |= (char)(joyResetMask | m_kpdResetMask);
		m_rgwJoyState[ m_CV_SECONDARY_CONTROLLER ] |= (char)(joyResetMask | m_kpdResetMask);
		return	true;
	}

	//	IInputController
	public char SetJoyPos( char nControllerID, char nPos )
	{
		// Reset joystick to center position and then update bitmask to indicate
		//		new position.
		m_rgwJoyState[ nControllerID ] |= joyResetMask;
		if (nPos==m_CV_JOYPOS_CENTER)
			return	1;
		m_rgwJoyState[ nControllerID ] &= ~m_rgwJoyPosMasks[ nPos ];

		return	1;
	}

	public char SetButtonState( char nControllerID, char nButtonID, char nButtonState )
	{
		// Reset keypad state to nothing pressed and then set new keydown state
		//		if supplied.
		m_rgwJoyState[ nControllerID ] |= m_kpdResetMask;
		if ((nButtonState==m_CV_BUTTON_NOTPRESSED) || (nButtonID==m_CV_NUMPAD_NOTHING))
			return	1;
		if (nButtonID==m_CV_FIRE1 || nButtonID==m_CV_FIRE2)
			m_rgwJoyState[ nControllerID ] &= ~m_rgwKeypadStateMasks[ nButtonID ];
		else
		{
			m_rgwJoyState[ nControllerID ] &= 0xFFF0;
			m_rgwJoyState[ nControllerID ] |= m_rgwKeypadStateMasks[ nButtonID ];
		}

		return	1;
	}

	public char GetJoyPos( char nControllerID )
	{
		return	m_rgwJoyPosMap[ ((~m_rgwJoyState[ nControllerID ]) & joyResetMask) >> 8 ];
	}


	public char GetButtonState( char nControllerID, char nButtonID )
	{
		char	wMask = (char)(((char)~m_rgwJoyState[ nControllerID ]) & m_kpdResetMask);

		if ((nButtonID == m_CV_FIRE1) && ((wMask & m_kpdFire1Mask)==m_kpdFire1Mask))
			return	1;

		if ((nButtonID == m_CV_FIRE2) && ((wMask & m_kpdFire2Mask)==m_kpdFire2Mask))
			return	1;

		if (m_kpdKeyMap[ wMask & 0x000F ] == nButtonID)
			return	1;

		return	0;
	}
}

