import I8BitIoDevice;

class SN76489 implements I8BitIoDevice
{
	private static final char m_STATE_IDLE = 0;
	private static final char m_STATE_WAITSECONDBYTE = 1;

	private static final char m_SF_TONE1_FREQ = 0;
	private static final char m_SF_TONE2_FREQ = 1;
	private static final char m_SF_TONE3_FREQ = 2;

	private static final char m_CHANNEL_1 = 0;
	private static final char m_CHANNEL_2 = 1;
	private static final char m_CHANNEL_3 = 2;
	private static final char m_CHANNEL_NOISE = 3;

	private static final char m_NUM_SOUND_CHANNELS = 4; //(3 tone, 1 noise)

	private static final char m_NOISETYPE_PERIODIC	= 0;
	private static final char m_NOISETYPE_WHITE = 1;

	private static final char m_NUM_NOISE_TYPES = 2;

	private char m_nPortBase = 0;	
	private int m_nWriteState = 0;
	private char m_nSndFunc = 0;
	private char m_bFirstByteData = 0;


	public String toString()
	{
		return	"Texas Instruments SN76489 sound chip emulation";
	}

	// I8BitIoDevice implementation
	public boolean init( char nPortBase )
	{
		m_nPortBase = nPortBase;
		reset();
		return	true;
	}

	public char read( char nPort )
	{
		return	0xFF;
	}

	public void write( char nPort, char bData )
	{
		switch( m_nWriteState )
		{
			case	m_STATE_IDLE:
				switch ((bData>>4)&0x07)
				{
					// Tone 1 - Update frequency
					case	0:
						m_nWriteState = m_STATE_WAITSECONDBYTE;
						m_bFirstByteData = bData;		
						m_nSndFunc = m_CHANNEL_1;
						break;
					
					// Tone 1 - Update volume
					case	1:
						SetChannelVolume(m_CHANNEL_1, bData & 0x0F);
						break;

					// Tone 2 - Update frequency
					case	2:
						m_nWriteState = m_STATE_WAITSECONDBYTE;
						m_bFirstByteData = bData;
						m_nSndFunc = m_CHANNEL_2;
						break;

					// Tone 2 - Update volume
					case	3:
						SetChannelVolume(m_CHANNEL_2, bData & 0x0F);
						break;

					// Tone 3 - Update frequency
					case	4:
						m_nWriteState = m_STATE_WAITSECONDBYTE;
						m_bFirstByteData = bData;
						m_nSndFunc = m_CHANNEL_3;
						break;

					// Tone 3 - Update volume
					case	5:
						SetChannelVolume(m_CHANNEL_3, bData & 0x0F);
						break;				

					// noise type
					case	6:
						bData &= 0x03;
						SetNoiseType((char)((bData>>2)&0x0F), (char)(bData & 0x02));
						break;

					// noise volume
					case	7:
						SetChannelVolume(m_CHANNEL_NOISE, bData & 0x0F);
						break;
				}
				break;

			case	m_STATE_WAITSECONDBYTE:
				{
					int nFreq = ((bData & 0x3F)<< 4) | (m_bFirstByteData & 0x0F);
				
					SetChannelFrequency(m_nSndFunc, nFreq);

					// Reset to initial state...
					m_nWriteState = m_STATE_IDLE;
					m_bFirstByteData = 0x00;
				}
				break;
		}
	}

	public boolean reset()
	{
		return	true;
	}


	private void SetChannelVolume(int nChannel, int nVolume)
	{
	}

	private void SetChannelFrequency(int nChannel, int nVolume)
	{
	}

	private void SetNoiseType(char bType, char bShiftRate)
	{
	}
}
