//******************************************************************************
// VColecoJava.java:	Applet
//
//******************************************************************************
import java.io.*;
import java.applet.*;
import java.awt.*;
import java.net.*;
import VColecoFrame;
import ColecoVision;
import IVideoRefreshNotify;
import EmulPropDlg;
import AboutDlg;


//==============================================================================
// Main Class for applet VColecoJava
//
//==============================================================================
public class VColecoJava extends Applet implements IVideoRefreshNotify, Runnable
{
	private ColecoVision	m_ColecoVision;


	private Thread m_VColecoJava;

	// STANDALONE APPLICATION SUPPORT:
	//		m_fStandAlone will be set to true if applet is run standalone
	//--------------------------------------------------------------------------
	private boolean m_fStandAlone = false;
	private boolean m_fInitialized = false;
	private boolean m_fPaused = false;

	private String m_szMsg;

	// PARAMETER SUPPORT:
	//		Parameters allow an HTML author to pass information to the applet;
	// the HTML author specifies them using the <PARAM> tag within the <APPLET>
	// tag.  The following variables are used to store the values of the
	// parameters.
    //--------------------------------------------------------------------------

    // Members for applet parameters
    // <type>       <MemberVar>    = <Default Value>
    //--------------------------------------------------------------------------
	private String m_OSImage = "";
	private String m_ROMImage = "";
	private int m_nFrameSkipRate = 4;
	private boolean m_fLockCPUSpeed = false;
	private int m_SyncMethod = ColecoVision.m_TSM_TIMEDCPU;
	private int m_NMIFrequency = 60;

   // Parameter names.  To change a name of a parameter, you need only make
	// a single change.  Simply modify the value of the parameter string below.
   //--------------------------------------------------------------------------
	private final String PARAM_OSImage = "OSImage";
	private final String PARAM_ROMImage = "ROMImage";
	private final String PARAM_SkipRate = "SkipRate";
	private final String PARAM_LockSpeed = "LockSpeed";
	private final String PARAM_SyncMethod = "SyncMethod";
	private final String PARAM_NMIFrequency = "NMIFrequency";

	// STANDALONE APPLICATION SUPPORT
	// 	The GetParameter() method is a replacement for the getParameter() method
	// defined by Applet. This method returns the value of the specified parameter;
	// unlike the original getParameter() method, this method works when the applet
	// is run as a standalone application, as well as when run within an HTML page.
	// This method is called by GetParameters().
	//---------------------------------------------------------------------------
	String GetParameter(String strName, String args[])
	{
		if (args == null)
		{
			// Running within an HTML page, so call original getParameter().
			//-------------------------------------------------------------------
			return getParameter(strName);
		}

		// Running as standalone application, so parameter values are obtained from
		// the command line. The user specifies them as follows:
		//
		//	JView VColecoJava param1=<val> param2=<"val with spaces"> ...
		//-----------------------------------------------------------------------
		int    i;
		String strArg	= strName + "=";
		String strValue = null;
		int    nLength  = strArg.length();

		try
		{
			for (i = 0; i < args.length; i++)
			{
				String strParam = "" + args[i].substring(0, nLength);

				if (strArg.equalsIgnoreCase(strParam))
				{
					// Found matching parameter on command line, so extract its value.
					// If in double quotes, remove the quotes.
					//---------------------------------------------------------------
					strValue = "" + args[i].substring(nLength);
					if (strValue.startsWith("\""))
					{
						strValue = "" + strValue.substring(1);
						if (strValue.endsWith("\""))
							strValue = "" + strValue.substring(0, strValue.length() - 1);
					}
					break;
				}
			}
		}
		catch (Exception e)
		{
			// TODO: Place exception-handling code here in case an
		}

		return strValue;
	}

	// STANDALONE APPLICATION SUPPORT
	// 	The GetParameters() method retrieves the values of each of the applet's
	// parameters and stores them in variables. This method works both when the
	// applet is run as a standalone application and when it's run within an HTML
	// page.  When the applet is run as a standalone application, this method is
	// called by the main() method, which passes it the command-line arguments.
	// When the applet is run within an HTML page, this method is called by the
	// init() method with args == null.
	//---------------------------------------------------------------------------
	void GetParameters(String args[])
	{
		// Query values of all Parameters
		//--------------------------------------------------------------
		String param;

		// OSImage: URL of OS ROM to load (COLECO.ROM)
		//--------------------------------------------------------------
		param = GetParameter(PARAM_OSImage, args);
		if (param != null)
			m_OSImage = param;

		// ROMImage: URL of cartridge ROM to play
		//--------------------------------------------------------------
		param = GetParameter(PARAM_ROMImage, args);
		if (param != null)
			m_ROMImage = param;

		// FrameRate: Frames per second.
		//--------------------------------------------------------------
		param = GetParameter(PARAM_SkipRate, args);
		if (param != null)
			m_nFrameSkipRate = Integer.parseInt(param);

		// LockSpeed: Lock CPU emulation @ 3.58Mhz
		//--------------------------------------------------------------
		param = GetParameter(PARAM_LockSpeed, args);
		if (param != null)
			m_fLockCPUSpeed = Boolean.valueOf(param).booleanValue();

		// SyncMethod: NMI Timing Method
		//--------------------------------------------------------------
		param = GetParameter(PARAM_SyncMethod, args);
		if (param != null)
		{
			if (param.equalsIgnoreCase("dynamic"))
				m_SyncMethod = ColecoVision.m_TSM_TIMEDCPU;
			else if (param.equalsIgnoreCase("none"))
				m_SyncMethod = ColecoVision.m_TSM_EMULATEDCPU;
			else
				m_SyncMethod = Integer.parseInt(param);
		}

		param = GetParameter(PARAM_NMIFrequency, args);
		if (param != null)
			m_NMIFrequency = Integer.parseInt(param);
	}

	void	SetMsgText(String szMsg)
	{
		m_szMsg = szMsg;
		repaint();
	}

	// STANDALONE APPLICATION SUPPORT
	// 	The main() method acts as the applet's entry point when it is run
	// as a standalone application. It is ignored if the applet is run from
	// within an HTML page.
	//--------------------------------------------------------------------------
	public static void main(String args[])
	{
		// Create Toplevel Window to contain applet VColecoJava
		//----------------------------------------------------------------------
		VColecoFrame frame = new VColecoFrame("Virtual ColecoVision (JavaVM)");

		// Must show Frame before we size it so insets() will return valid values
		//----------------------------------------------------------------------
		frame.show();
      frame.hide();
		frame.resize(frame.insets().left + frame.insets().right  + 272,
					 frame.insets().top  + frame.insets().bottom + 208);
		frame.setResizable(false);

		// The following code starts the applet running within the frame window.
		// It also calls GetParameters() to retrieve parameter values from the
		// command line, and sets m_fStandAlone to true to prevent init() from
		// trying to get them from the HTML page.
		//----------------------------------------------------------------------
		VColecoJava applet_VColecoJava = new VColecoJava();

		frame.add("Center", applet_VColecoJava);
		applet_VColecoJava.m_fStandAlone = true;
		applet_VColecoJava.GetParameters(args);
		applet_VColecoJava.init();
		applet_VColecoJava.start();
      frame.show();
	}

	// VColecoJava Class Constructor
	//--------------------------------------------------------------------------
	public VColecoJava()
	{
	}

	// APPLET INFO SUPPORT:
	//		The getAppletInfo() method returns a string describing the applet's
	// author, copyright date, or miscellaneous information.
    //--------------------------------------------------------------------------
	public String getAppletInfo()
	{
		return "Name: Virtual ColecoVision for Java\r\n" +
		       "Author: Neal Danner\r\n" +
		       "Created with Microsoft Visual J++ Version 1.1";
	}

	// PARAMETER SUPPORT
	//		The getParameterInfo() method returns an array of strings describing
	// the parameters understood by this applet.
	//
    // VColecoJava Parameter Information:
    //  { "Name", "Type", "Description" },
    //--------------------------------------------------------------------------
	public String[][] getParameterInfo()
	{
		String[][] info =
		{
			{ PARAM_OSImage, "String", "URL of OS ROM to load (COLECO.ROM)" },
			{ PARAM_ROMImage, "String", "URL of cartridge ROM to play" },
			{ PARAM_SkipRate, "int", "Video output frame skip rate." },
			{ PARAM_LockSpeed, "boolean", "Lock CPU emulation @ 3.58Mhz" },
			{ PARAM_SyncMethod, "int", "NMI Timing Method (native,emulated)" },
			{ PARAM_NMIFrequency, "int", "NMI Frequency setting (Hz)" }
		};
		return info;		
	}

	// The init() method is called by the AWT when an applet is first loaded or
	// reloaded.  Override this method to perform whatever initialization your
	// applet needs, such as initializing data structures, loading images or
	// fonts, creating frame windows, setting the layout manager, or adding UI
	// components.
    //--------------------------------------------------------------------------
	public void init()
	{
		if (!m_fStandAlone)
			GetParameters(null);
        // If you use a ResourceWizard-generated "control creator" class to
        // arrange controls in your applet, you may want to call its
        // CreateControls() method from within this method. Remove the following
        // call to resize() before adding the call to CreateControls();
        // CreateControls() does its own resizing.
        //----------------------------------------------------------------------
    	resize(272, 208);

		setBackground(Color.black);
		setForeground(Color.white);
	}

	// Place additional applet clean up code here.  destroy() is called when
	// when you applet is terminating and being unloaded.
	//-------------------------------------------------------------------------
	public void destroy()
	{
		m_ColecoVision.shutdown();
	}

	// VColecoJava Paint Handler
	//--------------------------------------------------------------------------
	public void paint(Graphics g)
	{
		if (m_fInitialized)
			m_ColecoVision.paint(g, true);
		else
			g.drawString(m_szMsg, 5, 10);
	}

	public void update(Graphics g)
	{
		if (m_fInitialized)
			paint(g);
		else
			super.update(g);
	}
	
	public void RefreshNotification()
	{
		Graphics g = getGraphics();
		if (null!=g)
			m_ColecoVision.paint(g, false);
	}

	public void run()
	{
		SetMsgText("Initializing core emulation...");
		m_ColecoVision = new ColecoVision();
		SetMsgText("Preparing core emulation...");
		if (!(m_fInitialized = m_ColecoVision.init(this, this)))
			SetMsgText("Core emulation preparation failed.");
		else
			SetMsgText("Core emulation started.");

		// Setup the screen refresh stuff...
		m_ColecoVision.setVideoSyncMethod(m_SyncMethod);
		m_ColecoVision.setVideoRefreshFreq(m_NMIFrequency);
		m_ColecoVision.setVideoRefreshSkipRate(m_nFrameSkipRate);
		m_ColecoVision.lockCPUSpeed(m_fLockCPUSpeed);

		// Load operating system ROM if specified
		if (m_OSImage.length()>0)
		{
			try
			{
				URL urlOS = new URL(m_OSImage);
				m_ColecoVision.loadOS(urlOS);
			}
			catch (Exception e)
			{
				SetMsgText(e.toString());
			}
		}

		// Load cartridge ROM if specified
		if (m_ROMImage.length()>0)
		{
			try
			{
				URL urlCart = new URL(m_ROMImage);
				m_ColecoVision.loadCartridge(urlCart);
			}
			catch (Exception e)
			{
				SetMsgText(e.toString());
			}
		}
		
		// Start the emulation thread.
		m_ColecoVision.start();
	}

	//		The start() method is called when the page containing the applet
	// first appears on the screen. The AppletWizard's initial implementation
	// of this method starts execution of the applet's thread.
	//--------------------------------------------------------------------------
	public void start()
	{
		if (null==m_VColecoJava)
		{
			m_VColecoJava = new Thread(this);
			m_VColecoJava.start();
		}
	}
	
	//		The stop() method is called when the page containing the applet is
	// no longer on the screen. The AppletWizard's initial implementation of
	// this method stops execution of the applet's thread.
	//--------------------------------------------------------------------------
	public void stop()
	{
		m_ColecoVision.stop();
		if (null!=m_VColecoJava)
			m_VColecoJava.stop();
	}

	public boolean keyDown(Event evt, int nKey)
	{
		if (null!=m_ColecoVision)
			return	m_ColecoVision.keyDown(evt, nKey);
		return	false;
	}

	public boolean keyUp(Event evt, int nKey)
	{
		if (null!=m_ColecoVision)
			return	m_ColecoVision.keyUp(evt, nKey);
		return	false;
	}

	public boolean OnFILE_OPEN(Frame parent)
	{
		FileDialog	fd = new FileDialog(parent, "Open Cartridge ROM", FileDialog.LOAD);

		fd.setFile("*.rom");
		fd.show();
		
		if (null!=fd.getFile())
		{
			String	szFileToLoad = fd.getDirectory() + fd.getFile();

			try
			{
				DataInputStream dis = new DataInputStream(new FileInputStream(szFileToLoad));
				m_ColecoVision.loadCartridge(dis);
				dis.close();
			}
			catch(Exception e)
			{
				System.out.println(e.toString());
			}
		}

		return true;
	}

	public boolean OnEMULATION_RESET()
	{
		m_ColecoVision.reset();
		return	true;
	}

	public boolean OnEMULATION_PAUSE()
	{
		m_ColecoVision.pause(m_fPaused = !m_fPaused);
		return	true;
	}

	public boolean OnEMULATION_PROPERTIES(Frame parent)
	{
		EmulPropDlg	pd = new EmulPropDlg(parent, true, m_ColecoVision);
		pd.show();
		return true;
	}

	public boolean OnHELP_ABOUT(Frame parent)
	{
		AboutDlg ad = new AboutDlg(parent, true);
		ad.show();
		return true;
	}
}