import I8BitIoDevice;

class Z80SystemBus
{
	private static final char m_MAX_IO_DEVICES = 16;

	private char m_Memory[] = null;
	private I8BitIoDevice m_Devices[] = null;
	private char m_DeviceMapRead[] = null;
	private char m_DeviceMapWrite[] = null;
	
	public Z80SystemBus()
	{
	}

	public boolean init()
	{
		m_Memory = new char[0x10000];	
		m_Devices = new I8BitIoDevice[m_MAX_IO_DEVICES];

		m_DeviceMapRead = new char[256];
		m_DeviceMapWrite = new char[256];

		int i;

		for (i = 0; i < m_MAX_IO_DEVICES; i++)
			m_Devices[ i ] = null;

		for (i = 0; i < 256; i++)
		{
			m_DeviceMapRead[i] = 0xFF;
			m_DeviceMapWrite[i] = 0xFF;
		}
		return	true;
	}

	public boolean shutdown()
	{
		return	true;
	}

	public boolean reset()
	{
		int i;

		// Empty system memory
		for (i = 0; i < 65536; i++)
			m_Memory[i] = 0xFF;

		// Reset devices...
		for (i = 0; i < m_MAX_IO_DEVICES; i++)
			if (null != m_Devices[ i ])
				m_Devices[ i ].reset();

		return	true;
	}

	//	Memory I/O
	public char MemReadByte(char wOffset)
	{
		return m_Memory[ wOffset ];
	}

	public void MemWriteByte(char wOffset, char bData)
	{
		m_Memory[ wOffset ] = (char)(bData & 0xFF);
	}
	public char MemReadWord(char wOffset)
	{
		char wData =(char)(m_Memory[ wOffset++ ] & 0xFF);
		wData |= (char)(m_Memory[ wOffset ] << 8);
		return	wData;
	}

	public void MemWriteWord(char wOffset, char wData)
	{
		m_Memory[ wOffset++ ] = (char)(wData & 0xFF);
		m_Memory[ wOffset ] = (char)((wData >> 8) & 0xFF);
	}

	// Port I/O
	public char ReadDevice(char bPort)
	{
		char bPortOwner = m_DeviceMapRead[ bPort ];
		if (bPortOwner != (char)0xFF)
			return	m_Devices[ bPortOwner ].read(bPort);

		return	(char)0xFF;
	}

	public void WriteDevice(char bPort, char bData)
	{
		char bPortOwner = m_DeviceMapWrite[ bPort ];
		if (bPortOwner != 0xFF)
			m_Devices[ bPortOwner ].write(bPort, bData);
	}

	public boolean MapIoDevice(char nPortBase, char nPortSpan,
		I8BitIoDevice Device, boolean fInport, boolean fOutport)
	{
		//	See if device is already registered
		boolean fDeviceRegistered = false;
		int i, nDeviceSlot = m_MAX_IO_DEVICES;
		char	DeviceMapArea[] = null;

		// Gotta map something!!!
		if ((!fInport && !fOutport) || 0==nPortSpan)
			return	false;

		if (	nPortBase < 0		||
				(nPortBase+nPortSpan) > 256	||
				nPortSpan < 1
			)
			return	false;
		
		for (i = 0; i < m_MAX_IO_DEVICES; i++)
		{
			if (m_Devices[ i ] == Device)
			{
				nDeviceSlot = i;
				fDeviceRegistered = true;
				break;
			}

			if (	m_Devices[ i ] == null	&&
					i < nDeviceSlot
				)
			{
				nDeviceSlot = i;
			}
		}

		//	Exit now if no room to map device
		if (nDeviceSlot >= m_MAX_IO_DEVICES)
			return	false;

		//	Check to see if port areas are already mapped...
		if (fInport)
		{
			for (i = nPortBase; i < (nPortBase+nPortSpan); i++)
			{
				if (m_DeviceMapRead[ i ] != 0xFF)
					return	false;
			}
		}

		if (fOutport)
		{
			for (i = nPortBase; i < (nPortBase+nPortSpan); i++)
			{
				if (m_DeviceMapWrite[ i ] != 0xFF)
					return	false;
			}
		}

		// Map device in if it isn't already registered
		if (!fDeviceRegistered)
		{
			if (!Device.init(nPortBase))
				return	false;
			m_Devices[ nDeviceSlot ] = Device;
		}

		// Map device in as requested
		for (;;)
		{
			if (fInport)
			{
				DeviceMapArea = m_DeviceMapRead;
			}
			else if (fOutport)
			{
				DeviceMapArea = m_DeviceMapWrite;
			}
			else
				break;
				
				//	Register the device!!!
			for (i = nPortBase; i < (nPortBase+nPortSpan); i++)
				DeviceMapArea[ i ] = (char)(nDeviceSlot);

			if (!fInport && fOutport)
				fOutport = false;

			fInport = false;	// done mapping inport if any
		}

		return	true;
	}
}
