import java.io.*;
import java.awt.*;
import java.net.*;
import Z80SystemBus;
import Z80CPU;
import TMS9928A;
import SN76489;
import CVHandCtrls;
//import DemoCartridge;
import IVideoRefreshNotify;

//
//
// ColecoVision
//
//
class ColecoVision extends Z80SystemBus implements Runnable
{
	private Component m_parent;
	private IVideoRefreshNotify m_refreshSink = null;

	private static final int m_CVISION_ORIGINAL_HZ = 3580000;
	private static final int m_CVISION_ORIGINAL_FPS = 60;
	private static final int m_CYCLES_TO_RUN = m_CVISION_ORIGINAL_HZ / 150/*m_CVISION_ORIGINAL_FPS*/;

	public static final int m_TSM_EMULATEDCPU = 0;
	public static final int m_TSM_TIMEDCPU = 1;

	private static final int m_MIN_CYCLES_DYNAMIC = 3580000; // 1.5Mhz
	private static final int m_NMI_RESYNC_TIME	= 1000;
	private static final int m_NMI_RESYNC_ADJUSTMENT = m_NMI_RESYNC_TIME / 1000;

	private Thread m_coreEmulation = null;
	private String m_szMsg = "";

	//	holder for ROM image
	private byte m_rgbCartridgeROM[] = null;

	// Core emulation pieces
	private Z80SystemBus		m_z80bus = null;
	private Z80CPU				m_z80cpu = null;
	private TMS9928A			m_tms9928a = null;
	private SN76489			m_sn76489 = null;
	private CVHandCtrls		m_handCtls = null;

	private int m_nTimingSyncMethod = m_TSM_TIMEDCPU;	// NMI freq timing method
	private int m_nRefreshFreq = m_CVISION_ORIGINAL_FPS;
	private int m_nRefreshCounter = m_CVISION_ORIGINAL_HZ / m_nRefreshFreq;

	//	TSM_EMULATEDCPU && TSM_TIMEDCPU
	private int m_nDynamicRefreshCounter = m_nRefreshCounter;
	private int m_nNextRefreshTime = 0;
	private int m_nLastRefreshCycle = 0;
	private int m_nCyclesSinceRefresh = 0;

	private boolean m_fInitialized = false;
	private long m_lMsgHideTime = 0;

	private boolean m_fLockCPUSpeed = false;
	private long m_lSliceEndConst = 0;
	private int m_nLockedCPUCyclesToBurn = m_CVISION_ORIGINAL_HZ / 1000; //	(CV's per ms rate)

	private int m_nFrameSkipRate = 4;
	private int m_nFrameSkipCount = m_nFrameSkipRate;

	private boolean m_fPaused = false;

	private byte m_rgbOSROM[] = null;
	private static final byte m_rgbCOLECO_ROM[] =
	{
		(byte)0x31, (byte)0xB9, (byte)0x73, (byte)0xC3, (byte)0x6E, (byte)0x00, (byte)0xFF, (byte)0xFF, (byte)0xC3, (byte)0x0C, (byte)0x80, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF,
		(byte)0xC3, (byte)0x0F, (byte)0x80, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xC3, (byte)0x12, (byte)0x80, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF,
		(byte)0xC3, (byte)0x15, (byte)0x80, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xC3, (byte)0x18, (byte)0x80, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF,
		(byte)0xC3, (byte)0x1B, (byte)0x80, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xC3, (byte)0x1E, (byte)0x80, (byte)0x2A, (byte)0xC8, (byte)0x73, (byte)0xCB, (byte)0x7C,
		(byte)0x28, (byte)0x06, (byte)0xCB, (byte)0x44, (byte)0x28, (byte)0x06, (byte)0x18, (byte)0x07, (byte)0xCB, (byte)0x44, (byte)0x28, (byte)0x03, (byte)0x37, (byte)0x18, (byte)0x01, (byte)0xB7,
		(byte)0xCB, (byte)0x15, (byte)0xCB, (byte)0x14, (byte)0x22, (byte)0xC8, (byte)0x73, (byte)0x7D, (byte)0xC9, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF,
		(byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xC3, (byte)0x21, (byte)0x80, (byte)0x3C, (byte)0xAB, (byte)0x16, (byte)0x23, (byte)0x16, (byte)0x2A, (byte)0x00,
		(byte)0x80, (byte)0x7D, (byte)0xFE, (byte)0x55, (byte)0xC2, (byte)0x81, (byte)0x00, (byte)0x7C, (byte)0xFE, (byte)0xAA, (byte)0xC2, (byte)0x81, (byte)0x00, (byte)0x2A, (byte)0x0A, (byte)0x80,
		(byte)0xE9, (byte)0xCD, (byte)0xD6, (byte)0x1F, (byte)0x21, (byte)0x33, (byte)0x00, (byte)0x22, (byte)0xC8, (byte)0x73, (byte)0xCD, (byte)0x05, (byte)0x11, (byte)0x3E, (byte)0x00, (byte)0x32,
		(byte)0xC6, (byte)0x73, (byte)0x32, (byte)0xC7, (byte)0x73, (byte)0xC3, (byte)0x19, (byte)0x13, (byte)0xE1, (byte)0xE3, (byte)0xE5, (byte)0x0A, (byte)0x6F, (byte)0x03, (byte)0x0A, (byte)0x03,
		(byte)0x67, (byte)0xE3, (byte)0xD5, (byte)0x5E, (byte)0x23, (byte)0x56, (byte)0x23, (byte)0xE5, (byte)0x7B, (byte)0xB2, (byte)0xC2, (byte)0xB7, (byte)0x00, (byte)0xE1, (byte)0x5E, (byte)0x23,
		(byte)0x56, (byte)0x23, (byte)0xE5, (byte)0xEB, (byte)0x5E, (byte)0x23, (byte)0x56, (byte)0x03, (byte)0x0A, (byte)0x07, (byte)0xD2, (byte)0xDA, (byte)0x00, (byte)0x03, (byte)0xE1, (byte)0xE3,
		(byte)0x73, (byte)0x23, (byte)0x72, (byte)0x23, (byte)0xD1, (byte)0xE3, (byte)0x2B, (byte)0xAF, (byte)0xBC, (byte)0xC2, (byte)0xD0, (byte)0x00, (byte)0xBD, (byte)0xCA, (byte)0xD6, (byte)0x00,
		(byte)0xE3, (byte)0xE5, (byte)0xEB, (byte)0xC3, (byte)0xA3, (byte)0x00, (byte)0xE1, (byte)0xEB, (byte)0xE3, (byte)0xE9, (byte)0xE1, (byte)0xE3, (byte)0xE5, (byte)0x0F, (byte)0x67, (byte)0x0B,
		(byte)0x0A, (byte)0x6F, (byte)0xE3, (byte)0x03, (byte)0x03, (byte)0x1A, (byte)0x77, (byte)0x23, (byte)0x13, (byte)0xE3, (byte)0x2B, (byte)0xAF, (byte)0xBD, (byte)0xC2, (byte)0xF4, (byte)0x00,
		(byte)0xBC, (byte)0xCA, (byte)0xF8, (byte)0x00, (byte)0xE3, (byte)0xC3, (byte)0xE5, (byte)0x00, (byte)0xE1, (byte)0xC3, (byte)0xC4, (byte)0x00, (byte)0xDD, (byte)0x7E, (byte)0x07, (byte)0xFE,
		(byte)0x00, (byte)0x20, (byte)0x09, (byte)0xDD, (byte)0x7E, (byte)0x05, (byte)0x3D, (byte)0xC8, (byte)0xDD, (byte)0x77, (byte)0x05, (byte)0xC9, (byte)0xDD, (byte)0xE5, (byte)0xE1, (byte)0x1E,
		(byte)0x06, (byte)0x16, (byte)0x00, (byte)0x19, (byte)0xCD, (byte)0x90, (byte)0x01, (byte)0x20, (byte)0x15, (byte)0xCD, (byte)0xA6, (byte)0x01, (byte)0x2B, (byte)0x7E, (byte)0x3D, (byte)0xC8,
		(byte)0x77, (byte)0x2B, (byte)0x2B, (byte)0xDD, (byte)0x7E, (byte)0x07, (byte)0xCD, (byte)0xB1, (byte)0x01, (byte)0x23, (byte)0xCB, (byte)0x96, (byte)0xF6, (byte)0xFF, (byte)0xC9, (byte)0xDD,
		(byte)0x7E, (byte)0x08, (byte)0xFE, (byte)0x00, (byte)0xC8, (byte)0xDD, (byte)0xE5, (byte)0xE1, (byte)0x16, (byte)0x00, (byte)0x1E, (byte)0x09, (byte)0x19, (byte)0xCD, (byte)0x90, (byte)0x01,
		(byte)0x20, (byte)0x21, (byte)0xCD, (byte)0xA6, (byte)0x01, (byte)0x2B, (byte)0xCD, (byte)0x90, (byte)0x01, (byte)0x28, (byte)0x16, (byte)0x7E, (byte)0xE6, (byte)0xF0, (byte)0x5F, (byte)0x2B,
		(byte)0x2B, (byte)0x2B, (byte)0x2B, (byte)0x7E, (byte)0xE6, (byte)0xF0, (byte)0x83, (byte)0x5F, (byte)0x7E, (byte)0xE6, (byte)0x0F, (byte)0xB3, (byte)0x77, (byte)0xF6, (byte)0xFF, (byte)0x18,
		(byte)0x02, (byte)0x36, (byte)0x00, (byte)0xC9, (byte)0xDD, (byte)0x7E, (byte)0x04, (byte)0xCB, (byte)0x61, (byte)0x28, (byte)0x04, (byte)0x0F, (byte)0x0F, (byte)0x0F, (byte)0x0F, (byte)0xE6,
		(byte)0x0F, (byte)0xB1, (byte)0xD3, (byte)0xFF, (byte)0xC9, (byte)0xDD, (byte)0x7E, (byte)0x03, (byte)0xE6, (byte)0x0F, (byte)0xB2, (byte)0xD3, (byte)0xFF, (byte)0xDD, (byte)0x7E, (byte)0x03,
		(byte)0xE6, (byte)0xF0, (byte)0x57, (byte)0xDD, (byte)0x7E, (byte)0x04, (byte)0xE6, (byte)0x0F, (byte)0xB2, (byte)0x0F, (byte)0x0F, (byte)0x0F, (byte)0x0F, (byte)0xD3, (byte)0xFF, (byte)0xC9,
		(byte)0x3E, (byte)0x00, (byte)0xED, (byte)0x67, (byte)0xD6, (byte)0x01, (byte)0xF5, (byte)0xED, (byte)0x6F, (byte)0xF1, (byte)0xC9, (byte)0x3E, (byte)0x00, (byte)0xED, (byte)0x6F, (byte)0xD6,
		(byte)0x01, (byte)0xF5, (byte)0xED, (byte)0x67, (byte)0xF1, (byte)0xC9, (byte)0x7E, (byte)0xE6, (byte)0xF0, (byte)0x47, (byte)0x0F, (byte)0x0F, (byte)0x0F, (byte)0x0F, (byte)0xB0, (byte)0x77,
		(byte)0xC9, (byte)0x06, (byte)0x00, (byte)0xCB, (byte)0x7F, (byte)0x28, (byte)0x02, (byte)0x06, (byte)0xFF, (byte)0x86, (byte)0x77, (byte)0x23, (byte)0x7E, (byte)0x88, (byte)0x77, (byte)0x2B,
		(byte)0xC9, (byte)0x2A, (byte)0x20, (byte)0x70, (byte)0x2B, (byte)0x2B, (byte)0x48, (byte)0x06, (byte)0x00, (byte)0xCB, (byte)0x01, (byte)0xCB, (byte)0x01, (byte)0x09, (byte)0x5E, (byte)0x23,
		(byte)0x56, (byte)0xD5, (byte)0xDD, (byte)0xE1, (byte)0xC9, (byte)0xDD, (byte)0x75, (byte)0x01, (byte)0xDD, (byte)0x74, (byte)0x02, (byte)0x1A, (byte)0xE6, (byte)0x3F, (byte)0x47, (byte)0xDD,
		(byte)0x7E, (byte)0x00, (byte)0xE6, (byte)0xC0, (byte)0xB0, (byte)0xDD, (byte)0x77, (byte)0x00, (byte)0xC9, (byte)0xDD, (byte)0x7E, (byte)0x00, (byte)0xFE, (byte)0xFF, (byte)0xC8, (byte)0xE6,
		(byte)0x3F, (byte)0xFE, (byte)0x3E, (byte)0xC0, (byte)0xDD, (byte)0xE5, (byte)0xE1, (byte)0x23, (byte)0x5E, (byte)0x23, (byte)0x56, (byte)0xEB, (byte)0xC9, (byte)0x02, (byte)0x00, (byte)0x01,
		(byte)0x00, (byte)0x02, (byte)0x00, (byte)0x01, (byte)0xFD, (byte)0x01, (byte)0x11, (byte)0xBA, (byte)0x73, (byte)0xCD, (byte)0x98, (byte)0x00, (byte)0x3A, (byte)0xBA, (byte)0x73, (byte)0x47,
		(byte)0x2A, (byte)0xBB, (byte)0x73, (byte)0x22, (byte)0x20, (byte)0x70, (byte)0x23, (byte)0x23, (byte)0x5E, (byte)0x23, (byte)0x56, (byte)0xEB, (byte)0x1E, (byte)0x0A, (byte)0x16, (byte)0x00,
		(byte)0x36, (byte)0xFF, (byte)0x19, (byte)0x10, (byte)0xFB, (byte)0x36, (byte)0x00, (byte)0x21, (byte)0x4C, (byte)0x02, (byte)0x22, (byte)0x22, (byte)0x70, (byte)0x22, (byte)0x24, (byte)0x70,
		(byte)0x22, (byte)0x26, (byte)0x70, (byte)0x22, (byte)0x28, (byte)0x70, (byte)0x3E, (byte)0xFF, (byte)0x32, (byte)0x2A, (byte)0x70, (byte)0x3E, (byte)0x9F, (byte)0xD3, (byte)0xFF, (byte)0x3E,
		(byte)0xBF, (byte)0xD3, (byte)0xFF, (byte)0x3E, (byte)0xDF, (byte)0xD3, (byte)0xFF, (byte)0x3E, (byte)0xFF, (byte)0xD3, (byte)0xFF, (byte)0xC9, (byte)0xFF, (byte)0x01, (byte)0x00, (byte)0x01,
		(byte)0x00, (byte)0x01, (byte)0x4D, (byte)0x02, (byte)0x11, (byte)0xBA, (byte)0x73, (byte)0xCD, (byte)0x98, (byte)0x00, (byte)0x3A, (byte)0xBA, (byte)0x73, (byte)0x47, (byte)0xC5, (byte)0xCD,
		(byte)0xC1, (byte)0x01, (byte)0xDD, (byte)0x7E, (byte)0x00, (byte)0xE6, (byte)0x3F, (byte)0xC1, (byte)0xB8, (byte)0xC8, (byte)0xDD, (byte)0x70, (byte)0x00, (byte)0x2B, (byte)0x2B, (byte)0x56,
		(byte)0x2B, (byte)0x5E, (byte)0xDD, (byte)0x73, (byte)0x01, (byte)0xDD, (byte)0x72, (byte)0x02, (byte)0xCD, (byte)0x5F, (byte)0x03, (byte)0xCD, (byte)0x95, (byte)0x02, (byte)0xC9, (byte)0x06,
		(byte)0x01, (byte)0xCD, (byte)0xC1, (byte)0x01, (byte)0x3E, (byte)0x00, (byte)0xDD, (byte)0xBE, (byte)0x00, (byte)0xC8, (byte)0xCD, (byte)0xD6, (byte)0x02, (byte)0x1E, (byte)0x0A, (byte)0x16,
		(byte)0x00, (byte)0xDD, (byte)0x19, (byte)0x18, (byte)0xEF, (byte)0xDD, (byte)0xE5, (byte)0x21, (byte)0x4C, (byte)0x02, (byte)0x22, (byte)0x22, (byte)0x70, (byte)0x22, (byte)0x24, (byte)0x70,
		(byte)0x22, (byte)0x26, (byte)0x70, (byte)0x22, (byte)0x28, (byte)0x70, (byte)0x06, (byte)0x01, (byte)0xCD, (byte)0xC1, (byte)0x01, (byte)0xDD, (byte)0x7E, (byte)0x00, (byte)0xFE, (byte)0x00,
		(byte)0x28, (byte)0x21, (byte)0xFE, (byte)0xFF, (byte)0x28, (byte)0x15, (byte)0xDD, (byte)0x7E, (byte)0x00, (byte)0xE6, (byte)0xC0, (byte)0x07, (byte)0x07, (byte)0x07, (byte)0x5F, (byte)0x16,
		(byte)0x00, (byte)0x21, (byte)0x22, (byte)0x70, (byte)0x19, (byte)0xDD, (byte)0xE5, (byte)0xD1, (byte)0x73, (byte)0x23, (byte)0x72, (byte)0x1E, (byte)0x0A, (byte)0x16, (byte)0x00, (byte)0xDD,
		(byte)0x19, (byte)0x18, (byte)0xD8, (byte)0xDD, (byte)0xE1, (byte)0xC9, (byte)0xCD, (byte)0xE9, (byte)0x01, (byte)0xFE, (byte)0xFF, (byte)0xC8, (byte)0xFE, (byte)0x3E, (byte)0x20, (byte)0x06,
		(byte)0x1E, (byte)0x07, (byte)0x16, (byte)0x00, (byte)0x19, (byte)0xE9, (byte)0xCD, (byte)0x2F, (byte)0x01, (byte)0xCD, (byte)0xFC, (byte)0x00, (byte)0x20, (byte)0x11, (byte)0xDD, (byte)0x7E,
		(byte)0x00, (byte)0xF5, (byte)0xCD, (byte)0x5F, (byte)0x03, (byte)0xC1, (byte)0xDD, (byte)0x7E, (byte)0x00, (byte)0xB8, (byte)0x28, (byte)0x03, (byte)0xCD, (byte)0x95, (byte)0x02, (byte)0xC9,
		(byte)0x3E, (byte)0x9F, (byte)0x0E, (byte)0x90, (byte)0x16, (byte)0x80, (byte)0xDD, (byte)0x2A, (byte)0x24, (byte)0x70, (byte)0xCD, (byte)0x4E, (byte)0x03, (byte)0x3E, (byte)0xBF, (byte)0x0E,
		(byte)0xB0, (byte)0x16, (byte)0xA0, (byte)0xDD, (byte)0x2A, (byte)0x26, (byte)0x70, (byte)0xCD, (byte)0x4E, (byte)0x03, (byte)0x3E, (byte)0xDF, (byte)0x0E, (byte)0xD0, (byte)0x16, (byte)0xC0,
		(byte)0xDD, (byte)0x2A, (byte)0x28, (byte)0x70, (byte)0xCD, (byte)0x4E, (byte)0x03, (byte)0x3E, (byte)0xFF, (byte)0x0E, (byte)0xF0, (byte)0xDD, (byte)0x2A, (byte)0x22, (byte)0x70, (byte)0xDD,
		(byte)0x5E, (byte)0x00, (byte)0x1C, (byte)0x20, (byte)0x04, (byte)0xD3, (byte)0xFF, (byte)0x18, (byte)0x14, (byte)0xCD, (byte)0x64, (byte)0x01, (byte)0xDD, (byte)0x7E, (byte)0x04, (byte)0xE6,
		(byte)0x0F, (byte)0x21, (byte)0x2A, (byte)0x70, (byte)0xBE, (byte)0x28, (byte)0x06, (byte)0x77, (byte)0x0E, (byte)0xE0, (byte)0xCD, (byte)0x64, (byte)0x01, (byte)0xC9, (byte)0xDD, (byte)0x5E,
		(byte)0x00, (byte)0x1C, (byte)0x20, (byte)0x04, (byte)0xD3, (byte)0xFF, (byte)0x18, (byte)0x06, (byte)0xCD, (byte)0x64, (byte)0x01, (byte)0xCD, (byte)0x75, (byte)0x01, (byte)0xC9, (byte)0xDD,
		(byte)0x7E, (byte)0x00, (byte)0xE6, (byte)0x3F, (byte)0xF5, (byte)0xDD, (byte)0x36, (byte)0x00, (byte)0xFF, (byte)0xDD, (byte)0x6E, (byte)0x01, (byte)0xDD, (byte)0x66, (byte)0x02, (byte)0x7E,
		(byte)0x47, (byte)0xCB, (byte)0x6F, (byte)0x28, (byte)0x1C, (byte)0xC5, (byte)0xE6, (byte)0x1F, (byte)0x23, (byte)0xDD, (byte)0x75, (byte)0x01, (byte)0xDD, (byte)0x74, (byte)0x02, (byte)0xDD,
		(byte)0x36, (byte)0x04, (byte)0xF0, (byte)0xDD, (byte)0x77, (byte)0x05, (byte)0xDD, (byte)0x36, (byte)0x07, (byte)0x00, (byte)0xDD, (byte)0x36, (byte)0x08, (byte)0x00, (byte)0xC3, (byte)0x61,
		(byte)0x04, (byte)0xCB, (byte)0x67, (byte)0x28, (byte)0x0F, (byte)0xCB, (byte)0x5F, (byte)0x28, (byte)0x05, (byte)0xC1, (byte)0xCD, (byte)0x5E, (byte)0x02, (byte)0xC9, (byte)0x3E, (byte)0xFF,
		(byte)0xF5, (byte)0xC3, (byte)0x61, (byte)0x04, (byte)0xE6, (byte)0x3C, (byte)0xFE, (byte)0x04, (byte)0x20, (byte)0x28, (byte)0xFD, (byte)0xE1, (byte)0xFD, (byte)0xE5, (byte)0xC5, (byte)0x23,
		(byte)0x5E, (byte)0xDD, (byte)0x73, (byte)0x01, (byte)0x23, (byte)0x56, (byte)0xDD, (byte)0x72, (byte)0x02, (byte)0x23, (byte)0xFD, (byte)0xE5, (byte)0xF1, (byte)0xD5, (byte)0xFD, (byte)0xE1,
		(byte)0x11, (byte)0xC6, (byte)0x03, (byte)0xD5, (byte)0xFD, (byte)0xE9, (byte)0x16, (byte)0x00, (byte)0x1E, (byte)0x07, (byte)0xFD, (byte)0x19, (byte)0x11, (byte)0x61, (byte)0x04, (byte)0xD5,
		(byte)0xFD, (byte)0xE9, (byte)0xC5, (byte)0x78, (byte)0xE6, (byte)0x03, (byte)0xFE, (byte)0x00, (byte)0x20, (byte)0x20, (byte)0x23, (byte)0x23, (byte)0x23, (byte)0x23, (byte)0xDD, (byte)0x75,
		(byte)0x01, (byte)0xDD, (byte)0x74, (byte)0x02, (byte)0x2B, (byte)0x11, (byte)0x05, (byte)0x00, (byte)0xCD, (byte)0x78, (byte)0x04, (byte)0x01, (byte)0x03, (byte)0x00, (byte)0xED, (byte)0xB8,
		(byte)0xDD, (byte)0x36, (byte)0x07, (byte)0x00, (byte)0xDD, (byte)0x36, (byte)0x08, (byte)0x00, (byte)0x18, (byte)0x67, (byte)0xFE, (byte)0x01, (byte)0x20, (byte)0x1B, (byte)0x1E, (byte)0x06,
		(byte)0x16, (byte)0x00, (byte)0x19, (byte)0xDD, (byte)0x75, (byte)0x01, (byte)0xDD, (byte)0x74, (byte)0x02, (byte)0x2B, (byte)0x1C, (byte)0xCD, (byte)0x78, (byte)0x04, (byte)0x01, (byte)0x05,
		(byte)0x00, (byte)0xED, (byte)0xB8, (byte)0xDD, (byte)0x36, (byte)0x08, (byte)0x00, (byte)0x18, (byte)0x48, (byte)0xFE, (byte)0x02, (byte)0x20, (byte)0x28, (byte)0x1E, (byte)0x06, (byte)0x16,
		(byte)0x00, (byte)0x19, (byte)0xF1, (byte)0xF5, (byte)0xE6, (byte)0xC0, (byte)0x20, (byte)0x01, (byte)0x2B, (byte)0xDD, (byte)0x75, (byte)0x01, (byte)0xDD, (byte)0x74, (byte)0x02, (byte)0x2B,
		(byte)0x1E, (byte)0x09, (byte)0xCD, (byte)0x78, (byte)0x04, (byte)0x01, (byte)0x02, (byte)0x00, (byte)0xED, (byte)0xB8, (byte)0x3E, (byte)0x00, (byte)0x12, (byte)0x1B, (byte)0x1B, (byte)0x0E,
		(byte)0x03, (byte)0xED, (byte)0xB8, (byte)0x18, (byte)0x1C, (byte)0x1E, (byte)0x08, (byte)0x16, (byte)0x00, (byte)0x19, (byte)0xDD, (byte)0x75, (byte)0x01, (byte)0xDD, (byte)0x74, (byte)0x02,
		(byte)0x2B, (byte)0xDD, (byte)0xE5, (byte)0xFD, (byte)0xE1, (byte)0x1E, (byte)0x09, (byte)0xFD, (byte)0x19, (byte)0xFD, (byte)0xE5, (byte)0xD1, (byte)0x01, (byte)0x07, (byte)0x00, (byte)0xED,
		(byte)0xB8, (byte)0xDD, (byte)0xE5, (byte)0xE1, (byte)0xF1, (byte)0xC1, (byte)0xFE, (byte)0xFF, (byte)0xC8, (byte)0x57, (byte)0xE6, (byte)0x3F, (byte)0xFE, (byte)0x04, (byte)0x20, (byte)0x02,
		(byte)0x06, (byte)0x3E, (byte)0x7A, (byte)0xE6, (byte)0xC0, (byte)0xB0, (byte)0x77, (byte)0xC9, (byte)0xDD, (byte)0xE5, (byte)0xFD, (byte)0xE1, (byte)0xFD, (byte)0x19, (byte)0xFD, (byte)0xE5,
		(byte)0xD1, (byte)0xC9, (byte)0x02, (byte)0x00, (byte)0xFE, (byte)0xFF, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0x82, (byte)0x04, (byte)0x11, (byte)0xBA, (byte)0x73, (byte)0xCD, (byte)0x98,
		(byte)0x00, (byte)0x2A, (byte)0xBA, (byte)0x73, (byte)0x5E, (byte)0x23, (byte)0x56, (byte)0xEB, (byte)0x3A, (byte)0xBC, (byte)0x73, (byte)0xFE, (byte)0x00, (byte)0x28, (byte)0x03, (byte)0x37,
		(byte)0x18, (byte)0x01, (byte)0xB7, (byte)0x5E, (byte)0x23, (byte)0x56, (byte)0x23, (byte)0x4E, (byte)0x23, (byte)0x46, (byte)0x23, (byte)0x3E, (byte)0x00, (byte)0x02, (byte)0x1A, (byte)0xF5,
		(byte)0xE6, (byte)0x0F, (byte)0xCA, (byte)0xE7, (byte)0x04, (byte)0x3D, (byte)0xCA, (byte)0xF1, (byte)0x05, (byte)0x3D, (byte)0xCA, (byte)0x00, (byte)0x06, (byte)0x3D, (byte)0xCA, (byte)0x00,
		(byte)0x06, (byte)0x3D, (byte)0x28, (byte)0x02, (byte)0xF1, (byte)0xC9, (byte)0x1A, (byte)0x1F, (byte)0x1F, (byte)0x1F, (byte)0x1F, (byte)0xE6, (byte)0x0F, (byte)0x47, (byte)0x5E, (byte)0x23,
		(byte)0x56, (byte)0x23, (byte)0xB7, (byte)0x28, (byte)0x10, (byte)0xF1, (byte)0xF5, (byte)0xE5, (byte)0xC5, (byte)0xEB, (byte)0xCD, (byte)0xA3, (byte)0x04, (byte)0xC1, (byte)0xE1, (byte)0x5E,
		(byte)0x23, (byte)0x56, (byte)0x23, (byte)0x10, (byte)0xF0, (byte)0xF1, (byte)0xC9, (byte)0xCD, (byte)0x72, (byte)0x05, (byte)0x1A, (byte)0x6F, (byte)0x13, (byte)0x1A, (byte)0x85, (byte)0xFD,
		(byte)0x77, (byte)0x05, (byte)0x26, (byte)0x00, (byte)0xF1, (byte)0x30, (byte)0x38, (byte)0xF5, (byte)0x3A, (byte)0xC3, (byte)0x73, (byte)0xCB, (byte)0x4F, (byte)0x28, (byte)0x31, (byte)0xEB,
		(byte)0x44, (byte)0x4D, (byte)0x6E, (byte)0x26, (byte)0x00, (byte)0xE5, (byte)0x29, (byte)0x29, (byte)0x29, (byte)0xE5, (byte)0x03, (byte)0x0A, (byte)0x6F, (byte)0x03, (byte)0x0A, (byte)0x67,
		(byte)0xC1, (byte)0xFD, (byte)0xE1, (byte)0xF1, (byte)0xCB, (byte)0x7F, (byte)0x28, (byte)0x03, (byte)0xCD, (byte)0x94, (byte)0x05, (byte)0xCD, (byte)0xE8, (byte)0x05, (byte)0xCB, (byte)0x77,
		(byte)0x28, (byte)0x03, (byte)0xCD, (byte)0x94, (byte)0x05, (byte)0xCD, (byte)0xE8, (byte)0x05, (byte)0xCB, (byte)0x6F, (byte)0x28, (byte)0x03, (byte)0xCD, (byte)0x94, (byte)0x05, (byte)0xC9,
		(byte)0xEB, (byte)0x4E, (byte)0x06, (byte)0x00, (byte)0xC5, (byte)0xFD, (byte)0xE1, (byte)0x23, (byte)0x7E, (byte)0x23, (byte)0x66, (byte)0x6F, (byte)0xE5, (byte)0xC5, (byte)0xD5, (byte)0xFD,
		(byte)0xE5, (byte)0x3E, (byte)0x03, (byte)0xCD, (byte)0x27, (byte)0x1C, (byte)0xC1, (byte)0xE1, (byte)0x5D, (byte)0x54, (byte)0x09, (byte)0x2B, (byte)0xCB, (byte)0x3C, (byte)0xCB, (byte)0x1D,
		(byte)0xCB, (byte)0x3C, (byte)0xCB, (byte)0x1D, (byte)0xCB, (byte)0x3C, (byte)0xCB, (byte)0x1D, (byte)0xCB, (byte)0x2B, (byte)0xCB, (byte)0x2B, (byte)0xCB, (byte)0x2B, (byte)0xB7, (byte)0xED,
		(byte)0x52, (byte)0x23, (byte)0xE5, (byte)0xFD, (byte)0xE1, (byte)0xE1, (byte)0x29, (byte)0x29, (byte)0x29, (byte)0xC1, (byte)0x09, (byte)0x3E, (byte)0x04, (byte)0xCD, (byte)0x27, (byte)0x1C,
		(byte)0xF1, (byte)0xC9, (byte)0xC5, (byte)0xFD, (byte)0xE1, (byte)0xD5, (byte)0x5E, (byte)0x23, (byte)0x56, (byte)0xCB, (byte)0x7A, (byte)0x20, (byte)0x14, (byte)0x7A, (byte)0xFE, (byte)0x70,
		(byte)0x38, (byte)0x06, (byte)0x3E, (byte)0x80, (byte)0x12, (byte)0x18, (byte)0x0A, (byte)0x80, (byte)0x21, (byte)0x87, (byte)0x05, (byte)0x01, (byte)0x01, (byte)0x00, (byte)0xCD, (byte)0x01,
		(byte)0x1D, (byte)0xD1, (byte)0x13, (byte)0xC9, (byte)0xF5, (byte)0xC5, (byte)0xFD, (byte)0xE5, (byte)0xD5, (byte)0xE5, (byte)0x3E, (byte)0x03, (byte)0xCD, (byte)0x27, (byte)0x1C, (byte)0xE1,
		(byte)0xD1, (byte)0xFD, (byte)0xE1, (byte)0xC1, (byte)0xF1, (byte)0xF5, (byte)0xC5, (byte)0xFD, (byte)0xE5, (byte)0xD5, (byte)0xE5, (byte)0xCB, (byte)0x67, (byte)0x20, (byte)0x0D, (byte)0x09,
		(byte)0x3E, (byte)0x04, (byte)0xCD, (byte)0x27, (byte)0x1C, (byte)0xE1, (byte)0xD1, (byte)0xFD, (byte)0xE1, (byte)0xC1, (byte)0xF1, (byte)0xC9, (byte)0x09, (byte)0x4D, (byte)0x44, (byte)0xFD,
		(byte)0xE5, (byte)0xE1, (byte)0xE5, (byte)0x0A, (byte)0xC5, (byte)0x01, (byte)0x08, (byte)0x00, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x09, (byte)0x06, (byte)0x08, (byte)0x2B, (byte)0x77,
		(byte)0x10, (byte)0xFC, (byte)0xD5, (byte)0xFD, (byte)0x21, (byte)0x01, (byte)0x00, (byte)0x3E, (byte)0x04, (byte)0xCD, (byte)0x27, (byte)0x1C, (byte)0xD1, (byte)0xC1, (byte)0x13, (byte)0x03,
		(byte)0xE1, (byte)0x2B, (byte)0x7C, (byte)0xB5, (byte)0x20, (byte)0xDC, (byte)0x18, (byte)0xCD, (byte)0xC5, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0xEB, (byte)0x09, (byte)0xEB, (byte)0xC1,
		(byte)0xC9, (byte)0xCD, (byte)0x72, (byte)0x05, (byte)0x13, (byte)0x1A, (byte)0xFD, (byte)0x77, (byte)0x05, (byte)0x13, (byte)0x1A, (byte)0xFD, (byte)0x77, (byte)0x06, (byte)0xF1, (byte)0xC9,
		(byte)0x03, (byte)0x03, (byte)0x03, (byte)0x03, (byte)0x03, (byte)0xEB, (byte)0x23, (byte)0x7E, (byte)0x5F, (byte)0x16, (byte)0x00, (byte)0xD5, (byte)0x23, (byte)0x5E, (byte)0x23, (byte)0x56,
		(byte)0x23, (byte)0x86, (byte)0x02, (byte)0x4E, (byte)0x06, (byte)0x00, (byte)0xC5, (byte)0xFD, (byte)0xE1, (byte)0xEB, (byte)0xD1, (byte)0xF1, (byte)0xD0, (byte)0x3E, (byte)0x01, (byte)0xCD,
		(byte)0x27, (byte)0x1C, (byte)0xC9, (byte)0xDD, (byte)0xE5, (byte)0x2A, (byte)0xCD, (byte)0x73, (byte)0xD1, (byte)0x73, (byte)0x23, (byte)0x72, (byte)0x23, (byte)0x70, (byte)0x23, (byte)0xEB,
		(byte)0x3A, (byte)0xCB, (byte)0x73, (byte)0x3C, (byte)0x21, (byte)0xCA, (byte)0x73, (byte)0xBE, (byte)0x20, (byte)0x0D, (byte)0x3E, (byte)0x00, (byte)0x32, (byte)0xCB, (byte)0x73, (byte)0x2A,
		(byte)0xD1, (byte)0x73, (byte)0x22, (byte)0xCD, (byte)0x73, (byte)0x18, (byte)0x07, (byte)0x32, (byte)0xCB, (byte)0x73, (byte)0xED, (byte)0x53, (byte)0xCD, (byte)0x73, (byte)0xC9, (byte)0x02,
		(byte)0x00, (byte)0x01, (byte)0x00, (byte)0xFE, (byte)0xFF, (byte)0x01, (byte)0x4F, (byte)0x06, (byte)0x11, (byte)0xBA, (byte)0x73, (byte)0xCD, (byte)0x98, (byte)0x00, (byte)0x3A, (byte)0xBA,
		(byte)0x73, (byte)0x2A, (byte)0xBB, (byte)0x73, (byte)0x32, (byte)0xCA, (byte)0x73, (byte)0x3E, (byte)0x00, (byte)0x32, (byte)0xCB, (byte)0x73, (byte)0x32, (byte)0xCC, (byte)0x73, (byte)0x22,
		(byte)0xD1, (byte)0x73, (byte)0x22, (byte)0xCD, (byte)0x73, (byte)0x22, (byte)0xCF, (byte)0x73, (byte)0xC9, (byte)0x3A, (byte)0xC6, (byte)0x73, (byte)0xF5, (byte)0x3E, (byte)0x00, (byte)0x32,
		(byte)0xC6, (byte)0x73, (byte)0x3A, (byte)0xCC, (byte)0x73, (byte)0x21, (byte)0xCB, (byte)0x73, (byte)0xBE, (byte)0x28, (byte)0x31, (byte)0x2A, (byte)0xCF, (byte)0x73, (byte)0x5E, (byte)0x23,
		(byte)0x56, (byte)0x23, (byte)0x46, (byte)0x23, (byte)0xD5, (byte)0xDD, (byte)0xE1, (byte)0xE5, (byte)0xCD, (byte)0xE3, (byte)0x06, (byte)0x3A, (byte)0xCC, (byte)0x73, (byte)0x3C, (byte)0x21,
		(byte)0xCA, (byte)0x73, (byte)0xBE, (byte)0x20, (byte)0x0E, (byte)0x3E, (byte)0x00, (byte)0x32, (byte)0xCC, (byte)0x73, (byte)0x2A, (byte)0xD1, (byte)0x73, (byte)0x22, (byte)0xCF, (byte)0x73,
		(byte)0xE1, (byte)0x18, (byte)0x07, (byte)0x32, (byte)0xCC, (byte)0x73, (byte)0xE1, (byte)0x22, (byte)0xCF, (byte)0x73, (byte)0x18, (byte)0xC6, (byte)0xF1, (byte)0x32, (byte)0xC6, (byte)0x73,
		(byte)0xC9, (byte)0x02, (byte)0x00, (byte)0x02, (byte)0x00, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0xC1, (byte)0x06, (byte)0x11, (byte)0xBA, (byte)0x73, (byte)0xCD, (byte)0x98, (byte)0x00,
		(byte)0xDD, (byte)0x2A, (byte)0xBA, (byte)0x73, (byte)0x3A, (byte)0xBC, (byte)0x73, (byte)0x47, (byte)0x3A, (byte)0xC6, (byte)0x73, (byte)0xFE, (byte)0x01, (byte)0x20, (byte)0x04, (byte)0xCD,
		(byte)0x23, (byte)0x06, (byte)0xC9, (byte)0xDD, (byte)0x66, (byte)0x01, (byte)0xDD, (byte)0x6E, (byte)0x00, (byte)0x7E, (byte)0x4F, (byte)0xE6, (byte)0x0F, (byte)0xCA, (byte)0xFF, (byte)0x06,
		(byte)0x3D, (byte)0xCA, (byte)0x87, (byte)0x0A, (byte)0x3D, (byte)0xCA, (byte)0xDF, (byte)0x08, (byte)0x3D, (byte)0xCA, (byte)0x55, (byte)0x09, (byte)0xC3, (byte)0xA2, (byte)0x0E, (byte)0xDD,
		(byte)0x56, (byte)0x03, (byte)0xDD, (byte)0x5E, (byte)0x02, (byte)0xD5, (byte)0xFD, (byte)0xE1, (byte)0xFD, (byte)0x56, (byte)0x02, (byte)0xFD, (byte)0x5E, (byte)0x01, (byte)0xCD, (byte)0xE8,
		(byte)0x07, (byte)0x4B, (byte)0xFD, (byte)0x56, (byte)0x04, (byte)0xFD, (byte)0x5E, (byte)0x03, (byte)0xCD, (byte)0xE8, (byte)0x07, (byte)0x43, (byte)0xFD, (byte)0x5E, (byte)0x00, (byte)0x16,
		(byte)0x00, (byte)0x19, (byte)0x19, (byte)0x1E, (byte)0x05, (byte)0x19, (byte)0x5E, (byte)0x23, (byte)0x56, (byte)0xEB, (byte)0xC5, (byte)0xD1, (byte)0x4E, (byte)0x23, (byte)0x46, (byte)0x23,
		(byte)0xDD, (byte)0x7E, (byte)0x05, (byte)0xCB, (byte)0x7F, (byte)0x28, (byte)0x04, (byte)0xCD, (byte)0x0B, (byte)0x08, (byte)0xC9, (byte)0xC5, (byte)0xD5, (byte)0xE5, (byte)0xFE, (byte)0x70,
		(byte)0x28, (byte)0x02, (byte)0x38, (byte)0x07, (byte)0x67, (byte)0xDD, (byte)0x6E, (byte)0x04, (byte)0x7E, (byte)0x18, (byte)0x35, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0xDD, (byte)0x56,
		(byte)0x05, (byte)0xDD, (byte)0x5E, (byte)0x04, (byte)0xE5, (byte)0xD5, (byte)0xE5, (byte)0x01, (byte)0x04, (byte)0x00, (byte)0xCD, (byte)0x3E, (byte)0x1D, (byte)0xE1, (byte)0x7E, (byte)0xFE,
		(byte)0x80, (byte)0x20, (byte)0x03, (byte)0xD1, (byte)0x18, (byte)0x19, (byte)0x23, (byte)0x23, (byte)0x46, (byte)0x23, (byte)0x5E, (byte)0x16, (byte)0x00, (byte)0x23, (byte)0xEB, (byte)0x18,
		(byte)0x01, (byte)0x29, (byte)0x10, (byte)0xFD, (byte)0xE5, (byte)0xC1, (byte)0xEB, (byte)0xD1, (byte)0x13, (byte)0x13, (byte)0x13, (byte)0x13, (byte)0xCD, (byte)0x3E, (byte)0x1D, (byte)0xE1,
		(byte)0x7E, (byte)0xFE, (byte)0x80, (byte)0x28, (byte)0x0F, (byte)0x5E, (byte)0x23, (byte)0x56, (byte)0x23, (byte)0x4E, (byte)0x23, (byte)0x46, (byte)0x23, (byte)0xDD, (byte)0xE5, (byte)0xCD,
		(byte)0x0B, (byte)0x08, (byte)0xDD, (byte)0xE1, (byte)0xE1, (byte)0xD1, (byte)0xC1, (byte)0xC5, (byte)0xD5, (byte)0xE5, (byte)0xDD, (byte)0x66, (byte)0x05, (byte)0xDD, (byte)0x6E, (byte)0x04,
		(byte)0x3E, (byte)0x70, (byte)0xBC, (byte)0x38, (byte)0x03, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x73, (byte)0x23, (byte)0x72, (byte)0x23, (byte)0x71, (byte)0x23, (byte)0x70, (byte)0x23,
		(byte)0xDD, (byte)0xE5, (byte)0xCD, (byte)0x98, (byte)0x08, (byte)0xDD, (byte)0xE1, (byte)0xE1, (byte)0xD1, (byte)0xC1, (byte)0xDD, (byte)0xE5, (byte)0xCD, (byte)0x0B, (byte)0x08, (byte)0xDD,
		(byte)0xE1, (byte)0xDD, (byte)0x56, (byte)0x05, (byte)0x3E, (byte)0x70, (byte)0xBA, (byte)0x28, (byte)0x1E, (byte)0x38, (byte)0x1C, (byte)0xDD, (byte)0x5E, (byte)0x04, (byte)0xD9, (byte)0x2A,
		(byte)0x06, (byte)0x80, (byte)0xE5, (byte)0x23, (byte)0x23, (byte)0x5E, (byte)0x16, (byte)0x00, (byte)0x23, (byte)0x46, (byte)0xEB, (byte)0x18, (byte)0x01, (byte)0x29, (byte)0x10, (byte)0xFD,
		(byte)0xE5, (byte)0xD9, (byte)0xC1, (byte)0xE1, (byte)0xCD, (byte)0x01, (byte)0x1D, (byte)0xC9, (byte)0xE5, (byte)0xCB, (byte)0x2A, (byte)0xCB, (byte)0x1B, (byte)0xCB, (byte)0x2A, (byte)0xCB,
		(byte)0x1B, (byte)0xCB, (byte)0x2A, (byte)0xCB, (byte)0x1B, (byte)0xCB, (byte)0x7A, (byte)0x20, (byte)0x09, (byte)0x21, (byte)0x80, (byte)0xFF, (byte)0x19, (byte)0xE1, (byte)0xD0, (byte)0x1E,
		(byte)0x7F, (byte)0xC9, (byte)0x21, (byte)0x80, (byte)0x00, (byte)0x19, (byte)0xE1, (byte)0xD8, (byte)0x1E, (byte)0x80, (byte)0xC9, (byte)0xC5, (byte)0xD5, (byte)0xE5, (byte)0xD9, (byte)0xE1,
		(byte)0xD1, (byte)0xC1, (byte)0xCD, (byte)0xC0, (byte)0x08, (byte)0xD9, (byte)0x7B, (byte)0xCB, (byte)0x7F, (byte)0x20, (byte)0x03, (byte)0xFE, (byte)0x20, (byte)0xD0, (byte)0x81, (byte)0xCB,
		(byte)0x7F, (byte)0xC0, (byte)0xB7, (byte)0xC8, (byte)0xCB, (byte)0x7B, (byte)0x28, (byte)0x20, (byte)0x79, (byte)0x83, (byte)0xD5, (byte)0xFE, (byte)0x21, (byte)0x38, (byte)0x02, (byte)0x3E,
		(byte)0x20, (byte)0x5F, (byte)0x16, (byte)0x00, (byte)0xD5, (byte)0xFD, (byte)0xE1, (byte)0xD1, (byte)0x7B, (byte)0xD9, (byte)0xC5, (byte)0xED, (byte)0x44, (byte)0x4F, (byte)0x06, (byte)0x00,
		(byte)0x09, (byte)0xEB, (byte)0x09, (byte)0xEB, (byte)0xC1, (byte)0xD9, (byte)0x18, (byte)0x1C, (byte)0x7B, (byte)0x81, (byte)0xFE, (byte)0x1F, (byte)0x28, (byte)0x0F, (byte)0x38, (byte)0x0D,
		(byte)0x3E, (byte)0x20, (byte)0x93, (byte)0xD5, (byte)0x5F, (byte)0x16, (byte)0x00, (byte)0xD5, (byte)0xFD, (byte)0xE1, (byte)0xD1, (byte)0x18, (byte)0x07, (byte)0xC5, (byte)0x06, (byte)0x00,
		(byte)0xC5, (byte)0xFD, (byte)0xE1, (byte)0xC1, (byte)0x1E, (byte)0x00, (byte)0x7A, (byte)0x83, (byte)0xCB, (byte)0x7F, (byte)0x20, (byte)0x19, (byte)0xFE, (byte)0x18, (byte)0x30, (byte)0x15,
		(byte)0xC5, (byte)0xD5, (byte)0xD9, (byte)0xC5, (byte)0xD5, (byte)0xE5, (byte)0xFD, (byte)0xE5, (byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0x27, (byte)0x1C, (byte)0xFD, (byte)0xE1, (byte)0xE1,
		(byte)0xD1, (byte)0xC1, (byte)0xD9, (byte)0xD1, (byte)0xC1, (byte)0xD9, (byte)0xC5, (byte)0x06, (byte)0x00, (byte)0x09, (byte)0xEB, (byte)0x01, (byte)0x20, (byte)0x00, (byte)0x09, (byte)0xEB,
		(byte)0xC1, (byte)0xD9, (byte)0x1C, (byte)0x7B, (byte)0xB8, (byte)0x20, (byte)0xCF, (byte)0xC9, (byte)0xCD, (byte)0xC0, (byte)0x08, (byte)0xC5, (byte)0x06, (byte)0x00, (byte)0xC5, (byte)0xFD,
		(byte)0xE1, (byte)0xC1, (byte)0xC5, (byte)0xD5, (byte)0xE5, (byte)0xFD, (byte)0xE5, (byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0xA3, (byte)0x1B, (byte)0xFD, (byte)0xE1, (byte)0xE1, (byte)0xD1,
		(byte)0xC1, (byte)0xC5, (byte)0x06, (byte)0x00, (byte)0x09, (byte)0x01, (byte)0x20, (byte)0x00, (byte)0xEB, (byte)0x09, (byte)0xEB, (byte)0xC1, (byte)0x05, (byte)0x20, (byte)0xE3, (byte)0xC9,
		(byte)0xE5, (byte)0xCB, (byte)0x7A, (byte)0x28, (byte)0x04, (byte)0x26, (byte)0xFF, (byte)0x18, (byte)0x02, (byte)0x26, (byte)0x00, (byte)0x6A, (byte)0x29, (byte)0x29, (byte)0x29, (byte)0x29,
		(byte)0x29, (byte)0xCB, (byte)0x7B, (byte)0x28, (byte)0x04, (byte)0x16, (byte)0xFF, (byte)0x18, (byte)0x02, (byte)0x16, (byte)0x00, (byte)0x19, (byte)0xEB, (byte)0xE1, (byte)0xC9, (byte)0xFD,
		(byte)0x2A, (byte)0x06, (byte)0x80, (byte)0xDD, (byte)0x6E, (byte)0x02, (byte)0xDD, (byte)0x66, (byte)0x03, (byte)0x11, (byte)0x01, (byte)0x00, (byte)0x19, (byte)0x4E, (byte)0x23, (byte)0x46,
		(byte)0x78, (byte)0xFE, (byte)0x00, (byte)0x28, (byte)0x0B, (byte)0xFE, (byte)0xFF, (byte)0xC2, (byte)0x54, (byte)0x0A, (byte)0x79, (byte)0xFE, (byte)0xF9, (byte)0xFA, (byte)0x54, (byte)0x0A,
		(byte)0x23, (byte)0x4E, (byte)0x23, (byte)0x46, (byte)0x78, (byte)0xFE, (byte)0x00, (byte)0x28, (byte)0x0B, (byte)0xFE, (byte)0xFF, (byte)0xC2, (byte)0x54, (byte)0x0A, (byte)0x79, (byte)0xFE,
		(byte)0xF9, (byte)0xFA, (byte)0x54, (byte)0x0A, (byte)0x2B, (byte)0x2B, (byte)0x7E, (byte)0xFE, (byte)0x00, (byte)0xCA, (byte)0xCA, (byte)0x09, (byte)0x2B, (byte)0x4E, (byte)0x23, (byte)0x46,
		(byte)0x21, (byte)0x08, (byte)0x00, (byte)0x09, (byte)0x7D, (byte)0xFD, (byte)0x77, (byte)0x01, (byte)0xDD, (byte)0x6E, (byte)0x00, (byte)0xDD, (byte)0x66, (byte)0x01, (byte)0x11, (byte)0x05,
		(byte)0x00, (byte)0x19, (byte)0xEB, (byte)0x1A, (byte)0x6F, (byte)0x13, (byte)0x1A, (byte)0x67, (byte)0xE5, (byte)0xDD, (byte)0x6E, (byte)0x02, (byte)0xDD, (byte)0x66, (byte)0x03, (byte)0x11,
		(byte)0x00, (byte)0x00, (byte)0x19, (byte)0x7E, (byte)0xCB, (byte)0x27, (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x4F, (byte)0xE1, (byte)0x09, (byte)0x7E, (byte)0xF6, (byte)0x80, (byte)0xFD,
		(byte)0x77, (byte)0x03, (byte)0xC3, (byte)0x00, (byte)0x0A, (byte)0xFD, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0xDD, (byte)0x6E, (byte)0x02, (byte)0xDD, (byte)0x66, (byte)0x03, (byte)0x11,
		(byte)0x01, (byte)0x00, (byte)0x19, (byte)0x4E, (byte)0x23, (byte)0x46, (byte)0x78, (byte)0xFE, (byte)0x00, (byte)0x28, (byte)0x0B, (byte)0xFE, (byte)0xFF, (byte)0xC2, (byte)0x54, (byte)0x0A,
		(byte)0x79, (byte)0xFE, (byte)0xE1, (byte)0xFA, (byte)0x54, (byte)0x0A, (byte)0x23, (byte)0x4E, (byte)0x23, (byte)0x46, (byte)0x78, (byte)0xFE, (byte)0x00, (byte)0x28, (byte)0x0B, (byte)0xFE,
		(byte)0xFF, (byte)0xC2, (byte)0x54, (byte)0x0A, (byte)0x79, (byte)0xFE, (byte)0xE1, (byte)0xFA, (byte)0x54, (byte)0x0A, (byte)0x2B, (byte)0x2B, (byte)0x7E, (byte)0xFE, (byte)0x00, (byte)0xCA,
		(byte)0xCA, (byte)0x09, (byte)0x2B, (byte)0x4E, (byte)0x23, (byte)0x46, (byte)0x21, (byte)0x20, (byte)0x00, (byte)0x09, (byte)0x7D, (byte)0xFD, (byte)0x77, (byte)0x01, (byte)0xDD, (byte)0x6E,
		(byte)0x00, (byte)0xDD, (byte)0x66, (byte)0x01, (byte)0x11, (byte)0x05, (byte)0x00, (byte)0x19, (byte)0xEB, (byte)0x1A, (byte)0x6F, (byte)0x13, (byte)0x1A, (byte)0x67, (byte)0xE5, (byte)0xDD,
		(byte)0x6E, (byte)0x02, (byte)0xDD, (byte)0x66, (byte)0x03, (byte)0x11, (byte)0x00, (byte)0x00, (byte)0x19, (byte)0x7E, (byte)0xCB, (byte)0x27, (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x4F,
		(byte)0xE1, (byte)0x09, (byte)0x7E, (byte)0xF6, (byte)0x80, (byte)0xFD, (byte)0x77, (byte)0x03, (byte)0x18, (byte)0x36, (byte)0xDD, (byte)0x6E, (byte)0x02, (byte)0xDD, (byte)0x66, (byte)0x03,
		(byte)0x11, (byte)0x01, (byte)0x00, (byte)0x19, (byte)0x7E, (byte)0xFD, (byte)0x77, (byte)0x01, (byte)0xDD, (byte)0x6E, (byte)0x00, (byte)0xDD, (byte)0x66, (byte)0x01, (byte)0x11, (byte)0x05,
		(byte)0x00, (byte)0x19, (byte)0xEB, (byte)0x1A, (byte)0x6F, (byte)0x13, (byte)0x1A, (byte)0x67, (byte)0xE5, (byte)0xDD, (byte)0x6E, (byte)0x02, (byte)0xDD, (byte)0x66, (byte)0x03, (byte)0x11,
		(byte)0x00, (byte)0x00, (byte)0x19, (byte)0x7E, (byte)0xCB, (byte)0x27, (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x4F, (byte)0xE1, (byte)0x09, (byte)0x7E, (byte)0xFD, (byte)0x77, (byte)0x03,
		(byte)0xDD, (byte)0x6E, (byte)0x02, (byte)0xDD, (byte)0x66, (byte)0x03, (byte)0x11, (byte)0x03, (byte)0x00, (byte)0x19, (byte)0x7E, (byte)0xFD, (byte)0x77, (byte)0x00, (byte)0xDD, (byte)0x6E,
		(byte)0x00, (byte)0xDD, (byte)0x66, (byte)0x01, (byte)0x11, (byte)0x05, (byte)0x00, (byte)0x19, (byte)0xEB, (byte)0x1A, (byte)0x6F, (byte)0x13, (byte)0x1A, (byte)0x67, (byte)0xE5, (byte)0xDD,
		(byte)0x6E, (byte)0x02, (byte)0xDD, (byte)0x66, (byte)0x03, (byte)0x11, (byte)0x00, (byte)0x00, (byte)0x19, (byte)0x7E, (byte)0xCB, (byte)0x27, (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x4F,
		(byte)0xE1, (byte)0x09, (byte)0x23, (byte)0x7E, (byte)0xDD, (byte)0x6E, (byte)0x00, (byte)0xDD, (byte)0x66, (byte)0x01, (byte)0x11, (byte)0x01, (byte)0x00, (byte)0x19, (byte)0x86, (byte)0xFD,
		(byte)0x77, (byte)0x02, (byte)0xAF, (byte)0x16, (byte)0x00, (byte)0xDD, (byte)0x5E, (byte)0x04, (byte)0xFD, (byte)0xE5, (byte)0xE1, (byte)0xFD, (byte)0x21, (byte)0x01, (byte)0x00, (byte)0xCD,
		(byte)0xBE, (byte)0x1F, (byte)0x18, (byte)0x32, (byte)0xFD, (byte)0xE5, (byte)0xDD, (byte)0xE5, (byte)0xFD, (byte)0xE5, (byte)0xFD, (byte)0xE5, (byte)0xAF, (byte)0x16, (byte)0x00, (byte)0xDD,
		(byte)0x5E, (byte)0x04, (byte)0xE1, (byte)0xFD, (byte)0x21, (byte)0x01, (byte)0x00, (byte)0xCD, (byte)0xBB, (byte)0x1F, (byte)0x3E, (byte)0x00, (byte)0xFD, (byte)0xE1, (byte)0xFD, (byte)0x77,
		(byte)0x01, (byte)0x3E, (byte)0x80, (byte)0xFD, (byte)0x77, (byte)0x03, (byte)0xAF, (byte)0x16, (byte)0x00, (byte)0xDD, (byte)0xE1, (byte)0xDD, (byte)0x5E, (byte)0x04, (byte)0xE1, (byte)0xFD,
		(byte)0x21, (byte)0x01, (byte)0x00, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0xC9, (byte)0xFD, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x3A, (byte)0xC3, (byte)0x73, (byte)0xCB, (byte)0x4F,
		(byte)0x20, (byte)0x04, (byte)0xCB, (byte)0xB8, (byte)0x18, (byte)0x02, (byte)0xCB, (byte)0xF8, (byte)0xFD, (byte)0x70, (byte)0x03, (byte)0xE5, (byte)0xDD, (byte)0x66, (byte)0x03, (byte)0xDD,
		(byte)0x6E, (byte)0x02, (byte)0x7E, (byte)0xFD, (byte)0x77, (byte)0x04, (byte)0xEE, (byte)0x80, (byte)0x77, (byte)0x23, (byte)0x5E, (byte)0x7B, (byte)0xE6, (byte)0x07, (byte)0xED, (byte)0x44,
		(byte)0xC6, (byte)0x08, (byte)0xFD, (byte)0x77, (byte)0x01, (byte)0x23, (byte)0x56, (byte)0xCD, (byte)0xE8, (byte)0x07, (byte)0xFD, (byte)0x73, (byte)0x11, (byte)0x23, (byte)0x5E, (byte)0x7B,
		(byte)0xE6, (byte)0x07, (byte)0xFD, (byte)0x77, (byte)0x00, (byte)0x23, (byte)0x56, (byte)0xCD, (byte)0xE8, (byte)0x07, (byte)0xFD, (byte)0x73, (byte)0x12, (byte)0x2A, (byte)0x06, (byte)0x80,
		(byte)0x11, (byte)0x13, (byte)0x00, (byte)0x19, (byte)0xFD, (byte)0x56, (byte)0x12, (byte)0xFD, (byte)0x5E, (byte)0x11, (byte)0x01, (byte)0x03, (byte)0x03, (byte)0xCD, (byte)0x98, (byte)0x08,
		(byte)0xDD, (byte)0x56, (byte)0x05, (byte)0xDD, (byte)0x5E, (byte)0x04, (byte)0xDD, (byte)0x7E, (byte)0x06, (byte)0xDD, (byte)0xE1, (byte)0xFD, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0xFD,
		(byte)0x77, (byte)0x05, (byte)0xD5, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x01, (byte)0x06, (byte)0x00, (byte)0x09, (byte)0x01, (byte)0x0B, (byte)0x00, (byte)0x7A, (byte)0xFE, (byte)0x70,
		(byte)0x30, (byte)0x05, (byte)0xCD, (byte)0xE2, (byte)0x1F, (byte)0x18, (byte)0x03, (byte)0xEB, (byte)0xED, (byte)0xB0, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x11, (byte)0x13, (byte)0x00,
		(byte)0x19, (byte)0xD9, (byte)0xED, (byte)0x5B, (byte)0x06, (byte)0x80, (byte)0x21, (byte)0x08, (byte)0x00, (byte)0x19, (byte)0xEB, (byte)0xD9, (byte)0xFD, (byte)0x2A, (byte)0x06, (byte)0x80,
		(byte)0xFD, (byte)0x4E, (byte)0x05, (byte)0x06, (byte)0x09, (byte)0x7E, (byte)0x91, (byte)0xFE, (byte)0x12, (byte)0x30, (byte)0x0E, (byte)0xFE, (byte)0x09, (byte)0x38, (byte)0x02, (byte)0xD6,
		(byte)0x09, (byte)0xD9, (byte)0x6F, (byte)0x26, (byte)0x00, (byte)0x19, (byte)0x7E, (byte)0xD9, (byte)0x77, (byte)0x23, (byte)0x10, (byte)0xE9, (byte)0xD1, (byte)0x2A, (byte)0x06, (byte)0x80,
		(byte)0x01, (byte)0x11, (byte)0x00, (byte)0x09, (byte)0x01, (byte)0x0B, (byte)0x00, (byte)0x7A, (byte)0xFE, (byte)0x70, (byte)0x30, (byte)0x05, (byte)0xCD, (byte)0xDF, (byte)0x1F, (byte)0x18,
		(byte)0x02, (byte)0xED, (byte)0xB0, (byte)0xDD, (byte)0xE5, (byte)0xED, (byte)0x5B, (byte)0x06, (byte)0x80, (byte)0x21, (byte)0x13, (byte)0x00, (byte)0x19, (byte)0xEB, (byte)0x01, (byte)0x14,
		(byte)0x00, (byte)0x09, (byte)0x06, (byte)0x09, (byte)0x1A, (byte)0x13, (byte)0xD5, (byte)0x11, (byte)0x08, (byte)0x00, (byte)0x19, (byte)0xE5, (byte)0x5F, (byte)0x16, (byte)0x00, (byte)0x4F,
		(byte)0xC5, (byte)0x3E, (byte)0x09, (byte)0x90, (byte)0x06, (byte)0x00, (byte)0xD6, (byte)0x03, (byte)0x38, (byte)0x03, (byte)0x04, (byte)0x18, (byte)0xF9, (byte)0x78, (byte)0xFD, (byte)0x2A,
		(byte)0x06, (byte)0x80, (byte)0xFD, (byte)0x86, (byte)0x12, (byte)0xFD, (byte)0xCB, (byte)0x03, (byte)0x7E, (byte)0xFD, (byte)0x21, (byte)0x01, (byte)0x00, (byte)0x20, (byte)0x29, (byte)0x3E,
		(byte)0x03, (byte)0xCD, (byte)0xBB, (byte)0x1F, (byte)0xC1, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0xC5, (byte)0x11, (byte)0x84, (byte)0x00, (byte)0x19, (byte)0x59, (byte)0xCB, (byte)0x3B,
		(byte)0xCB, (byte)0x3B, (byte)0xCB, (byte)0x3B, (byte)0x16, (byte)0x00, (byte)0x3E, (byte)0x09, (byte)0x90, (byte)0x4F, (byte)0x06, (byte)0x00, (byte)0x09, (byte)0xFD, (byte)0x21, (byte)0x01,
		(byte)0x00, (byte)0x3E, (byte)0x04, (byte)0xCD, (byte)0xBB, (byte)0x1F, (byte)0x18, (byte)0x21, (byte)0xCB, (byte)0x2F, (byte)0xCB, (byte)0x2F, (byte)0xCB, (byte)0x2F, (byte)0xFE, (byte)0x03,
		(byte)0x30, (byte)0x17, (byte)0x57, (byte)0xD5, (byte)0xE5, (byte)0x3E, (byte)0x03, (byte)0xCD, (byte)0xBB, (byte)0x1F, (byte)0xE1, (byte)0x11, (byte)0x68, (byte)0x00, (byte)0x19, (byte)0xD1,
		(byte)0xFD, (byte)0x21, (byte)0x01, (byte)0x00, (byte)0x3E, (byte)0x04, (byte)0xCD, (byte)0xBB, (byte)0x1F, (byte)0xC1, (byte)0xE1, (byte)0xD1, (byte)0x10, (byte)0x86, (byte)0xDD, (byte)0xE1,
		(byte)0xD9, (byte)0xDD, (byte)0x56, (byte)0x03, (byte)0xDD, (byte)0x5E, (byte)0x02, (byte)0xDD, (byte)0x46, (byte)0x05, (byte)0xDD, (byte)0x4E, (byte)0x04, (byte)0xD9, (byte)0xDD, (byte)0xE5,
		(byte)0xE1, (byte)0xFD, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0xFD, (byte)0x7E, (byte)0x04, (byte)0x87, (byte)0x4F, (byte)0x06, (byte)0x00, (byte)0x11, (byte)0x06, (byte)0x00, (byte)0x19,
		(byte)0x09, (byte)0x5E, (byte)0x23, (byte)0x56, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x01, (byte)0x7C, (byte)0x00, (byte)0x09, (byte)0xE5, (byte)0xC5, (byte)0x01, (byte)0x05, (byte)0x00,
		(byte)0x7A, (byte)0xFE, (byte)0x70, (byte)0x30, (byte)0x05, (byte)0xCD, (byte)0xE2, (byte)0x1F, (byte)0x18, (byte)0x03, (byte)0xEB, (byte)0xED, (byte)0xB0, (byte)0xFD, (byte)0x2A, (byte)0x06,
		(byte)0x80, (byte)0xC1, (byte)0xFD, (byte)0x09, (byte)0xFD, (byte)0x7E, (byte)0x04, (byte)0xFD, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0xFD, (byte)0x77, (byte)0x02, (byte)0xD1, (byte)0x2A,
		(byte)0x06, (byte)0x80, (byte)0x01, (byte)0x64, (byte)0x00, (byte)0x09, (byte)0x06, (byte)0x04, (byte)0x1A, (byte)0xDD, (byte)0xBE, (byte)0x01, (byte)0xD5, (byte)0x30, (byte)0x16, (byte)0xD9,
		(byte)0x87, (byte)0x87, (byte)0x87, (byte)0x6F, (byte)0x26, (byte)0x00, (byte)0x09, (byte)0xE5, (byte)0xD9, (byte)0xD1, (byte)0xEB, (byte)0xC5, (byte)0x01, (byte)0x08, (byte)0x00, (byte)0xED,
		(byte)0xB0, (byte)0xC1, (byte)0xEB, (byte)0x18, (byte)0x30, (byte)0xDD, (byte)0x96, (byte)0x01, (byte)0xD9, (byte)0x87, (byte)0x87, (byte)0x87, (byte)0x6F, (byte)0x26, (byte)0x00, (byte)0x19,
		(byte)0xE5, (byte)0xD9, (byte)0xD1, (byte)0x7A, (byte)0xFE, (byte)0x70, (byte)0x30, (byte)0x14, (byte)0xC5, (byte)0xE5, (byte)0xD5, (byte)0x01, (byte)0x08, (byte)0x00, (byte)0xCD, (byte)0xE2,
		(byte)0x1F, (byte)0x01, (byte)0x08, (byte)0x00, (byte)0xE1, (byte)0x09, (byte)0xEB, (byte)0xE1, (byte)0x09, (byte)0xC1, (byte)0x18, (byte)0x09, (byte)0xEB, (byte)0xC5, (byte)0x01, (byte)0x08,
		(byte)0x00, (byte)0xED, (byte)0xB0, (byte)0xC1, (byte)0xEB, (byte)0xD1, (byte)0x13, (byte)0x10, (byte)0xAF, (byte)0xFD, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0xED, (byte)0x5B, (byte)0x06,
		(byte)0x80, (byte)0x21, (byte)0x1C, (byte)0x00, (byte)0x19, (byte)0xFD, (byte)0x4E, (byte)0x00, (byte)0x06, (byte)0x00, (byte)0x09, (byte)0xE5, (byte)0xDD, (byte)0xE1, (byte)0x21, (byte)0x64,
		(byte)0x00, (byte)0x19, (byte)0xE5, (byte)0x3E, (byte)0x10, (byte)0x08, (byte)0xE1, (byte)0x56, (byte)0x23, (byte)0xE5, (byte)0x01, (byte)0x0F, (byte)0x00, (byte)0x09, (byte)0x5E, (byte)0xEB,
		(byte)0xFD, (byte)0x46, (byte)0x01, (byte)0xAF, (byte)0x05, (byte)0xFA, (byte)0xBC, (byte)0x0C, (byte)0x29, (byte)0x17, (byte)0x18, (byte)0xF8, (byte)0x5F, (byte)0xCD, (byte)0x2F, (byte)0x0E,
		(byte)0xFD, (byte)0x7E, (byte)0x00, (byte)0x3C, (byte)0xFD, (byte)0x77, (byte)0x00, (byte)0xFE, (byte)0x08, (byte)0x28, (byte)0x04, (byte)0xFE, (byte)0x10, (byte)0x20, (byte)0x05, (byte)0x01,
		(byte)0x10, (byte)0x00, (byte)0xDD, (byte)0x09, (byte)0xDD, (byte)0x23, (byte)0x08, (byte)0x3D, (byte)0x28, (byte)0x03, (byte)0x08, (byte)0x18, (byte)0xC9, (byte)0xE1, (byte)0xFD, (byte)0xCB,
		(byte)0x03, (byte)0x7E, (byte)0x20, (byte)0x1F, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x01, (byte)0x84, (byte)0x00, (byte)0x09, (byte)0xFD, (byte)0x56, (byte)0x02, (byte)0xFD, (byte)0xCB,
		(byte)0x03, (byte)0x4E, (byte)0x20, (byte)0x04, (byte)0x0E, (byte)0x0F, (byte)0x18, (byte)0x02, (byte)0x0E, (byte)0x00, (byte)0x06, (byte)0x09, (byte)0x7E, (byte)0xA1, (byte)0xB2, (byte)0x77,
		(byte)0x23, (byte)0x10, (byte)0xF9, (byte)0xFD, (byte)0x7E, (byte)0x05, (byte)0xFD, (byte)0xCB, (byte)0x04, (byte)0x7E, (byte)0x28, (byte)0x02, (byte)0xC6, (byte)0x09, (byte)0x4F, (byte)0x2A,
		(byte)0x06, (byte)0x80, (byte)0x11, (byte)0x13, (byte)0x00, (byte)0x19, (byte)0x06, (byte)0x09, (byte)0x77, (byte)0x3C, (byte)0x23, (byte)0x10, (byte)0xFB, (byte)0xFD, (byte)0xCB, (byte)0x03,
		(byte)0x7E, (byte)0x20, (byte)0x73, (byte)0x59, (byte)0x16, (byte)0x00, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x01, (byte)0x1C, (byte)0x00, (byte)0x09, (byte)0xFD, (byte)0x21, (byte)0x09,
		(byte)0x00, (byte)0x3E, (byte)0x03, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0xFD, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x01, (byte)0x84, (byte)0x00,
		(byte)0x09, (byte)0xDD, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x01, (byte)0x13, (byte)0x00, (byte)0xDD, (byte)0x09, (byte)0x06, (byte)0x09, (byte)0xDD, (byte)0x7E, (byte)0x00, (byte)0xDD,
		(byte)0x23, (byte)0xCB, (byte)0x3F, (byte)0xCB, (byte)0x3F, (byte)0xCB, (byte)0x3F, (byte)0x5F, (byte)0x16, (byte)0x00, (byte)0xC5, (byte)0x3E, (byte)0x09, (byte)0x90, (byte)0x06, (byte)0x00,
		(byte)0xFE, (byte)0x03, (byte)0x38, (byte)0x05, (byte)0xD6, (byte)0x03, (byte)0x04, (byte)0x18, (byte)0xF7, (byte)0xFD, (byte)0x86, (byte)0x11, (byte)0xFE, (byte)0x20, (byte)0x30, (byte)0x17,
		(byte)0x78, (byte)0xFD, (byte)0x86, (byte)0x12, (byte)0xFE, (byte)0x18, (byte)0x30, (byte)0x0F, (byte)0xDD, (byte)0xE5, (byte)0xE5, (byte)0x3E, (byte)0x04, (byte)0xFD, (byte)0x21, (byte)0x01,
		(byte)0x00, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0xE1, (byte)0xDD, (byte)0xE1, (byte)0xC1, (byte)0x23, (byte)0x13, (byte)0x05, (byte)0x78, (byte)0xFE, (byte)0x00, (byte)0x20, (byte)0xBC,
		(byte)0xFD, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x18, (byte)0x54, (byte)0x06, (byte)0x00, (byte)0xC5, (byte)0x79, (byte)0x80, (byte)0x80, (byte)0x80, (byte)0x4F, (byte)0x21, (byte)0x00,
		(byte)0x00, (byte)0x11, (byte)0x18, (byte)0x00, (byte)0x78, (byte)0x3D, (byte)0xFA, (byte)0xAC, (byte)0x0D, (byte)0x19, (byte)0x18, (byte)0xF9, (byte)0xFD, (byte)0x2A, (byte)0x06, (byte)0x80,
		(byte)0xFD, (byte)0x7E, (byte)0x12, (byte)0x80, (byte)0xFE, (byte)0x18, (byte)0x30, (byte)0x2B, (byte)0xCB, (byte)0x3F, (byte)0xCB, (byte)0x3F, (byte)0xCB, (byte)0x3F, (byte)0x57, (byte)0x59,
		(byte)0xD5, (byte)0x01, (byte)0x1C, (byte)0x00, (byte)0x09, (byte)0xED, (byte)0x4B, (byte)0x06, (byte)0x80, (byte)0x09, (byte)0xE5, (byte)0xFD, (byte)0x21, (byte)0x03, (byte)0x00, (byte)0x3E,
		(byte)0x03, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0xE1, (byte)0x11, (byte)0x68, (byte)0x00, (byte)0x19, (byte)0xD1, (byte)0xFD, (byte)0x21, (byte)0x03, (byte)0x00, (byte)0x3E, (byte)0x04,
		(byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0xC1, (byte)0x04, (byte)0x78, (byte)0xFE, (byte)0x03, (byte)0x20, (byte)0xAE, (byte)0xFD, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0xFD, (byte)0x46,
		(byte)0x06, (byte)0x78, (byte)0xFE, (byte)0x80, (byte)0x28, (byte)0x21, (byte)0xFD, (byte)0x4E, (byte)0x07, (byte)0xFD, (byte)0x66, (byte)0x11, (byte)0xFD, (byte)0x6E, (byte)0x12, (byte)0xB7,
		(byte)0xED, (byte)0x42, (byte)0x28, (byte)0x13, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x11, (byte)0x08, (byte)0x00, (byte)0x19, (byte)0xFD, (byte)0x5E, (byte)0x06, (byte)0xFD, (byte)0x56,
		(byte)0x07, (byte)0x01, (byte)0x03, (byte)0x03, (byte)0xCD, (byte)0x0B, (byte)0x08, (byte)0xFD, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x11, (byte)0x13,
		(byte)0x00, (byte)0x19, (byte)0xFD, (byte)0x5E, (byte)0x11, (byte)0xFD, (byte)0x56, (byte)0x12, (byte)0x01, (byte)0x03, (byte)0x03, (byte)0xCD, (byte)0x0B, (byte)0x08, (byte)0xC9, (byte)0xFD,
		(byte)0xCB, (byte)0x03, (byte)0x46, (byte)0x20, (byte)0x16, (byte)0xDD, (byte)0xB6, (byte)0x00, (byte)0xDD, (byte)0x77, (byte)0x00, (byte)0x7C, (byte)0xDD, (byte)0xB6, (byte)0x08, (byte)0xDD,
		(byte)0x77, (byte)0x08, (byte)0x7D, (byte)0xDD, (byte)0xB6, (byte)0x10, (byte)0xDD, (byte)0x77, (byte)0x10, (byte)0x18, (byte)0x14, (byte)0xB7, (byte)0x28, (byte)0x03, (byte)0xDD, (byte)0x77,
		(byte)0x00, (byte)0x7C, (byte)0xB7, (byte)0x28, (byte)0x03, (byte)0xDD, (byte)0x77, (byte)0x08, (byte)0x7D, (byte)0xB7, (byte)0x28, (byte)0x03, (byte)0xDD, (byte)0x77, (byte)0x10, (byte)0xFD,
		(byte)0xCB, (byte)0x03, (byte)0x7E, (byte)0x28, (byte)0x3C, (byte)0xDD, (byte)0xE5, (byte)0x01, (byte)0x68, (byte)0x00, (byte)0xDD, (byte)0x09, (byte)0xFD, (byte)0x46, (byte)0x02, (byte)0xFD,
		(byte)0xCB, (byte)0x03, (byte)0x4E, (byte)0x20, (byte)0x04, (byte)0x0E, (byte)0x0F, (byte)0x18, (byte)0x02, (byte)0x0E, (byte)0x00, (byte)0x7B, (byte)0xB7, (byte)0x28, (byte)0x08, (byte)0xDD,
		(byte)0x7E, (byte)0x00, (byte)0xA1, (byte)0xB0, (byte)0xDD, (byte)0x77, (byte)0x00, (byte)0x7C, (byte)0xB7, (byte)0x28, (byte)0x08, (byte)0xDD, (byte)0x7E, (byte)0x08, (byte)0xA1, (byte)0xB0,
		(byte)0xDD, (byte)0x77, (byte)0x08, (byte)0x7D, (byte)0xB7, (byte)0x28, (byte)0x08, (byte)0xDD, (byte)0x7E, (byte)0x10, (byte)0xA1, (byte)0xB0, (byte)0xDD, (byte)0x77, (byte)0x10, (byte)0xDD,
		(byte)0xE1, (byte)0xC9, (byte)0xC5, (byte)0xD9, (byte)0xDD, (byte)0x66, (byte)0x03, (byte)0xDD, (byte)0x6E, (byte)0x02, (byte)0x7E, (byte)0x23, (byte)0x4E, (byte)0x23, (byte)0x46, (byte)0x23,
		(byte)0x5E, (byte)0x23, (byte)0x56, (byte)0xD9, (byte)0x87, (byte)0x87, (byte)0x5F, (byte)0x16, (byte)0x00, (byte)0x23, (byte)0x19, (byte)0x4E, (byte)0x23, (byte)0x46, (byte)0x23, (byte)0x5E,
		(byte)0x23, (byte)0x56, (byte)0x60, (byte)0x69, (byte)0xC1, (byte)0x79, (byte)0x48, (byte)0xCB, (byte)0x3F, (byte)0xCB, (byte)0x3F, (byte)0xCB, (byte)0x3F, (byte)0xCB, (byte)0x3F, (byte)0x47,
		(byte)0xC5, (byte)0xDD, (byte)0xE5, (byte)0xE5, (byte)0xD5, (byte)0xDD, (byte)0x6E, (byte)0x04, (byte)0xDD, (byte)0x66, (byte)0x05, (byte)0xDD, (byte)0x23, (byte)0xDD, (byte)0x23, (byte)0x23,
		(byte)0x23, (byte)0x5E, (byte)0x23, (byte)0x56, (byte)0xD5, (byte)0xFD, (byte)0xE1, (byte)0xD1, (byte)0xE1, (byte)0x7E, (byte)0xFD, (byte)0xCB, (byte)0x00, (byte)0x7E, (byte)0x28, (byte)0x02,
		(byte)0xCB, (byte)0xFF, (byte)0xFD, (byte)0x77, (byte)0x00, (byte)0x23, (byte)0x1A, (byte)0xD9, (byte)0x6F, (byte)0x26, (byte)0x00, (byte)0x09, (byte)0xFD, (byte)0x75, (byte)0x01, (byte)0xFD,
		(byte)0x74, (byte)0x02, (byte)0xD9, (byte)0x13, (byte)0x1A, (byte)0xD9, (byte)0x6F, (byte)0x26, (byte)0x00, (byte)0x19, (byte)0xFD, (byte)0x75, (byte)0x03, (byte)0xFD, (byte)0x74, (byte)0x04,
		(byte)0xD9, (byte)0x13, (byte)0x10, (byte)0xBF, (byte)0xFD, (byte)0xE1, (byte)0x01, (byte)0x04, (byte)0x00, (byte)0xFD, (byte)0x09, (byte)0xD1, (byte)0xFD, (byte)0x6E, (byte)0x00, (byte)0xFD,
		(byte)0x66, (byte)0x01, (byte)0xFD, (byte)0x23, (byte)0xFD, (byte)0x23, (byte)0xE5, (byte)0xDD, (byte)0xE1, (byte)0xFD, (byte)0xE5, (byte)0xD5, (byte)0x43, (byte)0xCD, (byte)0xFA, (byte)0x1F,
		(byte)0xD1, (byte)0xFD, (byte)0xE1, (byte)0x15, (byte)0x20, (byte)0xE6, (byte)0xC9, (byte)0x2A, (byte)0xD3, (byte)0x73, (byte)0xCB, (byte)0x6E, (byte)0xCC, (byte)0x49, (byte)0x0F, (byte)0xCB,
		(byte)0x66, (byte)0x20, (byte)0x05, (byte)0x23, (byte)0x23, (byte)0x23, (byte)0x18, (byte)0xF2, (byte)0xC9, (byte)0xE5, (byte)0xCB, (byte)0x5E, (byte)0x28, (byte)0x2B, (byte)0xCB, (byte)0x76,
		(byte)0x20, (byte)0x0D, (byte)0x23, (byte)0x5E, (byte)0x23, (byte)0x56, (byte)0x1B, (byte)0x7B, (byte)0xB2, (byte)0x20, (byte)0x34, (byte)0xE1, (byte)0xE5, (byte)0x18, (byte)0x2C, (byte)0x23,
		(byte)0x5E, (byte)0x23, (byte)0x56, (byte)0xEB, (byte)0x5E, (byte)0x23, (byte)0x56, (byte)0x1B, (byte)0x7B, (byte)0xB2, (byte)0x20, (byte)0x23, (byte)0x23, (byte)0x5E, (byte)0x23, (byte)0x56,
		(byte)0x2B, (byte)0x2B, (byte)0x72, (byte)0x2B, (byte)0x73, (byte)0xE1, (byte)0xE5, (byte)0x18, (byte)0x12, (byte)0x23, (byte)0x35, (byte)0x20, (byte)0x10, (byte)0xE1, (byte)0xE5, (byte)0xCB,
		(byte)0x76, (byte)0x28, (byte)0x08, (byte)0x23, (byte)0x23, (byte)0x7E, (byte)0x2B, (byte)0x77, (byte)0x2B, (byte)0xE1, (byte)0xE5, (byte)0xCB, (byte)0xFE, (byte)0xE1, (byte)0xC9, (byte)0x72,
		(byte)0x2B, (byte)0x73, (byte)0x18, (byte)0xF9, (byte)0x02, (byte)0x00, (byte)0x02, (byte)0x00, (byte)0x02, (byte)0x00, (byte)0x01, (byte)0x94, (byte)0x0F, (byte)0x11, (byte)0xBA, (byte)0x73,
		(byte)0xCD, (byte)0x98, (byte)0x00, (byte)0x2A, (byte)0xBA, (byte)0x73, (byte)0xED, (byte)0x5B, (byte)0xBC, (byte)0x73, (byte)0x22, (byte)0xD3, (byte)0x73, (byte)0x36, (byte)0x30, (byte)0xEB,
		(byte)0x22, (byte)0xD5, (byte)0x73, (byte)0xC9, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0xB4, (byte)0x0F, (byte)0x11, (byte)0xBE, (byte)0x73, (byte)0xCD, (byte)0x98,
		(byte)0x00, (byte)0x3A, (byte)0xBE, (byte)0x73, (byte)0x4F, (byte)0x2A, (byte)0xD3, (byte)0x73, (byte)0x47, (byte)0x11, (byte)0x03, (byte)0x00, (byte)0xB7, (byte)0x28, (byte)0x08, (byte)0xCB,
		(byte)0x66, (byte)0x20, (byte)0x6A, (byte)0x19, (byte)0x0D, (byte)0x20, (byte)0xF8, (byte)0xCB, (byte)0x6E, (byte)0x20, (byte)0x62, (byte)0xCB, (byte)0xEE, (byte)0xCB, (byte)0x76, (byte)0x28,
		(byte)0x5C, (byte)0xCB, (byte)0x5E, (byte)0x28, (byte)0x58, (byte)0x23, (byte)0x5E, (byte)0x23, (byte)0x56, (byte)0xD5, (byte)0x2A, (byte)0xD3, (byte)0x73, (byte)0xE5, (byte)0xCB, (byte)0x66,
		(byte)0x20, (byte)0x2E, (byte)0xCB, (byte)0x6E, (byte)0x20, (byte)0x23, (byte)0x7E, (byte)0xE6, (byte)0x48, (byte)0xFE, (byte)0x48, (byte)0x20, (byte)0x1C, (byte)0x23, (byte)0x23, (byte)0x7E,
		(byte)0xBA, (byte)0x38, (byte)0x16, (byte)0x20, (byte)0x08, (byte)0x2B, (byte)0x7E, (byte)0xBB, (byte)0x38, (byte)0x0F, (byte)0x28, (byte)0x31, (byte)0x23, (byte)0x56, (byte)0x2B, (byte)0x5E,
		(byte)0x1B, (byte)0x1B, (byte)0x1B, (byte)0x1B, (byte)0x73, (byte)0x23, (byte)0x72, (byte)0x18, (byte)0x00, (byte)0xE1, (byte)0x23, (byte)0x23, (byte)0x23, (byte)0xE5, (byte)0x18, (byte)0xCE,
		(byte)0x06, (byte)0x00, (byte)0xB7, (byte)0xE1, (byte)0xD1, (byte)0xE5, (byte)0x2A, (byte)0xD5, (byte)0x73, (byte)0xED, (byte)0x52, (byte)0x4D, (byte)0x6B, (byte)0x62, (byte)0x23, (byte)0x23,
		(byte)0x23, (byte)0x23, (byte)0xED, (byte)0xB0, (byte)0x01, (byte)0x08, (byte)0x00, (byte)0xED, (byte)0x42, (byte)0x22, (byte)0xD5, (byte)0x73, (byte)0xE1, (byte)0xC9, (byte)0x02, (byte)0x00,
		(byte)0x01, (byte)0x00, (byte)0x02, (byte)0x00, (byte)0x01, (byte)0x3E, (byte)0x10, (byte)0x11, (byte)0xBF, (byte)0x73, (byte)0xCD, (byte)0x98, (byte)0x00, (byte)0x2A, (byte)0xC0, (byte)0x73,
		(byte)0x3A, (byte)0xBF, (byte)0x73, (byte)0x4F, (byte)0xEB, (byte)0x2A, (byte)0xD3, (byte)0x73, (byte)0xAF, (byte)0x47, (byte)0xCB, (byte)0x6E, (byte)0x28, (byte)0x3E, (byte)0xE5, (byte)0x7E,
		(byte)0xE6, (byte)0x10, (byte)0xF6, (byte)0x20, (byte)0x77, (byte)0xAF, (byte)0xB2, (byte)0x20, (byte)0x0B, (byte)0xB1, (byte)0x28, (byte)0x02, (byte)0xCB, (byte)0xF6, (byte)0x23, (byte)0x73,
		(byte)0x23, (byte)0x73, (byte)0x18, (byte)0x42, (byte)0xCB, (byte)0xDE, (byte)0x79, (byte)0xB7, (byte)0x28, (byte)0x1B, (byte)0xD5, (byte)0xEB, (byte)0x2A, (byte)0xD5, (byte)0x73, (byte)0xEB,
		(byte)0xCB, (byte)0xF6, (byte)0x23, (byte)0x73, (byte)0x23, (byte)0x72, (byte)0xEB, (byte)0xD1, (byte)0x73, (byte)0x23, (byte)0x72, (byte)0x23, (byte)0x73, (byte)0x23, (byte)0x72, (byte)0x23,
		(byte)0x22, (byte)0xD5, (byte)0x73, (byte)0x18, (byte)0x21, (byte)0x23, (byte)0x73, (byte)0x23, (byte)0x72, (byte)0x23, (byte)0x18, (byte)0x1A, (byte)0xCB, (byte)0x66, (byte)0x20, (byte)0x06,
		(byte)0x23, (byte)0x23, (byte)0x23, (byte)0x04, (byte)0x18, (byte)0xB4, (byte)0xD5, (byte)0xE5, (byte)0x23, (byte)0x23, (byte)0x23, (byte)0x04, (byte)0x36, (byte)0x30, (byte)0xEB, (byte)0xE1,
		(byte)0xCB, (byte)0xA6, (byte)0xEB, (byte)0xD1, (byte)0x18, (byte)0xA4, (byte)0xE1, (byte)0xCB, (byte)0xAE, (byte)0x78, (byte)0xC9, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0x00, (byte)0x01,
		(byte)0xBB, (byte)0x10, (byte)0x11, (byte)0xC2, (byte)0x73, (byte)0xCD, (byte)0x98, (byte)0x00, (byte)0x3A, (byte)0xC2, (byte)0x73, (byte)0x4F, (byte)0x2A, (byte)0xD3, (byte)0x73, (byte)0x47,
		(byte)0x11, (byte)0x03, (byte)0x00, (byte)0xB7, (byte)0x28, (byte)0x08, (byte)0xCB, (byte)0x66, (byte)0x20, (byte)0x0C, (byte)0x19, (byte)0x0D, (byte)0x20, (byte)0xF8, (byte)0xCB, (byte)0x6E,
		(byte)0x20, (byte)0x04, (byte)0xCB, (byte)0x7E, (byte)0x20, (byte)0x03, (byte)0xAF, (byte)0x18, (byte)0x0A, (byte)0xCB, (byte)0x76, (byte)0x20, (byte)0x02, (byte)0xCB, (byte)0xEE, (byte)0xCB,
		(byte)0xBE, (byte)0x3E, (byte)0x01, (byte)0xB7, (byte)0xC9, (byte)0x0F, (byte)0x06, (byte)0x01, (byte)0x03, (byte)0x09, (byte)0x00, (byte)0x0A, (byte)0x0F, (byte)0x02, (byte)0x0B, (byte)0x07,
		(byte)0x0F, (byte)0x05, (byte)0x04, (byte)0x08, (byte)0x0F, (byte)0xD3, (byte)0xC0, (byte)0xAF, (byte)0xDD, (byte)0x2A, (byte)0x08, (byte)0x80, (byte)0xDD, (byte)0x23, (byte)0xDD, (byte)0x23,
		(byte)0xFD, (byte)0x21, (byte)0xD7, (byte)0x73, (byte)0x06, (byte)0x0A, (byte)0xDD, (byte)0x77, (byte)0x00, (byte)0xDD, (byte)0x23, (byte)0xFD, (byte)0x77, (byte)0x00, (byte)0xFD, (byte)0x23,
		(byte)0xFD, (byte)0x77, (byte)0x00, (byte)0xFD, (byte)0x23, (byte)0x05, (byte)0x20, (byte)0xEE, (byte)0x32, (byte)0xEB, (byte)0x73, (byte)0x32, (byte)0xEC, (byte)0x73, (byte)0x32, (byte)0xEE,
		(byte)0x73, (byte)0x32, (byte)0xEF, (byte)0x73, (byte)0x32, (byte)0xF0, (byte)0x73, (byte)0x32, (byte)0xF1, (byte)0x73, (byte)0xC9, (byte)0x00, (byte)0xC9, (byte)0x7C, (byte)0xFE, (byte)0x00,
		(byte)0x20, (byte)0x04, (byte)0xDB, (byte)0xFC, (byte)0x18, (byte)0x02, (byte)0xDB, (byte)0xFF, (byte)0x2F, (byte)0xC9, (byte)0xDB, (byte)0xFC, (byte)0x2F, (byte)0x32, (byte)0xEE, (byte)0x73,
		(byte)0xDB, (byte)0xFF, (byte)0x2F, (byte)0x32, (byte)0xEF, (byte)0x73, (byte)0xD3, (byte)0x80, (byte)0xCD, (byte)0x3B, (byte)0x11, (byte)0xDB, (byte)0xFC, (byte)0x2F, (byte)0x32, (byte)0xF0,
		(byte)0x73, (byte)0xDB, (byte)0xFF, (byte)0x2F, (byte)0x32, (byte)0xF1, (byte)0x73, (byte)0xD3, (byte)0xC0, (byte)0xC9, (byte)0xDB, (byte)0xFC, (byte)0x21, (byte)0xEB, (byte)0x73, (byte)0xCB,
		(byte)0x67, (byte)0x20, (byte)0x08, (byte)0xCB, (byte)0x6F, (byte)0x20, (byte)0x03, (byte)0x35, (byte)0x18, (byte)0x01, (byte)0x34, (byte)0xDB, (byte)0xFF, (byte)0xCB, (byte)0x67, (byte)0x20,
		(byte)0x09, (byte)0x23, (byte)0xCB, (byte)0x6F, (byte)0x20, (byte)0x03, (byte)0x35, (byte)0x18, (byte)0x01, (byte)0x34, (byte)0xC9, (byte)0x7D, (byte)0xFE, (byte)0x01, (byte)0x28, (byte)0x1A,
		(byte)0x01, (byte)0xEB, (byte)0x73, (byte)0x7C, (byte)0xFE, (byte)0x00, (byte)0x28, (byte)0x01, (byte)0x03, (byte)0x0A, (byte)0x5F, (byte)0xAF, (byte)0x02, (byte)0xCD, (byte)0x3D, (byte)0x11,
		(byte)0x57, (byte)0xE6, (byte)0x0F, (byte)0x6F, (byte)0x7A, (byte)0xE6, (byte)0x40, (byte)0x67, (byte)0x18, (byte)0x16, (byte)0xD3, (byte)0x80, (byte)0xCD, (byte)0x3D, (byte)0x11, (byte)0x57,
		(byte)0xD3, (byte)0xC0, (byte)0xE6, (byte)0x0F, (byte)0x21, (byte)0xF5, (byte)0x10, (byte)0x06, (byte)0x00, (byte)0x4F, (byte)0x09, (byte)0x6E, (byte)0x7A, (byte)0xE6, (byte)0x40, (byte)0x67,
		(byte)0xC9, (byte)0xCD, (byte)0x4A, (byte)0x11, (byte)0xFD, (byte)0x21, (byte)0xD7, (byte)0x73, (byte)0xDD, (byte)0x2A, (byte)0x08, (byte)0x80, (byte)0xDD, (byte)0xE5, (byte)0xDD, (byte)0x7E,
		(byte)0x00, (byte)0xCB, (byte)0x7F, (byte)0x28, (byte)0x1E, (byte)0x47, (byte)0x11, (byte)0x02, (byte)0x00, (byte)0xDD, (byte)0x19, (byte)0xE6, (byte)0x07, (byte)0x28, (byte)0x09, (byte)0x3A,
		(byte)0xEE, (byte)0x73, (byte)0x21, (byte)0xEB, (byte)0x73, (byte)0xCD, (byte)0x20, (byte)0x12, (byte)0x78, (byte)0xE6, (byte)0x18, (byte)0x28, (byte)0x06, (byte)0x3A, (byte)0xF0, (byte)0x73,
		(byte)0xCD, (byte)0x3F, (byte)0x12, (byte)0xDD, (byte)0xE1, (byte)0xDD, (byte)0x7E, (byte)0x01, (byte)0xCB, (byte)0x7F, (byte)0x28, (byte)0x23, (byte)0x47, (byte)0x11, (byte)0x0A, (byte)0x00,
		(byte)0xFD, (byte)0x19, (byte)0x11, (byte)0x07, (byte)0x00, (byte)0xDD, (byte)0x19, (byte)0xE6, (byte)0x07, (byte)0x28, (byte)0x09, (byte)0x3A, (byte)0xEF, (byte)0x73, (byte)0x21, (byte)0xEC,
		(byte)0x73, (byte)0xCD, (byte)0x20, (byte)0x12, (byte)0x78, (byte)0xE6, (byte)0x18, (byte)0x28, (byte)0x06, (byte)0x3A, (byte)0xF1, (byte)0x73, (byte)0xCD, (byte)0x3F, (byte)0x12, (byte)0xC9,
		(byte)0x4F, (byte)0xCB, (byte)0x48, (byte)0x28, (byte)0x04, (byte)0xCD, (byte)0xB9, (byte)0x12, (byte)0x79, (byte)0xCB, (byte)0x40, (byte)0x28, (byte)0x04, (byte)0xCD, (byte)0x89, (byte)0x12,
		(byte)0x79, (byte)0xCB, (byte)0x50, (byte)0x28, (byte)0x09, (byte)0x7E, (byte)0xDD, (byte)0x86, (byte)0x02, (byte)0xDD, (byte)0x77, (byte)0x02, (byte)0xAF, (byte)0x77, (byte)0xC9, (byte)0x4F,
		(byte)0xCB, (byte)0x58, (byte)0x28, (byte)0x04, (byte)0xCD, (byte)0xE9, (byte)0x12, (byte)0x79, (byte)0xCB, (byte)0x60, (byte)0x28, (byte)0x03, (byte)0xCD, (byte)0x50, (byte)0x12, (byte)0xC9,
		(byte)0xC5, (byte)0xD5, (byte)0xE5, (byte)0xE6, (byte)0x0F, (byte)0x5F, (byte)0xFD, (byte)0x46, (byte)0x08, (byte)0xFD, (byte)0x7E, (byte)0x09, (byte)0xFE, (byte)0x00, (byte)0x20, (byte)0x1A,
		(byte)0x7B, (byte)0xB8, (byte)0x28, (byte)0x05, (byte)0xFD, (byte)0x73, (byte)0x08, (byte)0x18, (byte)0x1C, (byte)0x3E, (byte)0x01, (byte)0xFD, (byte)0x77, (byte)0x09, (byte)0x21, (byte)0xF5,
		(byte)0x10, (byte)0x16, (byte)0x00, (byte)0x19, (byte)0x7E, (byte)0xDD, (byte)0x77, (byte)0x04, (byte)0x18, (byte)0x0B, (byte)0x7B, (byte)0xB8, (byte)0x28, (byte)0x07, (byte)0xFD, (byte)0x73,
		(byte)0x08, (byte)0xAF, (byte)0xFD, (byte)0x77, (byte)0x09, (byte)0xE1, (byte)0xD1, (byte)0xC1, (byte)0xC9, (byte)0xC5, (byte)0xD5, (byte)0xE6, (byte)0x40, (byte)0x5F, (byte)0xFD, (byte)0x46,
		(byte)0x00, (byte)0xFD, (byte)0x7E, (byte)0x01, (byte)0xFE, (byte)0x00, (byte)0x20, (byte)0x13, (byte)0x7B, (byte)0xB8, (byte)0x28, (byte)0x05, (byte)0xFD, (byte)0x73, (byte)0x00, (byte)0x18,
		(byte)0x15, (byte)0x3E, (byte)0x01, (byte)0xFD, (byte)0x77, (byte)0x01, (byte)0xDD, (byte)0x73, (byte)0x00, (byte)0x18, (byte)0x0B, (byte)0x7B, (byte)0xB8, (byte)0x28, (byte)0x07, (byte)0xFD,
		(byte)0x73, (byte)0x00, (byte)0xAF, (byte)0xFD, (byte)0x77, (byte)0x01, (byte)0xD1, (byte)0xC1, (byte)0xC9, (byte)0xC5, (byte)0xD5, (byte)0xE6, (byte)0x0F, (byte)0x5F, (byte)0xFD, (byte)0x46,
		(byte)0x02, (byte)0xFD, (byte)0x7E, (byte)0x03, (byte)0xFE, (byte)0x00, (byte)0x20, (byte)0x13, (byte)0x7B, (byte)0xB8, (byte)0x28, (byte)0x05, (byte)0xFD, (byte)0x73, (byte)0x02, (byte)0x18,
		(byte)0x15, (byte)0x3E, (byte)0x01, (byte)0xFD, (byte)0x77, (byte)0x03, (byte)0xDD, (byte)0x73, (byte)0x01, (byte)0x18, (byte)0x0B, (byte)0x7B, (byte)0xB8, (byte)0x28, (byte)0x07, (byte)0xFD,
		(byte)0x73, (byte)0x02, (byte)0xAF, (byte)0xFD, (byte)0x77, (byte)0x03, (byte)0xD1, (byte)0xC1, (byte)0xC9, (byte)0xC5, (byte)0xD5, (byte)0xE6, (byte)0x40, (byte)0x5F, (byte)0xFD, (byte)0x46,
		(byte)0x06, (byte)0xFD, (byte)0x7E, (byte)0x07, (byte)0xFE, (byte)0x00, (byte)0x20, (byte)0x13, (byte)0x7B, (byte)0xB8, (byte)0x28, (byte)0x05, (byte)0xFD, (byte)0x73, (byte)0x06, (byte)0x18,
		(byte)0x15, (byte)0x3E, (byte)0x01, (byte)0xFD, (byte)0x77, (byte)0x07, (byte)0xDD, (byte)0x73, (byte)0x03, (byte)0x18, (byte)0x0B, (byte)0x7B, (byte)0xB8, (byte)0x28, (byte)0x07, (byte)0xFD,
		(byte)0x73, (byte)0x06, (byte)0xAF, (byte)0xFD, (byte)0x77, (byte)0x07, (byte)0xD1, (byte)0xC1, (byte)0xC9, (byte)0x21, (byte)0x00, (byte)0x00, (byte)0x11, (byte)0x00, (byte)0x40, (byte)0x3E,
		(byte)0x00, (byte)0xCD, (byte)0xD4, (byte)0x18, (byte)0xCD, (byte)0xE9, (byte)0x18, (byte)0xCD, (byte)0x27, (byte)0x19, (byte)0x21, (byte)0xA3, (byte)0x18, (byte)0x11, (byte)0x60, (byte)0x00,
		(byte)0xE5, (byte)0xD5, (byte)0x7E, (byte)0xFE, (byte)0xFF, (byte)0x28, (byte)0x1B, (byte)0x47, (byte)0x04, (byte)0x21, (byte)0xC3, (byte)0x14, (byte)0x11, (byte)0x08, (byte)0x00, (byte)0x10,
		(byte)0x15, (byte)0xD1, (byte)0xD5, (byte)0xFD, (byte)0x21, (byte)0x01, (byte)0x00, (byte)0x3E, (byte)0x03, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0xD1, (byte)0xE1, (byte)0x13, (byte)0x23,
		(byte)0x18, (byte)0xDE, (byte)0xD1, (byte)0xE1, (byte)0x18, (byte)0x03, (byte)0x19, (byte)0x18, (byte)0xE6, (byte)0x21, (byte)0x4D, (byte)0x14, (byte)0x11, (byte)0x85, (byte)0x00, (byte)0xFD,
		(byte)0x21, (byte)0x16, (byte)0x00, (byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0x21, (byte)0x63, (byte)0x14, (byte)0x11, (byte)0xA5, (byte)0x00, (byte)0xFD, (byte)0x21,
		(byte)0x16, (byte)0x00, (byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0x21, (byte)0xC1, (byte)0x14, (byte)0x11, (byte)0x9B, (byte)0x00, (byte)0xFD, (byte)0x21, (byte)0x02,
		(byte)0x00, (byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0x21, (byte)0xB4, (byte)0x14, (byte)0x11, (byte)0xAA, (byte)0x02, (byte)0xFD, (byte)0x21, (byte)0x0D, (byte)0x00,
		(byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0x21, (byte)0x3B, (byte)0x14, (byte)0x11, (byte)0x00, (byte)0x00, (byte)0x3E, (byte)0x04, (byte)0xFD, (byte)0x21, (byte)0x12,
		(byte)0x00, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0x06, (byte)0x01, (byte)0x0E, (byte)0xC0, (byte)0xCD, (byte)0xD9, (byte)0x1F, (byte)0x21, (byte)0x00, (byte)0x80, (byte)0x7E, (byte)0xFE,
		(byte)0xAA, (byte)0x20, (byte)0x4C, (byte)0x23, (byte)0x7E, (byte)0xFE, (byte)0x55, (byte)0x20, (byte)0x46, (byte)0x21, (byte)0x24, (byte)0x80, (byte)0xCD, (byte)0x46, (byte)0x19, (byte)0x11,
		(byte)0x24, (byte)0x80, (byte)0x21, (byte)0x01, (byte)0x02, (byte)0xCD, (byte)0x51, (byte)0x19, (byte)0x21, (byte)0x24, (byte)0x80, (byte)0xCD, (byte)0x46, (byte)0x19, (byte)0x23, (byte)0x54,
		(byte)0x5D, (byte)0xCD, (byte)0x46, (byte)0x19, (byte)0x21, (byte)0xC1, (byte)0x01, (byte)0xCD, (byte)0x51, (byte)0x19, (byte)0x21, (byte)0x24, (byte)0x80, (byte)0xCD, (byte)0x46, (byte)0x19,
		(byte)0x23, (byte)0xCD, (byte)0x46, (byte)0x19, (byte)0x23, (byte)0x11, (byte)0xAC, (byte)0x02, (byte)0xFD, (byte)0x21, (byte)0x04, (byte)0x00, (byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0xBE,
		(byte)0x1F, (byte)0xCD, (byte)0x68, (byte)0x19, (byte)0x06, (byte)0x01, (byte)0x0E, (byte)0x80, (byte)0xCD, (byte)0xD9, (byte)0x1F, (byte)0x2A, (byte)0x0A, (byte)0x80, (byte)0xE9, (byte)0x21,
		(byte)0x79, (byte)0x14, (byte)0x11, (byte)0xAA, (byte)0x01, (byte)0xFD, (byte)0x21, (byte)0x0D, (byte)0x00, (byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0x21, (byte)0x86,
		(byte)0x14, (byte)0x11, (byte)0xE4, (byte)0x01, (byte)0xFD, (byte)0x21, (byte)0x1A, (byte)0x00, (byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0x21, (byte)0xA0, (byte)0x14,
		(byte)0x11, (byte)0x27, (byte)0x02, (byte)0xFD, (byte)0x21, (byte)0x14, (byte)0x00, (byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0x21, (byte)0x00, (byte)0x8A, (byte)0xCD,
		(byte)0x6B, (byte)0x19, (byte)0x06, (byte)0x01, (byte)0x0E, (byte)0x80, (byte)0xCD, (byte)0xD9, (byte)0x1F, (byte)0x18, (byte)0xFE, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF0, (byte)0xF0,
		(byte)0xF0, (byte)0xF0, (byte)0xF0, (byte)0xF0, (byte)0xF0, (byte)0xF0, (byte)0xF0, (byte)0xD0, (byte)0x80, (byte)0x90, (byte)0xB0, (byte)0x30, (byte)0x40, (byte)0x60, (byte)0x61, (byte)0x68,
		(byte)0x69, (byte)0x70, (byte)0x71, (byte)0x78, (byte)0x79, (byte)0x80, (byte)0x81, (byte)0x88, (byte)0x89, (byte)0x64, (byte)0x65, (byte)0x6C, (byte)0x74, (byte)0x75, (byte)0x7C, (byte)0x84,
		(byte)0x85, (byte)0x8C, (byte)0x8D, (byte)0x62, (byte)0x63, (byte)0x6A, (byte)0x6B, (byte)0x72, (byte)0x73, (byte)0x7A, (byte)0x7B, (byte)0x82, (byte)0x83, (byte)0x8A, (byte)0x8B, (byte)0x66,
		(byte)0x67, (byte)0x6D, (byte)0x76, (byte)0x77, (byte)0x7D, (byte)0x86, (byte)0x87, (byte)0x8E, (byte)0x8F, (byte)0x54, (byte)0x55, (byte)0x52, (byte)0x4E, (byte)0x20, (byte)0x47, (byte)0x41,
		(byte)0x4D, (byte)0x45, (byte)0x20, (byte)0x4F, (byte)0x46, (byte)0x46, (byte)0x42, (byte)0x45, (byte)0x46, (byte)0x4F, (byte)0x52, (byte)0x45, (byte)0x20, (byte)0x49, (byte)0x4E, (byte)0x53,
		(byte)0x45, (byte)0x52, (byte)0x54, (byte)0x49, (byte)0x4E, (byte)0x47, (byte)0x20, (byte)0x43, (byte)0x41, (byte)0x52, (byte)0x54, (byte)0x52, (byte)0x49, (byte)0x44, (byte)0x47, (byte)0x45,
		(byte)0x4F, (byte)0x52, (byte)0x20, (byte)0x45, (byte)0x58, (byte)0x50, (byte)0x41, (byte)0x4E, (byte)0x53, (byte)0x49, (byte)0x4F, (byte)0x4E, (byte)0x20, (byte)0x4D, (byte)0x4F, (byte)0x44,
		(byte)0x55, (byte)0x4C, (byte)0x45, (byte)0x2E, (byte)0x1D, (byte)0x20, (byte)0x31, (byte)0x39, (byte)0x38, (byte)0x32, (byte)0x20, (byte)0x43, (byte)0x4F, (byte)0x4C, (byte)0x45, (byte)0x43,
		(byte)0x4F, (byte)0x1E, (byte)0x1F, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x3F, (byte)0x7F, (byte)0xFF, (byte)0xFF, (byte)0xF3,
		(byte)0xF3, (byte)0xF0, (byte)0xF0, (byte)0x00, (byte)0x80, (byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0x00, (byte)0x00, (byte)0x3F, (byte)0x7F, (byte)0xFF, (byte)0xFF, (byte)0xF3,
		(byte)0xF3, (byte)0xF3, (byte)0xF3, (byte)0x00, (byte)0x80, (byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0xF0, (byte)0xF0, (byte)0xF0, (byte)0xF0, (byte)0xF0,
		(byte)0xF0, (byte)0xF0, (byte)0xF0, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xF0, (byte)0xF0, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0x00, (byte)0x00,
		(byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF1, (byte)0xF1, (byte)0xF1, (byte)0x7B, (byte)0x7B, (byte)0x7B, (byte)0x3F, (byte)0x3F, (byte)0xE0, (byte)0xE0, (byte)0xE0, (byte)0xC0, (byte)0xC0,
		(byte)0xC0, (byte)0x80, (byte)0x80, (byte)0x1F, (byte)0x3F, (byte)0x7F, (byte)0x79, (byte)0x78, (byte)0x7F, (byte)0x7F, (byte)0x3F, (byte)0x80, (byte)0xC0, (byte)0xE0, (byte)0xE0, (byte)0x00,
		(byte)0x80, (byte)0xC0, (byte)0xE0, (byte)0xF3, (byte)0xF3, (byte)0xFB, (byte)0xFB, (byte)0xFB, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0xC0,
		(byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0xF3, (byte)0xF3, (byte)0xFF, (byte)0xFF, (byte)0x7F, (byte)0x3F, (byte)0x00, (byte)0x00, (byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0x80,
		(byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF0, (byte)0xF0, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0xFF, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xC0, (byte)0xC0, (byte)0xC0,
		(byte)0xC0, (byte)0x00, (byte)0x00, (byte)0x3F, (byte)0x1F, (byte)0x1F, (byte)0x1F, (byte)0x0E, (byte)0x0E, (byte)0x00, (byte)0x00, (byte)0x80, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,
		(byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF0, (byte)0xF0, (byte)0xF0, (byte)0xF0, (byte)0xF0, (byte)0xF0, (byte)0x00, (byte)0x00, (byte)0x1F, (byte)0x01, (byte)0x79, (byte)0x7F, (byte)0x3F,
		(byte)0x1F, (byte)0x00, (byte)0x00, (byte)0xE0, (byte)0xE0, (byte)0xE0, (byte)0xE0, (byte)0xC0, (byte)0x80, (byte)0x00, (byte)0x00, (byte)0xFF, (byte)0xF7, (byte)0xF7, (byte)0xF7, (byte)0xF3,
		(byte)0xF3, (byte)0x00, (byte)0x00, (byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0x00, (byte)0x00, (byte)0x7E, (byte)0x81, (byte)0xBD, (byte)0xA1, (byte)0xA1,
		(byte)0xBD, (byte)0x81, (byte)0x7E, (byte)0x1F, (byte)0x04, (byte)0x04, (byte)0x04, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x44, (byte)0x6C, (byte)0x54, (byte)0x54, (byte)0x00,
		(byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x20, (byte)0x20, (byte)0x20, (byte)0x20, (byte)0x20,
		(byte)0x00, (byte)0x20, (byte)0x00, (byte)0x50, (byte)0x50, (byte)0x50, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x50, (byte)0x50, (byte)0xF8, (byte)0x50, (byte)0xF8,
		(byte)0x50, (byte)0x50, (byte)0x00, (byte)0x20, (byte)0x78, (byte)0xA0, (byte)0x70, (byte)0x28, (byte)0xF0, (byte)0x20, (byte)0x00, (byte)0xC0, (byte)0xC8, (byte)0x10, (byte)0x20, (byte)0x40,
		(byte)0x98, (byte)0x18, (byte)0x00, (byte)0x40, (byte)0xA0, (byte)0xA0, (byte)0x40, (byte)0xA8, (byte)0x90, (byte)0x68, (byte)0x00, (byte)0x20, (byte)0x20, (byte)0x20, (byte)0x00, (byte)0x00,
		(byte)0x00, (byte)0x00, (byte)0x00, (byte)0x20, (byte)0x40, (byte)0x80, (byte)0x80, (byte)0x80, (byte)0x40, (byte)0x20, (byte)0x00, (byte)0x20, (byte)0x10, (byte)0x08, (byte)0x08, (byte)0x08,
		(byte)0x10, (byte)0x20, (byte)0x00, (byte)0x20, (byte)0xA8, (byte)0x70, (byte)0x20, (byte)0x70, (byte)0xA8, (byte)0x20, (byte)0x00, (byte)0x00, (byte)0x20, (byte)0x20, (byte)0xF8, (byte)0x20,
		(byte)0x20, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x20, (byte)0x20, (byte)0x40, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF8, (byte)0x00,
		(byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x20, (byte)0x00, (byte)0x00, (byte)0x08, (byte)0x10, (byte)0x20, (byte)0x40,
		(byte)0x80, (byte)0x00, (byte)0x00, (byte)0x70, (byte)0x88, (byte)0x98, (byte)0xA8, (byte)0xC8, (byte)0x88, (byte)0x70, (byte)0x00, (byte)0x20, (byte)0x60, (byte)0x20, (byte)0x20, (byte)0x20,
		(byte)0x20, (byte)0x70, (byte)0x00, (byte)0x70, (byte)0x88, (byte)0x08, (byte)0x30, (byte)0x40, (byte)0x80, (byte)0xF8, (byte)0x00, (byte)0xF8, (byte)0x08, (byte)0x10, (byte)0x30, (byte)0x08,
		(byte)0x88, (byte)0x70, (byte)0x00, (byte)0x10, (byte)0x30, (byte)0x50, (byte)0x90, (byte)0xF8, (byte)0x10, (byte)0x10, (byte)0x00, (byte)0xF8, (byte)0x80, (byte)0xF0, (byte)0x08, (byte)0x08,
		(byte)0x88, (byte)0x70, (byte)0x00, (byte)0x38, (byte)0x40, (byte)0x80, (byte)0xF0, (byte)0x88, (byte)0x88, (byte)0x70, (byte)0x00, (byte)0xF8, (byte)0x08, (byte)0x10, (byte)0x20, (byte)0x40,
		(byte)0x40, (byte)0x40, (byte)0x00, (byte)0x70, (byte)0x88, (byte)0x88, (byte)0x70, (byte)0x88, (byte)0x88, (byte)0x70, (byte)0x00, (byte)0x70, (byte)0x88, (byte)0x88, (byte)0x78, (byte)0x08,
		(byte)0x10, (byte)0xE0, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x20, (byte)0x00, (byte)0x20, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x20, (byte)0x00, (byte)0x20,
		(byte)0x20, (byte)0x40, (byte)0x00, (byte)0x10, (byte)0x20, (byte)0x40, (byte)0x80, (byte)0x40, (byte)0x20, (byte)0x10, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF8, (byte)0x00, (byte)0xF8,
		(byte)0x00, (byte)0x00, (byte)0x00, (byte)0x40, (byte)0x20, (byte)0x10, (byte)0x08, (byte)0x10, (byte)0x20, (byte)0x40, (byte)0x00, (byte)0x70, (byte)0x88, (byte)0x10, (byte)0x20, (byte)0x20,
		(byte)0x00, (byte)0x20, (byte)0x00, (byte)0x70, (byte)0x88, (byte)0xA8, (byte)0xB8, (byte)0xB0, (byte)0x80, (byte)0x78, (byte)0x00, (byte)0x20, (byte)0x50, (byte)0x88, (byte)0x88, (byte)0xF8,
		(byte)0x88, (byte)0x88, (byte)0x00, (byte)0xF0, (byte)0x88, (byte)0x88, (byte)0xF0, (byte)0x88, (byte)0x88, (byte)0xF0, (byte)0x00, (byte)0x70, (byte)0x88, (byte)0x80, (byte)0x80, (byte)0x80,
		(byte)0x88, (byte)0x70, (byte)0x00, (byte)0xF0, (byte)0x88, (byte)0x88, (byte)0x88, (byte)0x88, (byte)0x88, (byte)0xF0, (byte)0x00, (byte)0xF8, (byte)0x80, (byte)0x80, (byte)0xF0, (byte)0x80,
		(byte)0x80, (byte)0xF8, (byte)0x00, (byte)0xF8, (byte)0x80, (byte)0x80, (byte)0xF0, (byte)0x80, (byte)0x80, (byte)0x80, (byte)0x00, (byte)0x78, (byte)0x80, (byte)0x80, (byte)0x80, (byte)0x98,
		(byte)0x88, (byte)0x78, (byte)0x00, (byte)0x88, (byte)0x88, (byte)0x88, (byte)0xF8, (byte)0x88, (byte)0x88, (byte)0x88, (byte)0x00, (byte)0x70, (byte)0x20, (byte)0x20, (byte)0x20, (byte)0x20,
		(byte)0x20, (byte)0x70, (byte)0x00, (byte)0x08, (byte)0x08, (byte)0x08, (byte)0x08, (byte)0x08, (byte)0x88, (byte)0x70, (byte)0x00, (byte)0x88, (byte)0x90, (byte)0xA0, (byte)0xC0, (byte)0xA0,
		(byte)0x90, (byte)0x88, (byte)0x00, (byte)0x80, (byte)0x80, (byte)0x80, (byte)0x80, (byte)0x80, (byte)0x80, (byte)0xF8, (byte)0x00, (byte)0x88, (byte)0xD8, (byte)0xA8, (byte)0xA8, (byte)0x88,
		(byte)0x88, (byte)0x88, (byte)0x00, (byte)0x88, (byte)0x88, (byte)0xC8, (byte)0xA8, (byte)0x98, (byte)0x88, (byte)0x88, (byte)0x00, (byte)0x70, (byte)0x88, (byte)0x88, (byte)0x88, (byte)0x88,
		(byte)0x88, (byte)0x70, (byte)0x00, (byte)0xF0, (byte)0x88, (byte)0x88, (byte)0xF0, (byte)0x80, (byte)0x80, (byte)0x80, (byte)0x00, (byte)0x70, (byte)0x88, (byte)0x88, (byte)0x88, (byte)0xA8,
		(byte)0x90, (byte)0x68, (byte)0x00, (byte)0xF0, (byte)0x88, (byte)0x88, (byte)0xF0, (byte)0xA0, (byte)0x90, (byte)0x88, (byte)0x00, (byte)0x70, (byte)0x88, (byte)0x80, (byte)0x70, (byte)0x08,
		(byte)0x88, (byte)0x70, (byte)0x00, (byte)0xF8, (byte)0x20, (byte)0x20, (byte)0x20, (byte)0x20, (byte)0x20, (byte)0x20, (byte)0x00, (byte)0x88, (byte)0x88, (byte)0x88, (byte)0x88, (byte)0x88,
		(byte)0x88, (byte)0x70, (byte)0x00, (byte)0x88, (byte)0x88, (byte)0x88, (byte)0x88, (byte)0x88, (byte)0x50, (byte)0x20, (byte)0x00, (byte)0x88, (byte)0x88, (byte)0x88, (byte)0xA8, (byte)0xA8,
		(byte)0xD8, (byte)0x88, (byte)0x00, (byte)0x88, (byte)0x88, (byte)0x50, (byte)0x20, (byte)0x50, (byte)0x88, (byte)0x88, (byte)0x00, (byte)0x88, (byte)0x88, (byte)0x50, (byte)0x20, (byte)0x20,
		(byte)0x20, (byte)0x20, (byte)0x00, (byte)0xF8, (byte)0x08, (byte)0x10, (byte)0x20, (byte)0x40, (byte)0x80, (byte)0xF8, (byte)0x00, (byte)0xF8, (byte)0xC0, (byte)0xC0, (byte)0xC0, (byte)0xC0,
		(byte)0xC0, (byte)0xF8, (byte)0x00, (byte)0x00, (byte)0x80, (byte)0x40, (byte)0x20, (byte)0x10, (byte)0x08, (byte)0x00, (byte)0x00, (byte)0xF8, (byte)0x18, (byte)0x18, (byte)0x18, (byte)0x18,
		(byte)0x18, (byte)0xF8, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x20, (byte)0x50, (byte)0x88, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,
		(byte)0x00, (byte)0x00, (byte)0xF8, (byte)0x40, (byte)0x20, (byte)0x10, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x70, (byte)0x88, (byte)0xF8,
		(byte)0x88, (byte)0x88, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF0, (byte)0x48, (byte)0x70, (byte)0x48, (byte)0xF0, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x78, (byte)0x80, (byte)0x80,
		(byte)0x80, (byte)0x78, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF0, (byte)0x48, (byte)0x48, (byte)0x48, (byte)0xF0, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF0, (byte)0x80, (byte)0xE0,
		(byte)0x80, (byte)0xF0, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF0, (byte)0x80, (byte)0xE0, (byte)0x80, (byte)0x80, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x78, (byte)0x80, (byte)0xB8,
		(byte)0x88, (byte)0x70, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x88, (byte)0x88, (byte)0xF8, (byte)0x88, (byte)0x88, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF8, (byte)0x20, (byte)0x20,
		(byte)0x20, (byte)0xF8, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x70, (byte)0x20, (byte)0x20, (byte)0xA0, (byte)0xE0, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x90, (byte)0xA0, (byte)0xC0,
		(byte)0xA0, (byte)0x90, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x80, (byte)0x80, (byte)0x80, (byte)0x80, (byte)0xF8, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x88, (byte)0xD8, (byte)0xA8,
		(byte)0x88, (byte)0x88, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x88, (byte)0xC8, (byte)0xA8, (byte)0x98, (byte)0x88, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF8, (byte)0x88, (byte)0x88,
		(byte)0x88, (byte)0xF8, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF0, (byte)0x88, (byte)0xF0, (byte)0x80, (byte)0x80, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF8, (byte)0x88, (byte)0xA8,
		(byte)0x90, (byte)0xE0, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF8, (byte)0x88, (byte)0xF8, (byte)0xA0, (byte)0x90, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x78, (byte)0x80, (byte)0x70,
		(byte)0x08, (byte)0xF0, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF8, (byte)0x20, (byte)0x20, (byte)0x20, (byte)0x20, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x88, (byte)0x88, (byte)0x88,
		(byte)0x88, (byte)0x70, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x88, (byte)0x88, (byte)0x90, (byte)0xA0, (byte)0x40, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x88, (byte)0x88, (byte)0xA8,
		(byte)0xD8, (byte)0x88, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x88, (byte)0x60, (byte)0x20, (byte)0x60, (byte)0x88, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x88, (byte)0x50, (byte)0x20,
		(byte)0x20, (byte)0x20, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xF8, (byte)0x10, (byte)0x20, (byte)0x40, (byte)0xF8, (byte)0x00, (byte)0x38, (byte)0x40, (byte)0x20, (byte)0xC0, (byte)0x20,
		(byte)0x40, (byte)0x38, (byte)0x00, (byte)0x40, (byte)0x20, (byte)0x10, (byte)0x08, (byte)0x10, (byte)0x20, (byte)0x40, (byte)0x00, (byte)0xE0, (byte)0x10, (byte)0x20, (byte)0x18, (byte)0x20,
		(byte)0x10, (byte)0xE0, (byte)0x00, (byte)0x40, (byte)0xA8, (byte)0x10, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00, (byte)0xA8, (byte)0x50, (byte)0xA8, (byte)0x50, (byte)0xA8,
		(byte)0x50, (byte)0xA8, (byte)0x00, (byte)0x01, (byte)0x02, (byte)0x0E, (byte)0x0F, (byte)0x08, (byte)0x09, (byte)0x12, (byte)0x13, (byte)0x03, (byte)0x04, (byte)0x0E, (byte)0x0F, (byte)0x05,
		(byte)0x14, (byte)0x00, (byte)0x00, (byte)0x05, (byte)0x00, (byte)0x10, (byte)0x11, (byte)0x0A, (byte)0x0B, (byte)0x15, (byte)0x16, (byte)0x06, (byte)0x07, (byte)0x10, (byte)0x11, (byte)0x05,
		(byte)0x14, (byte)0x00, (byte)0x00, (byte)0x01, (byte)0x02, (byte)0x0E, (byte)0x0F, (byte)0x03, (byte)0x04, (byte)0x0E, (byte)0x0F, (byte)0x03, (byte)0x04, (byte)0x0E, (byte)0x0F, (byte)0x0C,
		(byte)0x0D, (byte)0x17, (byte)0x18, (byte)0xFF, (byte)0x4F, (byte)0x7D, (byte)0xD3, (byte)0xBF, (byte)0x7C, (byte)0xF6, (byte)0x40, (byte)0xD3, (byte)0xBF, (byte)0x79, (byte)0xD3, (byte)0xBE,
		(byte)0x1B, (byte)0x7A, (byte)0xB3, (byte)0x20, (byte)0xF8, (byte)0xCD, (byte)0xDC, (byte)0x1F, (byte)0xC9, (byte)0x06, (byte)0x00, (byte)0x0E, (byte)0x00, (byte)0xCD, (byte)0xD9, (byte)0x1F,
		(byte)0x06, (byte)0x01, (byte)0x0E, (byte)0x80, (byte)0xCD, (byte)0xD9, (byte)0x1F, (byte)0x3E, (byte)0x02, (byte)0x21, (byte)0x00, (byte)0x18, (byte)0xCD, (byte)0xB8, (byte)0x1F, (byte)0x3E,
		(byte)0x04, (byte)0x21, (byte)0x00, (byte)0x20, (byte)0xCD, (byte)0xB8, (byte)0x1F, (byte)0x3E, (byte)0x03, (byte)0x21, (byte)0x00, (byte)0x00, (byte)0xCD, (byte)0xB8, (byte)0x1F, (byte)0x3E,
		(byte)0x00, (byte)0x21, (byte)0x00, (byte)0x1B, (byte)0xCD, (byte)0xB8, (byte)0x1F, (byte)0x3E, (byte)0x01, (byte)0x21, (byte)0x00, (byte)0x38, (byte)0xCD, (byte)0xB8, (byte)0x1F, (byte)0x06,
		(byte)0x07, (byte)0x0E, (byte)0x00, (byte)0xCD, (byte)0xD9, (byte)0x1F, (byte)0xC9, (byte)0x21, (byte)0x8B, (byte)0x15, (byte)0x11, (byte)0x1D, (byte)0x00, (byte)0xFD, (byte)0x21, (byte)0x60,
		(byte)0x00, (byte)0x3E, (byte)0x03, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0x21, (byte)0xA3, (byte)0x15, (byte)0x11, (byte)0x00, (byte)0x00, (byte)0xFD, (byte)0x21, (byte)0x01, (byte)0x00,
		(byte)0x3E, (byte)0x03, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0xC9, (byte)0x01, (byte)0x00, (byte)0x00, (byte)0x7E, (byte)0xFE, (byte)0x2F, (byte)0xC8, (byte)0x23, (byte)0x03, (byte)0x18,
		(byte)0xF8, (byte)0xC5, (byte)0xFD, (byte)0xE1, (byte)0x3E, (byte)0x20, (byte)0x99, (byte)0x1F, (byte)0x06, (byte)0x00, (byte)0x4F, (byte)0x09, (byte)0x44, (byte)0x4D, (byte)0x62, (byte)0x6B,
		(byte)0x50, (byte)0x59, (byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0xC9, (byte)0x21, (byte)0x00, (byte)0x17, (byte)0x11, (byte)0xFF, (byte)0x00, (byte)0x1B, (byte)0x7A,
		(byte)0xB3, (byte)0x20, (byte)0xFB, (byte)0x2B, (byte)0x7C, (byte)0xB5, (byte)0x20, (byte)0xF3, (byte)0xC9, (byte)0x21, (byte)0x00, (byte)0x00, (byte)0x11, (byte)0x00, (byte)0x40, (byte)0x3E,
		(byte)0x00, (byte)0xCD, (byte)0x82, (byte)0x1F, (byte)0xCD, (byte)0x85, (byte)0x1F, (byte)0x06, (byte)0x0F, (byte)0x0E, (byte)0x04, (byte)0xCD, (byte)0xD9, (byte)0x1F, (byte)0xCD, (byte)0x7F,
		(byte)0x1F, (byte)0x21, (byte)0x7C, (byte)0x1A, (byte)0x11, (byte)0x25, (byte)0x00, (byte)0xFD, (byte)0x21, (byte)0x16, (byte)0x00, (byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0xBE, (byte)0x1F,
		(byte)0x21, (byte)0x92, (byte)0x1A, (byte)0x11, (byte)0x65, (byte)0x00, (byte)0xFD, (byte)0x21, (byte)0x17, (byte)0x00, (byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0x11,
		(byte)0xC5, (byte)0x00, (byte)0xCD, (byte)0xCA, (byte)0x1A, (byte)0x11, (byte)0x05, (byte)0x01, (byte)0xCD, (byte)0xCA, (byte)0x1A, (byte)0x11, (byte)0x45, (byte)0x01, (byte)0xCD, (byte)0xCA,
		(byte)0x1A, (byte)0x11, (byte)0x85, (byte)0x01, (byte)0xCD, (byte)0xCA, (byte)0x1A, (byte)0x11, (byte)0xE5, (byte)0x01, (byte)0xCD, (byte)0xCA, (byte)0x1A, (byte)0x11, (byte)0x25, (byte)0x02,
		(byte)0xCD, (byte)0xCA, (byte)0x1A, (byte)0x11, (byte)0x65, (byte)0x02, (byte)0xCD, (byte)0xCA, (byte)0x1A, (byte)0x11, (byte)0xA5, (byte)0x02, (byte)0xCD, (byte)0xCA, (byte)0x1A, (byte)0x11,
		(byte)0x05, (byte)0x01, (byte)0xCD, (byte)0xD7, (byte)0x1A, (byte)0x11, (byte)0x45, (byte)0x01, (byte)0xCD, (byte)0xDC, (byte)0x1A, (byte)0x11, (byte)0x85, (byte)0x01, (byte)0xCD, (byte)0xE1,
		(byte)0x1A, (byte)0x21, (byte)0xC2, (byte)0x1A, (byte)0x11, (byte)0xE5, (byte)0x01, (byte)0xCD, (byte)0xE4, (byte)0x1A, (byte)0x21, (byte)0xC3, (byte)0x1A, (byte)0x11, (byte)0x25, (byte)0x02,
		(byte)0xCD, (byte)0xE4, (byte)0x1A, (byte)0x21, (byte)0xC4, (byte)0x1A, (byte)0x11, (byte)0x65, (byte)0x02, (byte)0xCD, (byte)0xE4, (byte)0x1A, (byte)0x21, (byte)0xC5, (byte)0x1A, (byte)0x11,
		(byte)0xA5, (byte)0x02, (byte)0xCD, (byte)0xE4, (byte)0x1A, (byte)0x11, (byte)0x0F, (byte)0x01, (byte)0xCD, (byte)0xD7, (byte)0x1A, (byte)0x11, (byte)0x4F, (byte)0x01, (byte)0xCD, (byte)0xDC,
		(byte)0x1A, (byte)0x11, (byte)0x8F, (byte)0x01, (byte)0xCD, (byte)0xE1, (byte)0x1A, (byte)0x11, (byte)0xF1, (byte)0x01, (byte)0xCD, (byte)0xEE, (byte)0x1A, (byte)0x11, (byte)0x31, (byte)0x02,
		(byte)0xCD, (byte)0xEE, (byte)0x1A, (byte)0x11, (byte)0x71, (byte)0x02, (byte)0xCD, (byte)0xEE, (byte)0x1A, (byte)0x11, (byte)0xB1, (byte)0x02, (byte)0xCD, (byte)0xEE, (byte)0x1A, (byte)0x11,
		(byte)0x2F, (byte)0x02, (byte)0xCD, (byte)0xD7, (byte)0x1A, (byte)0x11, (byte)0x6F, (byte)0x02, (byte)0xCD, (byte)0xDC, (byte)0x1A, (byte)0x11, (byte)0xAF, (byte)0x02, (byte)0xCD, (byte)0xE1,
		(byte)0x1A, (byte)0x11, (byte)0xFB, (byte)0x01, (byte)0xCD, (byte)0xFB, (byte)0x1A, (byte)0x11, (byte)0x3B, (byte)0x02, (byte)0xCD, (byte)0xFB, (byte)0x1A, (byte)0x11, (byte)0x7B, (byte)0x02,
		(byte)0xCD, (byte)0xFB, (byte)0x1A, (byte)0x11, (byte)0xBB, (byte)0x02, (byte)0xCD, (byte)0xFB, (byte)0x1A, (byte)0x2A, (byte)0xFA, (byte)0x73, (byte)0x11, (byte)0x20, (byte)0x00, (byte)0x3E,
		(byte)0xF4, (byte)0xCD, (byte)0x82, (byte)0x1F, (byte)0x06, (byte)0x01, (byte)0x0E, (byte)0xC0, (byte)0xCD, (byte)0xD9, (byte)0x1F, (byte)0xC9, (byte)0x54, (byte)0x4F, (byte)0x20, (byte)0x53,
		(byte)0x45, (byte)0x4C, (byte)0x45, (byte)0x43, (byte)0x54, (byte)0x20, (byte)0x47, (byte)0x41, (byte)0x4D, (byte)0x45, (byte)0x20, (byte)0x4F, (byte)0x50, (byte)0x54, (byte)0x49, (byte)0x4F,
		(byte)0x4E, (byte)0x2C, (byte)0x50, (byte)0x52, (byte)0x45, (byte)0x53, (byte)0x53, (byte)0x20, (byte)0x42, (byte)0x55, (byte)0x54, (byte)0x54, (byte)0x4F, (byte)0x4E, (byte)0x20, (byte)0x4F,
		(byte)0x4E, (byte)0x20, (byte)0x4B, (byte)0x45, (byte)0x59, (byte)0x50, (byte)0x41, (byte)0x44, (byte)0x2E, (byte)0x31, (byte)0x20, (byte)0x3D, (byte)0x20, (byte)0x53, (byte)0x4B, (byte)0x49,
		(byte)0x4C, (byte)0x4C, (byte)0x20, (byte)0x31, (byte)0x2F, (byte)0x4F, (byte)0x4E, (byte)0x45, (byte)0x20, (byte)0x50, (byte)0x4C, (byte)0x41, (byte)0x59, (byte)0x45, (byte)0x52, (byte)0x32,
		(byte)0x33, (byte)0x34, (byte)0x35, (byte)0x36, (byte)0x37, (byte)0x38, (byte)0x54, (byte)0x57, (byte)0x4F, (byte)0x53, (byte)0x21, (byte)0xA9, (byte)0x1A, (byte)0xFD, (byte)0x21, (byte)0x16,
		(byte)0x00, (byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0xC9, (byte)0x21, (byte)0xBF, (byte)0x1A, (byte)0x18, (byte)0x08, (byte)0x21, (byte)0xC0, (byte)0x1A, (byte)0x18,
		(byte)0x03, (byte)0x21, (byte)0xC1, (byte)0x1A, (byte)0xFD, (byte)0x21, (byte)0x01, (byte)0x00, (byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0xC9, (byte)0x21, (byte)0xC6,
		(byte)0x1A, (byte)0xFD, (byte)0x21, (byte)0x03, (byte)0x00, (byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0xC9, (byte)0x21, (byte)0xC9, (byte)0x1A, (byte)0xFD, (byte)0x21,
		(byte)0x01, (byte)0x00, (byte)0x3E, (byte)0x02, (byte)0xCD, (byte)0xBE, (byte)0x1F, (byte)0xC9, (byte)0x02, (byte)0x00, (byte)0x01, (byte)0x00, (byte)0x02, (byte)0x00, (byte)0x01, (byte)0x08,
		(byte)0x1B, (byte)0x11, (byte)0xBA, (byte)0x73, (byte)0xCD, (byte)0x98, (byte)0x00, (byte)0x3A, (byte)0xBA, (byte)0x73, (byte)0x2A, (byte)0xBB, (byte)0x73, (byte)0x4F, (byte)0x06, (byte)0x00,
		(byte)0xDD, (byte)0x21, (byte)0xF2, (byte)0x73, (byte)0xDD, (byte)0x09, (byte)0xDD, (byte)0x09, (byte)0xDD, (byte)0x75, (byte)0x00, (byte)0xDD, (byte)0x74, (byte)0x01, (byte)0x3A, (byte)0xC3,
		(byte)0x73, (byte)0xCB, (byte)0x4F, (byte)0x28, (byte)0x27, (byte)0x79, (byte)0xFE, (byte)0x03, (byte)0x28, (byte)0x06, (byte)0xFE, (byte)0x04, (byte)0x28, (byte)0x10, (byte)0x18, (byte)0x1C,
		(byte)0x06, (byte)0x04, (byte)0x7D, (byte)0xB4, (byte)0x20, (byte)0x04, (byte)0x0E, (byte)0x03, (byte)0x18, (byte)0x28, (byte)0x0E, (byte)0x07, (byte)0x18, (byte)0x24, (byte)0x06, (byte)0x03,
		(byte)0x7D, (byte)0xB4, (byte)0x20, (byte)0x04, (byte)0x0E, (byte)0x7F, (byte)0x18, (byte)0x1A, (byte)0x0E, (byte)0xFF, (byte)0x18, (byte)0x16, (byte)0xFD, (byte)0x21, (byte)0x76, (byte)0x1B,
		(byte)0xFD, (byte)0x09, (byte)0xFD, (byte)0x09, (byte)0xFD, (byte)0x7E, (byte)0x00, (byte)0xFD, (byte)0x46, (byte)0x01, (byte)0xCB, (byte)0x3C, (byte)0xCB, (byte)0x1D, (byte)0x3D, (byte)0x20,
		(byte)0xF9, (byte)0x4D, (byte)0xCD, (byte)0xCA, (byte)0x1C, (byte)0xC9, (byte)0x07, (byte)0x05, (byte)0x0B, (byte)0x06, (byte)0x0A, (byte)0x02, (byte)0x0B, (byte)0x04, (byte)0x06, (byte)0x03,
		(byte)0x05, (byte)0x00, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0x00, (byte)0xFE, (byte)0xFF, (byte)0x02, (byte)0x00, (byte)0x01, (byte)0x80, (byte)0x1B, (byte)0x11,
		(byte)0xBA, (byte)0x73, (byte)0xCD, (byte)0x98, (byte)0x00, (byte)0x3A, (byte)0xBA, (byte)0x73, (byte)0xED, (byte)0x5B, (byte)0xBB, (byte)0x73, (byte)0xFD, (byte)0x2A, (byte)0xBF, (byte)0x73,
		(byte)0x2A, (byte)0xBD, (byte)0x73, (byte)0xCD, (byte)0xAA, (byte)0x1B, (byte)0xCD, (byte)0x3E, (byte)0x1D, (byte)0xC9, (byte)0xFD, (byte)0x22, (byte)0xFE, (byte)0x73, (byte)0xDD, (byte)0x21,
		(byte)0xF2, (byte)0x73, (byte)0x4F, (byte)0x06, (byte)0x00, (byte)0xFE, (byte)0x04, (byte)0x20, (byte)0x07, (byte)0x3A, (byte)0xC3, (byte)0x73, (byte)0xCB, (byte)0x4F, (byte)0x28, (byte)0x2C,
		(byte)0xFD, (byte)0x21, (byte)0xFF, (byte)0x1B, (byte)0xFD, (byte)0x09, (byte)0xFD, (byte)0x7E, (byte)0x00, (byte)0xFE, (byte)0x00, (byte)0x28, (byte)0x1F, (byte)0xCB, (byte)0x23, (byte)0xCB,
		(byte)0x12, (byte)0x3D, (byte)0x20, (byte)0xF9, (byte)0xC5, (byte)0xED, (byte)0x4B, (byte)0xFE, (byte)0x73, (byte)0xFD, (byte)0x7E, (byte)0x00, (byte)0xFE, (byte)0x00, (byte)0x28, (byte)0x0B,
		(byte)0xCB, (byte)0x21, (byte)0xCB, (byte)0x10, (byte)0x3D, (byte)0x20, (byte)0xF9, (byte)0xED, (byte)0x43, (byte)0xFE, (byte)0x73, (byte)0xC1, (byte)0xE5, (byte)0xDD, (byte)0x09, (byte)0xDD,
		(byte)0x09, (byte)0xDD, (byte)0x6E, (byte)0x00, (byte)0xDD, (byte)0x66, (byte)0x01, (byte)0x19, (byte)0xEB, (byte)0xE1, (byte)0xED, (byte)0x4B, (byte)0xFE, (byte)0x73, (byte)0xC9, (byte)0x02,
		(byte)0x03, (byte)0x00, (byte)0x03, (byte)0x03, (byte)0x05, (byte)0x00, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0x00, (byte)0xFE, (byte)0xFF, (byte)0x02, (byte)0x00,
		(byte)0x01, (byte)0x04, (byte)0x1C, (byte)0x11, (byte)0xBA, (byte)0x73, (byte)0xCD, (byte)0x98, (byte)0x00, (byte)0x3A, (byte)0xBA, (byte)0x73, (byte)0xED, (byte)0x5B, (byte)0xBB, (byte)0x73,
		(byte)0xFD, (byte)0x2A, (byte)0xBF, (byte)0x73, (byte)0x2A, (byte)0xBD, (byte)0x73, (byte)0xF5, (byte)0xFE, (byte)0x00, (byte)0x20, (byte)0x22, (byte)0x3A, (byte)0xC7, (byte)0x73, (byte)0xFE,
		(byte)0x01, (byte)0x20, (byte)0x1B, (byte)0xF1, (byte)0xE5, (byte)0x2A, (byte)0x02, (byte)0x80, (byte)0x7B, (byte)0xCB, (byte)0x27, (byte)0xCB, (byte)0x27, (byte)0x5F, (byte)0x19, (byte)0xEB,
		(byte)0xFD, (byte)0xE5, (byte)0xC1, (byte)0x79, (byte)0xCB, (byte)0x27, (byte)0xCB, (byte)0x27, (byte)0x4F, (byte)0xE1, (byte)0xED, (byte)0xB0, (byte)0x18, (byte)0x07, (byte)0xF1, (byte)0xCD,
		(byte)0xAA, (byte)0x1B, (byte)0xCD, (byte)0x01, (byte)0x1D, (byte)0xC9, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0x56, (byte)0x1C, (byte)0x11, (byte)0xBA, (byte)0x73,
		(byte)0xCD, (byte)0x98, (byte)0x00, (byte)0x3A, (byte)0xBA, (byte)0x73, (byte)0x47, (byte)0xAF, (byte)0x2A, (byte)0x04, (byte)0x80, (byte)0x77, (byte)0x23, (byte)0x3C, (byte)0xB8, (byte)0x20,
		(byte)0xFA, (byte)0xC9, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0x72, (byte)0x1C, (byte)0x11, (byte)0xBA, (byte)0x73, (byte)0xCD, (byte)0x98, (byte)0x00, (byte)0x3A,
		(byte)0xBA, (byte)0x73, (byte)0xDD, (byte)0x2A, (byte)0x04, (byte)0x80, (byte)0xF5, (byte)0xFD, (byte)0x21, (byte)0xF2, (byte)0x73, (byte)0xFD, (byte)0x5E, (byte)0x00, (byte)0xFD, (byte)0x56,
		(byte)0x01, (byte)0x7B, (byte)0xD3, (byte)0xBF, (byte)0x7A, (byte)0xF6, (byte)0x40, (byte)0xD3, (byte)0xBF, (byte)0xF1, (byte)0x2A, (byte)0x02, (byte)0x80, (byte)0xDD, (byte)0x4E, (byte)0x00,
		(byte)0xDD, (byte)0x23, (byte)0x06, (byte)0x00, (byte)0x09, (byte)0x09, (byte)0x09, (byte)0x09, (byte)0x06, (byte)0x04, (byte)0x0E, (byte)0xBE, (byte)0xED, (byte)0xA3, (byte)0x00, (byte)0x00,
		(byte)0x20, (byte)0xFA, (byte)0x3D, (byte)0x20, (byte)0xE5, (byte)0xC9, (byte)0x02, (byte)0x00, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0x00, (byte)0x01, (byte)0xB6, (byte)0x1C, (byte)0x11,
		(byte)0xBA, (byte)0x73, (byte)0xCD, (byte)0x98, (byte)0x00, (byte)0x2A, (byte)0xBA, (byte)0x73, (byte)0x4C, (byte)0x45, (byte)0x79, (byte)0xD3, (byte)0xBF, (byte)0x78, (byte)0xC6, (byte)0x80,
		(byte)0xD3, (byte)0xBF, (byte)0x78, (byte)0xFE, (byte)0x00, (byte)0x20, (byte)0x04, (byte)0x79, (byte)0x32, (byte)0xC3, (byte)0x73, (byte)0x78, (byte)0xFE, (byte)0x01, (byte)0x20, (byte)0x04,
		(byte)0x79, (byte)0x32, (byte)0xC4, (byte)0x73, (byte)0xC9, (byte)0x03, (byte)0x00, (byte)0xFE, (byte)0xFF, (byte)0x02, (byte)0x00, (byte)0x02, (byte)0x00, (byte)0x01, (byte)0xE5, (byte)0x1C,
		(byte)0x11, (byte)0xBA, (byte)0x73, (byte)0xCD, (byte)0x98, (byte)0x00, (byte)0x2A, (byte)0xBA, (byte)0x73, (byte)0xED, (byte)0x5B, (byte)0xBC, (byte)0x73, (byte)0xED, (byte)0x4B, (byte)0xBE,
		(byte)0x73, (byte)0xE5, (byte)0xD5, (byte)0xE1, (byte)0x11, (byte)0x00, (byte)0x40, (byte)0x19, (byte)0x7D, (byte)0xD3, (byte)0xBF, (byte)0x7C, (byte)0xD3, (byte)0xBF, (byte)0xC5, (byte)0xD1,
		(byte)0xE1, (byte)0x0E, (byte)0xBE, (byte)0x43, (byte)0xED, (byte)0xA3, (byte)0x00, (byte)0x00, (byte)0xC2, (byte)0x14, (byte)0x1D, (byte)0x15, (byte)0xFA, (byte)0x21, (byte)0x1D, (byte)0x20,
		(byte)0xF3, (byte)0xC9, (byte)0x03, (byte)0x00, (byte)0xFE, (byte)0xFF, (byte)0x02, (byte)0x00, (byte)0x02, (byte)0x00, (byte)0x01, (byte)0x22, (byte)0x1D, (byte)0x11, (byte)0xBA, (byte)0x73,
		(byte)0xCD, (byte)0x98, (byte)0x00, (byte)0x2A, (byte)0xBA, (byte)0x73, (byte)0xED, (byte)0x5B, (byte)0xBC, (byte)0x73, (byte)0xED, (byte)0x4B, (byte)0xBE, (byte)0x73, (byte)0x7B, (byte)0xD3,
		(byte)0xBF, (byte)0x7A, (byte)0xD3, (byte)0xBF, (byte)0xC5, (byte)0xD1, (byte)0x0E, (byte)0xBE, (byte)0x43, (byte)0xED, (byte)0xA2, (byte)0x00, (byte)0x00, (byte)0xC2, (byte)0x49, (byte)0x1D,
		(byte)0x15, (byte)0xFA, (byte)0x56, (byte)0x1D, (byte)0x20, (byte)0xF3, (byte)0xC9, (byte)0xDB, (byte)0xBF, (byte)0xC9, (byte)0xDD, (byte)0x21, (byte)0x96, (byte)0x1D, (byte)0x18, (byte)0x10,
		(byte)0xDD, (byte)0x21, (byte)0xB7, (byte)0x1D, (byte)0x18, (byte)0x0A, (byte)0xDD, (byte)0x21, (byte)0xE5, (byte)0x1D, (byte)0x18, (byte)0x04, (byte)0xDD, (byte)0x21, (byte)0x07, (byte)0x1E,
		(byte)0xD9, (byte)0x08, (byte)0xDD, (byte)0xE5, (byte)0x08, (byte)0xF5, (byte)0x08, (byte)0xF1, (byte)0xD9, (byte)0xD5, (byte)0xD9, (byte)0xD1, (byte)0xFD, (byte)0x21, (byte)0x01, (byte)0x00,
		(byte)0x2A, (byte)0x06, (byte)0x80, (byte)0xCD, (byte)0xA3, (byte)0x1B, (byte)0xDD, (byte)0xE1, (byte)0xDD, (byte)0xE5, (byte)0xDD, (byte)0xE9, (byte)0x13, (byte)0x0B, (byte)0x78, (byte)0xB1,
		(byte)0xD9, (byte)0x20, (byte)0xE1, (byte)0xDD, (byte)0xE1, (byte)0xC9, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x01, (byte)0x08, (byte)0x00, (byte)0xE5, (byte)0xD1, (byte)0x09, (byte)0xEB,
		(byte)0xCD, (byte)0x00, (byte)0x1F, (byte)0xCD, (byte)0x72, (byte)0x1E, (byte)0xCD, (byte)0x5D, (byte)0x1E, (byte)0xFE, (byte)0x01, (byte)0x20, (byte)0x06, (byte)0xCD, (byte)0x89, (byte)0x1E,
		(byte)0xCD, (byte)0x9A, (byte)0x1E, (byte)0xD9, (byte)0x23, (byte)0x18, (byte)0xD5, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x01, (byte)0x08, (byte)0x00, (byte)0xE5, (byte)0xD1, (byte)0x09,
		(byte)0xEB, (byte)0xCD, (byte)0x4E, (byte)0x1F, (byte)0xCD, (byte)0x72, (byte)0x1E, (byte)0xCD, (byte)0x5D, (byte)0x1E, (byte)0xFE, (byte)0x01, (byte)0x20, (byte)0x13, (byte)0xCD, (byte)0x89,
		(byte)0x1E, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x01, (byte)0x08, (byte)0x00, (byte)0xE5, (byte)0xD1, (byte)0x09, (byte)0xEB, (byte)0xCD, (byte)0x4E, (byte)0x1F, (byte)0xCD, (byte)0x9A,
		(byte)0x1E, (byte)0xD9, (byte)0x23, (byte)0x18, (byte)0xA7, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x01, (byte)0x08, (byte)0x00, (byte)0xE5, (byte)0xD1, (byte)0x09, (byte)0xEB, (byte)0xCD,
		(byte)0x12, (byte)0x1F, (byte)0xCD, (byte)0x72, (byte)0x1E, (byte)0xCD, (byte)0x5D, (byte)0x1E, (byte)0xFE, (byte)0x01, (byte)0x20, (byte)0x06, (byte)0xCD, (byte)0x89, (byte)0x1E, (byte)0xCD,
		(byte)0x9A, (byte)0x1E, (byte)0xD9, (byte)0x23, (byte)0xC3, (byte)0x8C, (byte)0x1D, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x01, (byte)0x08, (byte)0x00, (byte)0xE5, (byte)0xD1, (byte)0x09,
		(byte)0xEB, (byte)0xCD, (byte)0xAB, (byte)0x1E, (byte)0x08, (byte)0xF5, (byte)0x08, (byte)0xF1, (byte)0xD9, (byte)0xE5, (byte)0xD9, (byte)0xD1, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x01,
		(byte)0x08, (byte)0x00, (byte)0x09, (byte)0xFD, (byte)0x21, (byte)0x04, (byte)0x00, (byte)0xCD, (byte)0x27, (byte)0x1C, (byte)0xCD, (byte)0x5D, (byte)0x1E, (byte)0xFE, (byte)0x01, (byte)0x20,
		(byte)0x24, (byte)0xCD, (byte)0x89, (byte)0x1E, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x01, (byte)0x08, (byte)0x00, (byte)0xE5, (byte)0xD1, (byte)0x09, (byte)0xEB, (byte)0xCD, (byte)0xEA,
		(byte)0x1E, (byte)0x3E, (byte)0x04, (byte)0xD9, (byte)0xE5, (byte)0xD9, (byte)0xD1, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x01, (byte)0x08, (byte)0x00, (byte)0x09, (byte)0xFD, (byte)0x21,
		(byte)0x04, (byte)0x00, (byte)0xCD, (byte)0x27, (byte)0x1C, (byte)0xD9, (byte)0x23, (byte)0x23, (byte)0x23, (byte)0x23, (byte)0xC3, (byte)0x8C, (byte)0x1D, (byte)0x08, (byte)0xF5, (byte)0x08,
		(byte)0xF1, (byte)0xFE, (byte)0x03, (byte)0x20, (byte)0x0A, (byte)0x21, (byte)0xC3, (byte)0x73, (byte)0xCB, (byte)0x4E, (byte)0x28, (byte)0x03, (byte)0x3E, (byte)0x01, (byte)0xC9, (byte)0x3E,
		(byte)0x00, (byte)0xC9, (byte)0x08, (byte)0xF5, (byte)0x08, (byte)0xF1, (byte)0xD9, (byte)0xE5, (byte)0xD9, (byte)0xD1, (byte)0x2A, (byte)0x06, (byte)0x80, (byte)0x01, (byte)0x08, (byte)0x00,
		(byte)0x09, (byte)0xFD, (byte)0x21, (byte)0x01, (byte)0x00, (byte)0xCD, (byte)0x27, (byte)0x1C, (byte)0xC9, (byte)0x3E, (byte)0x04, (byte)0xD9, (byte)0xD5, (byte)0xD9, (byte)0xD1, (byte)0x2A,
		(byte)0x06, (byte)0x80, (byte)0xFD, (byte)0x21, (byte)0x01, (byte)0x00, (byte)0xCD, (byte)0xA3, (byte)0x1B, (byte)0xC9, (byte)0x3E, (byte)0x04, (byte)0xD9, (byte)0xE5, (byte)0xD9, (byte)0xD1,
		(byte)0x2A, (byte)0x06, (byte)0x80, (byte)0xFD, (byte)0x21, (byte)0x01, (byte)0x00, (byte)0xCD, (byte)0x27, (byte)0x1C, (byte)0xC9, (byte)0xE5, (byte)0xDD, (byte)0xE1, (byte)0xD5, (byte)0xFD,
		(byte)0xE1, (byte)0x01, (byte)0x08, (byte)0x00, (byte)0xDD, (byte)0x7E, (byte)0x00, (byte)0xDD, (byte)0x23, (byte)0x57, (byte)0x1E, (byte)0x04, (byte)0xCB, (byte)0x17, (byte)0xCB, (byte)0x14,
		(byte)0xCB, (byte)0x12, (byte)0xCB, (byte)0x14, (byte)0x1D, (byte)0x20, (byte)0xF5, (byte)0x1E, (byte)0x04, (byte)0xCB, (byte)0x17, (byte)0xCB, (byte)0x15, (byte)0xCB, (byte)0x12, (byte)0xCB,
		(byte)0x15, (byte)0x1D, (byte)0x20, (byte)0xF5, (byte)0xFD, (byte)0x74, (byte)0x00, (byte)0xFD, (byte)0x75, (byte)0x10, (byte)0xFD, (byte)0x23, (byte)0xFD, (byte)0x74, (byte)0x00, (byte)0xFD,
		(byte)0x75, (byte)0x10, (byte)0xFD, (byte)0x23, (byte)0x0B, (byte)0x79, (byte)0xB0, (byte)0x20, (byte)0xCB, (byte)0xC9, (byte)0x01, (byte)0x10, (byte)0x00, (byte)0xE5, (byte)0x7E, (byte)0x23,
		(byte)0x12, (byte)0x13, (byte)0x12, (byte)0x13, (byte)0x0B, (byte)0x79, (byte)0xFE, (byte)0x08, (byte)0x20, (byte)0x01, (byte)0xE1, (byte)0x79, (byte)0xB0, (byte)0x20, (byte)0xEF, (byte)0xC9,
		(byte)0x01, (byte)0x08, (byte)0x00, (byte)0x46, (byte)0x3E, (byte)0x80, (byte)0xCB, (byte)0x10, (byte)0x1F, (byte)0x30, (byte)0xFB, (byte)0x12, (byte)0x23, (byte)0x13, (byte)0x0D, (byte)0x20,
		(byte)0xF2, (byte)0xC9, (byte)0xE5, (byte)0xDD, (byte)0xE1, (byte)0xEB, (byte)0x01, (byte)0x08, (byte)0x00, (byte)0xDD, (byte)0xCB, (byte)0x00, (byte)0x16, (byte)0xCB, (byte)0x1E, (byte)0xDD,
		(byte)0xCB, (byte)0x01, (byte)0x16, (byte)0xCB, (byte)0x1E, (byte)0xDD, (byte)0xCB, (byte)0x02, (byte)0x16, (byte)0xCB, (byte)0x1E, (byte)0xDD, (byte)0xCB, (byte)0x03, (byte)0x16, (byte)0xCB,
		(byte)0x1E, (byte)0xDD, (byte)0xCB, (byte)0x04, (byte)0x16, (byte)0xCB, (byte)0x1E, (byte)0xDD, (byte)0xCB, (byte)0x05, (byte)0x16, (byte)0xCB, (byte)0x1E, (byte)0xDD, (byte)0xCB, (byte)0x06,
		(byte)0x16, (byte)0xCB, (byte)0x1E, (byte)0xDD, (byte)0xCB, (byte)0x07, (byte)0x16, (byte)0xCB, (byte)0x1E, (byte)0x23, (byte)0x0D, (byte)0x20, (byte)0xCC, (byte)0xC9, (byte)0x01, (byte)0x07,
		(byte)0x00, (byte)0x09, (byte)0x03, (byte)0x7E, (byte)0x12, (byte)0x13, (byte)0x2B, (byte)0x0B, (byte)0x78, (byte)0xB1, (byte)0x20, (byte)0xF7, (byte)0xC9, (byte)0xFF, (byte)0xFF, (byte)0xFF,
		(byte)0xFF, (byte)0xC3, (byte)0x00, (byte)0x03, (byte)0xC3, (byte)0x88, (byte)0x04, (byte)0xC3, (byte)0xC7, (byte)0x06, (byte)0xC3, (byte)0x5A, (byte)0x1D, (byte)0xC3, (byte)0x60, (byte)0x1D,
		(byte)0xC3, (byte)0x66, (byte)0x1D, (byte)0xC3, (byte)0x6C, (byte)0x1D, (byte)0xC3, (byte)0x4A, (byte)0x11, (byte)0xC3, (byte)0x8B, (byte)0x11, (byte)0xC3, (byte)0x79, (byte)0x19, (byte)0xC3,
		(byte)0x27, (byte)0x19, (byte)0xC3, (byte)0xD4, (byte)0x18, (byte)0xC3, (byte)0xE9, (byte)0x18, (byte)0xC3, (byte)0x6A, (byte)0x11, (byte)0xC3, (byte)0x0E, (byte)0x1B, (byte)0xC3, (byte)0x8C,
		(byte)0x1B, (byte)0xC3, (byte)0x10, (byte)0x1C, (byte)0xC3, (byte)0x5A, (byte)0x1C, (byte)0xC3, (byte)0x76, (byte)0x1C, (byte)0xC3, (byte)0x9A, (byte)0x0F, (byte)0xC3, (byte)0xB8, (byte)0x0F,
		(byte)0xC3, (byte)0x44, (byte)0x10, (byte)0xC3, (byte)0xBF, (byte)0x10, (byte)0xC3, (byte)0xBC, (byte)0x1C, (byte)0xC3, (byte)0xED, (byte)0x1C, (byte)0xC3, (byte)0x2A, (byte)0x1D, (byte)0xC3,
		(byte)0x55, (byte)0x06, (byte)0xC3, (byte)0x03, (byte)0x02, (byte)0xC3, (byte)0x51, (byte)0x02, (byte)0xC3, (byte)0x1D, (byte)0x1B, (byte)0xC3, (byte)0xA3, (byte)0x1B, (byte)0xC3, (byte)0x27,
		(byte)0x1C, (byte)0xC3, (byte)0x66, (byte)0x1C, (byte)0xC3, (byte)0x82, (byte)0x1C, (byte)0xC3, (byte)0xAA, (byte)0x0F, (byte)0xC3, (byte)0xC4, (byte)0x0F, (byte)0xC3, (byte)0x53, (byte)0x10,
		(byte)0xC3, (byte)0xCB, (byte)0x10, (byte)0xC3, (byte)0x37, (byte)0x0F, (byte)0xC3, (byte)0x3B, (byte)0x02, (byte)0xC3, (byte)0xCA, (byte)0x1C, (byte)0xC3, (byte)0x57, (byte)0x1D, (byte)0xC3,
		(byte)0x01, (byte)0x1D, (byte)0xC3, (byte)0x3E, (byte)0x1D, (byte)0xC3, (byte)0x64, (byte)0x06, (byte)0xC3, (byte)0x79, (byte)0x06, (byte)0xC3, (byte)0xC1, (byte)0x11, (byte)0xC3, (byte)0x13,
		(byte)0x02, (byte)0xC3, (byte)0x5E, (byte)0x02, (byte)0xC3, (byte)0x7F, (byte)0x02, (byte)0xC3, (byte)0xA3, (byte)0x04, (byte)0xC3, (byte)0xD8, (byte)0x06, (byte)0xC3, (byte)0x3B, (byte)0x00,
	};

	public ColecoVision()
	{
	}

	public void paint(Graphics g, boolean fHighPriority)
	{
		if (m_fInitialized)
		{
			if (0==m_nFrameSkipCount-- || 0!=m_lMsgHideTime || fHighPriority)
			{
				m_tms9928a.paint(g);
				m_nFrameSkipCount = m_nFrameSkipRate;
			}
		}
		else
			g.clearRect(0, 0, 272, 208);

		// If we're timing a message on screen...
		if (0!=m_lMsgHideTime)
		{
			g.drawString(m_szMsg, 5, 10);

			// See if time for message to go away...
			if (System.currentTimeMillis() >= m_lMsgHideTime)
				m_lMsgHideTime = 0;
		}
	}

	public boolean keyDown(Event evt, int nKey)
	{
		if (null!=m_handCtls)
		{
			switch(nKey)
			{
				case	Event.UP:
					m_handCtls.SetJoyPos(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_JOYPOS_NORTH);
					return true;
				case	Event.DOWN:
					m_handCtls.SetJoyPos(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_JOYPOS_SOUTH);
					return true;
				case	Event.LEFT:
					m_handCtls.SetJoyPos(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_JOYPOS_WEST);
					return true;
				case	Event.RIGHT:
					m_handCtls.SetJoyPos(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_JOYPOS_EAST);
					return true;
				case	' ':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_FIRE1, m_handCtls.m_CV_BUTTON_PRESSED);
					return true;
				case	'\t':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_FIRE2, m_handCtls.m_CV_BUTTON_PRESSED);
					return true;
				case	'1':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_1, m_handCtls.m_CV_BUTTON_PRESSED);
					return true;
				case	'2':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_2, m_handCtls.m_CV_BUTTON_PRESSED);
					return true;
				case	'3':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_3, m_handCtls.m_CV_BUTTON_PRESSED);
					return true;
				case	'4':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_4, m_handCtls.m_CV_BUTTON_PRESSED);
					return true;
				case	'5':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_5, m_handCtls.m_CV_BUTTON_PRESSED);
					return true;
				case	'6':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_6, m_handCtls.m_CV_BUTTON_PRESSED);
					return true;
				case	'7':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_7, m_handCtls.m_CV_BUTTON_PRESSED);
					return true;
				case	'8':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_8, m_handCtls.m_CV_BUTTON_PRESSED);
					return true;
				case	'9':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_9, m_handCtls.m_CV_BUTTON_PRESSED);
					return true;
				case	'0':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_0, m_handCtls.m_CV_BUTTON_PRESSED);
					return true;
				case	'#':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_POUND, m_handCtls.m_CV_BUTTON_PRESSED);
					return true;
				case	'*':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_ASTERISK, m_handCtls.m_CV_BUTTON_PRESSED);
					return true;
				case	0x12:
					if ((evt.modifiers & Event.CTRL_MASK)==Event.CTRL_MASK)
						reset();
					return true;
			}
		}
		return	false;
	}

	public boolean keyUp(Event evt, int key)
	{
		if (null!=m_handCtls)
		{
			switch(key)
			{
				case	Event.UP:
				case	Event.DOWN:
				case	Event.LEFT:
				case	Event.RIGHT:
					m_handCtls.SetJoyPos(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_JOYPOS_CENTER);
					return true;
				case	' ':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_FIRE1, m_handCtls.m_CV_BUTTON_NOTPRESSED);
					return true;
				case	'\t':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_FIRE2, m_handCtls.m_CV_BUTTON_NOTPRESSED);
					return true;
				case	'1':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_1, m_handCtls.m_CV_BUTTON_NOTPRESSED);
					return true;
				case	'2':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_2, m_handCtls.m_CV_BUTTON_NOTPRESSED);
					return true;
				case	'3':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_3, m_handCtls.m_CV_BUTTON_NOTPRESSED);
					return true;
				case	'4':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_4, m_handCtls.m_CV_BUTTON_NOTPRESSED);
					return true;
				case	'5':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_5, m_handCtls.m_CV_BUTTON_NOTPRESSED);
					return true;
				case	'6':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_6, m_handCtls.m_CV_BUTTON_NOTPRESSED);
					return true;
				case	'7':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_7, m_handCtls.m_CV_BUTTON_NOTPRESSED);
					return true;
				case	'8':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_8, m_handCtls.m_CV_BUTTON_NOTPRESSED);
					return true;
				case	'9':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_9, m_handCtls.m_CV_BUTTON_NOTPRESSED);
					return true;
				case	'0':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_0, m_handCtls.m_CV_BUTTON_NOTPRESSED);
					return true;
				case	'#':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_POUND, m_handCtls.m_CV_BUTTON_NOTPRESSED);
					return true;
				case	'*':
					m_handCtls.SetButtonState(m_handCtls.m_CV_PRIMARY_CONTROLLER, m_handCtls.m_CV_NUMPAD_ASTERISK, m_handCtls.m_CV_BUTTON_NOTPRESSED);
					return true;
			}
		}
		return	false;
	}

	public boolean reset()
	{
		synchronized(m_z80cpu)
		{
			super.reset();

			// Reset CPU...
			m_z80cpu.reset();

			//	Copy COLECO.ROM code into memory...
			for (int i = 0; i < m_rgbOSROM.length; i++)
				MemWriteByte((char)i, (char)m_rgbOSROM[i]);

			if (null != m_rgbCartridgeROM)
				//	Copy cartridge ROM code into cartridge memory...
				for (int i = 0; i < m_rgbCartridgeROM.length; i++)
					MemWriteByte((char)(i+0x8000), (char)m_rgbCartridgeROM[i]);
		}

		return	true;
	}

	public long getEstimatedCPUSpeed()
	{
		return	m_nCyclesSinceRefresh / m_NMI_RESYNC_TIME * 1000;
	}

	public int setVideoSyncMethod(int nSyncMethod)
	{
		int nOldSync = m_nTimingSyncMethod;
		m_nTimingSyncMethod = nSyncMethod;
		return	nOldSync;
	}

	public int getVideoSyncMethod()
	{
		return	m_nTimingSyncMethod;
	}

	public int setVideoRefreshFreq(int nFreq)
	{
		int	nOldFreq = m_nRefreshFreq;
		m_nRefreshFreq = nFreq;
		return	nOldFreq;
	}

	public int getVideoRefreshFreq()
	{
		return	m_nRefreshFreq;
	}

	public int setVideoRefreshSkipRate(int nRate)
	{
		int nOldRate = m_nFrameSkipRate;
		m_nFrameSkipRate = nRate;
		return	nOldRate;
	}

	public int getVideoRefreshSkipRate()
	{
		return	m_nFrameSkipRate;
	}

	public void lockCPUSpeed(boolean lockit)
	{
		m_fLockCPUSpeed = lockit;
	}

	public boolean isCPUSpeedLocked()
	{
		return	m_fLockCPUSpeed;
	}

	public void pause(boolean pause)
	{
		m_fPaused = pause;
	}

	public boolean loadOS(URL osURL)
	{
		try
		{
			InputStream osImage = osURL.openStream();
			byte[] rgbImage = getROM(osImage);
			osImage.close();
			if (null==rgbImage)
				return	false;
			m_rgbOSROM = rgbImage;
			reset();
			return true;
		}
		catch(Exception e)
		{
			return false;
		}
	}

	public boolean loadCartridge(URL cartURL)
	{
		try
		{
			InputStream	cartImage = cartURL.openStream();
			boolean fRetVal = loadCartridge(cartImage);
			cartImage.close();
			return	fRetVal;
		}
		catch(Exception e)
		{
			return false;
		}
	}

	public boolean loadCartridge(InputStream is)
	{
		byte[] rgbImage = getROM(is);
		if (null==rgbImage)
			return	false;

		if (!isCartridgeROM(rgbImage))
			return false;

		m_rgbCartridgeROM = rgbImage;
		reset();
		return true;
	}

	private boolean isCartridgeROM(byte rgbImage[])
	{
		if (	((rgbImage[0]==(byte)(0x55)) && (rgbImage[1]==(byte)(0xAA))) ||
				((rgbImage[0]==(byte)(0xAA)) && (rgbImage[1]==(byte)(0x55)))
			)
			return	true;

		return	false;
	}

	private byte[] getROM(InputStream ROMcode)
	{
		try
		{
			byte[]		rgbImage;
			byte			rgbTempImage[] = new byte[32768];
			int			nWriteOffset = 0;
			int			nBytesRead = 0;

			for(;;)
			{
				int	nBytesAvail = ROMcode.available();
				if (0==nBytesAvail)
					break;

				nBytesRead = ROMcode.read(rgbTempImage, nWriteOffset, nBytesAvail);
				if (-1==nBytesRead)
					break;
				
				nWriteOffset += nBytesRead;
				SetStatusText("Retrieving ROM image...(" + nWriteOffset + ")", 0);
				if (nWriteOffset > 32768)
					break;
			}

			if (	nWriteOffset<=0	||
					nWriteOffset>32768
				)
				return null;

			// Create true size array of ROM image
			rgbImage = new	byte[nWriteOffset];
			System.arraycopy(rgbTempImage, 0, rgbImage, 0, nWriteOffset);

			return	rgbImage;
		}

		catch (IOException ioe)
		{
			SetStatusText(ioe.toString(), 0);
			return null;
		}
	}

	public boolean init(Component c, IVideoRefreshNotify refreshSink)
	{
		SetStatusText("Initializing...", 0);

		super.init();
		if (null==c)
			return	false;
		
		m_parent = c;
		m_refreshSink = refreshSink;

		// Initialize the Z80 cpu emulation core (attaching us to it)
		m_z80cpu = new Z80CPU();
		if (!m_z80cpu.init(this))
			return false;

		// Create and map into device ports:
		//		o	TMS9928A video emulation device
		//		o	SN76489 sound emulation device
		//		o	ColecoVision hand controller emulation device
		m_tms9928a = new TMS9928A(m_parent);
		m_sn76489 = new SN76489();
		m_handCtls = new CVHandCtrls();

		// Register the video controller emulation core
		if (!MapIoDevice( (char)0xA0, (char)0x20, m_tms9928a, true, true ))
			return false;

		// Register the hand controller emulation core
		if (!MapIoDevice( (char)0xE0, (char)0x20, m_handCtls, true, false))
			return false;
		if (!MapIoDevice( (char)0xC0, (char)0x20, m_handCtls, false, true))
			return false;
		if (!MapIoDevice( (char)0x80, (char)0x20, m_handCtls, false, true))
			return false;

		// Register the sound controller emulation module
		if (!MapIoDevice( (char)0xE0, (char)0x20, m_sn76489, false, true ))
			return false;

		SetStatusText("Booting...", 0);
		// Default to original ColecoVision Operating Sytem ROM and
		//		the Donkey Kong Jr cartridge ROM. Illegal as hell!!! <BFG>
		m_rgbOSROM = m_rgbCOLECO_ROM;
//		DemoCartridge dc = new DemoCartridge();
//		m_rgbCartridgeROM = dc.getROMImage();

		// Reset the system...
		reset();

		// Determine m_fLockCPUSpeed control data
		m_lSliceEndConst = 1; // Any Java Equiv to ==> (PerfomanceCounterFreq / 1000L) ???
		m_nLockedCPUCyclesToBurn = (int)(m_CVISION_ORIGINAL_HZ / 1000L); //(CV's per ms rate)
		m_lSliceEndConst *= 10L;
		m_nLockedCPUCyclesToBurn *= 10L;

		m_fInitialized = true;

		return	true;
	}

	public boolean shutdown()
	{
		super.shutdown();
		return	true;
	}

	public void start()
	{
		if (null==m_coreEmulation)
			m_coreEmulation = new Thread(this);
		m_coreEmulation.start();		
	}

	public void stop()
	{
		m_coreEmulation.stop();
	}

	// Thread entry point
	public void run()
	{
		//	We have two similiar loops here.  They
		//		are almost identical except one of them
		//		has extra code in to keep the emulated
		//		Mhz at 3.58Mhz or below.  Two loops were needed
		//		so that we don't have the extra "if"
		//		checking happening if the emulation speed
		//		isn't supposed to be locked.
		if (m_fLockCPUSpeed)
		{
			while (true)
			{
				try
				{
					if (!m_fPaused)
					{
						long lContinueTime = System.currentTimeMillis() + m_lSliceEndConst;
						runEmulation(m_nLockedCPUCyclesToBurn);
						
						for (;;)
						{
							if (System.currentTimeMillis() >= lContinueTime)
								break;
							m_coreEmulation.yield();
						}
					}
					else
						m_coreEmulation.sleep(55);
				}
				catch(InterruptedException ie)
				{
					return;
				}
			}
		}
		else while (true)
		{
			try
			{
				if (!m_fPaused)
				{
					runEmulation(m_CYCLES_TO_RUN);
					m_coreEmulation.yield();
				}
				else
					m_coreEmulation.sleep(55);
			}
			catch(InterruptedException ie)
			{
				return;
			}
		}
	}

	private void SetStatusText(String s, long lPersistMilli)
	{
		m_szMsg = s;

		if (lPersistMilli==0)
			lPersistMilli = 2000;  // 2 seconds minimum by default

		m_lMsgHideTime = System.currentTimeMillis() + lPersistMilli;

		// Tell refresh sink we need a repaint...
		if (null!=m_refreshSink)
			m_refreshSink.RefreshNotification();
	}

	private char runEmulation(int nNumCyclesToRun)
	{
		synchronized(m_z80cpu)
		{
			// If cycles to execute will not cause a video refresh to happen atleast once
			//		then let emulation blaze...
			if (nNumCyclesToRun < m_nRefreshCounter)
			{
				m_nRefreshCounter -= nNumCyclesToRun;
				return	m_z80cpu.timedCycle( nNumCyclesToRun );
			}

			// Video refresh will definately occur during cycle execution
			while (nNumCyclesToRun > m_nRefreshCounter)
			{
				m_z80cpu.timedCycle( m_nRefreshCounter );
				nNumCyclesToRun -= m_nRefreshCounter;
				
				// Emulate a video refresh (possibly setting Z80 emulation NMI line)
				m_tms9928a.EventNotification(m_z80cpu);

				// Notify callback that video refresh is happening...
				m_refreshSink.RefreshNotification();
							
				// If we should adjust NMI frequency to be in
				//		sync with CPU cycles per second...
				if (m_TSM_TIMEDCPU == m_nTimingSyncMethod)
				{
					long lCurTick = System.currentTimeMillis();

					// If NMI_RESYNC_TIME has passed then recalculate
					//		the CPU NMI generation
					if (m_nNextRefreshTime <= lCurTick)
					{
						int	nCycleCur = m_z80cpu.getCycleCount();

						// This will work as long as an the NMI_RESYNC_TIME
						//		isn't too long of a period and the emulation
						//		doesn't go over 4,000Mhz.  Boy, we'll have to
						//		worry about going over 4,000 Mhz on today's
						//		Java VMs won't we?  NOT!!
						m_nCyclesSinceRefresh = nCycleCur - m_nLastRefreshCycle;

						// Don't do dynamic unless above minimum base speed
						if (m_nCyclesSinceRefresh >= m_MIN_CYCLES_DYNAMIC)
						{
							m_nDynamicRefreshCounter = m_nCyclesSinceRefresh / m_nRefreshFreq;
							m_nDynamicRefreshCounter /= m_NMI_RESYNC_ADJUSTMENT;

							// Setup for next timing...
							m_nNextRefreshTime = ((int)System.currentTimeMillis()) + m_NMI_RESYNC_TIME;
							m_nLastRefreshCycle = nCycleCur;

	//						System.out.println("Estimated CPU Mhz: " + 
	//							((double) m_nCyclesSinceRefresh) / ((double)m_NMI_RESYNC_TIME) / 1000.0);
	//								sprintf(	(char*) szBuf, "Estimated CPU Mhz: %3.2f\r\n"
	//											"   > m_nDynamicRefreshCounter = %ld\r\n",
	//											((double) m_nCyclesSinceRefresh) / ((double)NMI_RESYNC_TIME) / 1000.0,
	//											m_nDynamicRefreshCounter );
	//								OutputDebugString( szBuf );
						}
					}
				
					// Use last dynamic refresh timing result to reset refresh
					//		counter.
					m_nRefreshCounter = m_nDynamicRefreshCounter;
				}
				else
					m_nRefreshCounter = m_CVISION_ORIGINAL_HZ / m_nRefreshFreq;
			}

			// Run remaining cycles out...
			if (nNumCyclesToRun>0)
			{
				m_z80cpu.timedCycle( nNumCyclesToRun );
				m_nRefreshCounter -= nNumCyclesToRun;
			}

			return	m_z80cpu.getPC();
		}
	}
}

